/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "RMatrixBase.hh"
#include <float.h>
#include <Base/Common/CompareFloatingPoint.hh>
#include <Base/Geometry/Quaternion.hh>


#ifdef WIN32
#  define _USE_MATH_DEFINES
#endif //WIN32

#include <math.h>

// if rotation angle is below this value, assume zero rotation
#define MINIMUM_ROTATION 1e-6

// if real part of quaternion = cos(alpha) is larger than 1+thisvalue
// GetQuaternion produces an error and returns -1
#define MAX_QUATERNION_COS_ERROR 1e-12


#include <Base/Common/BIASpragma.hh>



using namespace BIAS;
using namespace std;

RMatrixBase::RMatrixBase(const RMatrixBase& r)
  : Matrix3x3<ROTATION_MATRIX_TYPE>(r)
{}

RMatrixBase::RMatrixBase(const TNT::Matrix<ROTATION_MATRIX_TYPE>& mat)
  : Matrix3x3<ROTATION_MATRIX_TYPE>(mat)
{
  if ((mat.num_rows() != 3) || (mat.num_cols() != 3)) {
    BIASERR("wrong matrix size "<<mat.num_rows()<<"x"
           << mat.num_cols());
  }
}

RMatrixBase::~RMatrixBase() 
{}

void
RMatrixBase::SetXYZ(ROTATION_MATRIX_TYPE PhiX, ROTATION_MATRIX_TYPE PhiY, 
                    ROTATION_MATRIX_TYPE PhiZ)
{
  RMatrixBase Rx; Rx.SetIdentity();
  RMatrixBase Ry; Ry.SetIdentity();
  RMatrixBase Rz; Rz.SetIdentity();
  
  /* set Rz, Ry, Rx as rotation matrices */
  Rz[0][0] = cos(PhiZ);
  Rz[0][1] = -sin(PhiZ);
  Rz[1][0] = -Rz[0][1];
  Rz[1][1] = Rz[0][0];
  //Rz[2][2] = 1;

  Ry[0][0] = cos(PhiY);
  Ry[0][2] = sin(PhiY);
  //Ry[1][1] = 1;
  Ry[2][0] = -Ry[0][2]; //-sin(PhiY);
  Ry[2][2] = Ry[0][0]; //cos(PhiY);
  
  //Rx[0][0] = 1;
  Rx[1][1] = cos(PhiX);
  Rx[1][2] = -sin(PhiX);
  Rx[2][1] = -Rx[1][2]; //sin(PhiX);
  Rx[2][2] = Rx[1][1]; //cos(PhiX);
  
//   Matrix3x3<ROTATION_MATRIX_TYPE> Result=Rx*Ry*Rz;
//   for (int i =0; i<9; i++)
//     this->GetData()[i]=Result.GetData()[i];
  (*this) = Rx*Ry*Rz;
}

void
RMatrixBase::SetZYX(ROTATION_MATRIX_TYPE PhiX, ROTATION_MATRIX_TYPE PhiY, 
                    ROTATION_MATRIX_TYPE PhiZ)
{
  RMatrixBase Rx; Rx.SetIdentity();
  RMatrixBase Ry; Ry.SetIdentity();
  RMatrixBase Rz; Rz.SetIdentity();
  
  /* set Rz, Ry, Rx as rotation matrices */
  Rz[0][0] = cos(PhiZ);
  Rz[0][1] = -sin(PhiZ);
  Rz[1][0] = -Rz[0][1];
  Rz[1][1] = Rz[0][0];
  //Rz[2][2] = 1;

  Ry[0][0] = cos(PhiY);
  Ry[0][2] = sin(PhiY);
  //Ry[1][1] = 1;
  Ry[2][0] = -Ry[0][2]; //-sin(PhiY);
  Ry[2][2] = Ry[0][0]; //cos(PhiY);
  
  //Rx[0][0] = 1;
  Rx[1][1] = cos(PhiX);
  Rx[1][2] = -sin(PhiX);
  Rx[2][1] = -Rx[1][2]; //sin(PhiX);
  Rx[2][2] = Rx[1][1]; //cos(PhiX);
  
  (*this) = Rz*Ry*Rx;
}


RMatrixBase::RMatrixBase(Vector3<ROTATION_MATRIX_TYPE> w, 
                         ROTATION_MATRIX_TYPE phi)
{
  Set(w, phi);
}


void RMatrixBase::Set(Vector3<ROTATION_MATRIX_TYPE> w, 
                      ROTATION_MATRIX_TYPE phi)
{
#ifdef BIAS_DEBUG
    double wnorm=1.0-w.NormL2();
    if (fabs(wnorm)>1e-6){
	BIASERR("Rotation axis w must be an unit vector (L2 norm is "
		<< wnorm << ").");
	BIASABORT;
    }
#endif

  Matrix3x3<ROTATION_MATRIX_TYPE> Omega, I, A;
  I.SetIdentity();

  Omega[0][0] = 0.0;
  Omega[0][1] = -w[2];
  Omega[0][2] = w[1];
  Omega[1][0] = w[2];
  Omega[1][1] = 0.0;
  Omega[1][2] = -w[0];
  Omega[2][0] = -w[1] ;
  Omega[2][1] = w[0];
  Omega[2][2] = 0.0;

  Matrix3x3<ROTATION_MATRIX_TYPE> Sin_O, O_O, Cos_O_O;

  Sin_O = Omega *  sin(phi);

  Cos_O_O = (Omega * Omega) * (1-cos(phi));
  
  A = Matrix3x3<ROTATION_MATRIX_TYPE>(I + Sin_O + Cos_O_O);
  (*this) = RMatrixBase(I + Sin_O + Cos_O_O);

}

void RMatrixBase::SetFromAxisAngle(Vector3<ROTATION_MATRIX_TYPE> w)
{
  double dAngle = w.NormL2();
  if (fabs(dAngle)<MINIMUM_ROTATION) {
    // we need a rotaion axis even in case of no rotation:
    w[0] = 1;
    dAngle = 0;
  }
  w.Normalize();
  Set(w, dAngle);
}

int RMatrixBase::GetQuaternion(Quaternion<ROTATION_MATRIX_TYPE>& quat) const
{
  int res=0;

#ifdef WIN32

  // nicked from OpenSG
  // corrected by fw, orientation of 'axis' = q0 to q2 was wrong
  double tr;
  double s;
  double qt[3];

  unsigned int i;
  unsigned int j;
  unsigned int k;

  unsigned int nxt[3] = { 1, 2, 0};

  tr = (*this)[0][0] + (*this)[1][1] + (*this)[2][2];

  if(tr > 0.0){
    s = sqrt(tr + 1.0);

    quat[3] = s * 0.5;

    //s = 0.5 / s;
//     quat[0] = ((*this)[2][1] - (*this)[1][2]) * s;
//     quat[1] = ((*this)[0][2] - (*this)[2][0]) * s;
//     quat[2] = ((*this)[1][0] - (*this)[0][1]) * s;
    s *= 2.0;
    quat[0] = ((*this)[2][1] - (*this)[1][2]) / s;
    quat[1] = ((*this)[0][2] - (*this)[2][0]) / s;
    quat[2] = ((*this)[1][0] - (*this)[0][1]) / s;
    
    //cerr << "trace > 0"<<endl;
  } else {
    //cerr << "trace <= 0"<<endl;
    if((*this)[1][1] > (*this)[0][0])
      i = 1;
    else
      i = 0;

    if((*this)[2][2] > (*this)[i][i])
      i = 2;

    j = nxt[i];
    k = nxt[j];

    s = sqrt((*this)[i][i] - ((*this)[j][j] + (*this)[k][k]) + 1.0 );

    qt[i] = s * 0.5;
//     s     = 0.5 / s;

//     quat[3] = ((*this)[k][j] - (*this)[j][k] ) * s;

//     qt[j] = ((*this)[i][j] + (*this)[j][i]) * s;
//     qt[k] = ((*this)[i][k] + (*this)[k][i]) * s;
    s     *= 2.0;

    quat[3] = ((*this)[k][j] - (*this)[j][k] ) / s;

    qt[j] = ((*this)[i][j] + (*this)[j][i]) / s;
    qt[k] = ((*this)[i][k] + (*this)[k][i]) / s;

    quat[0] = qt[0];
    quat[1] = qt[1];
    quat[2] = qt[2];
  }

  /**if(quat[3] > 1.0 || quat[3] < -1.0){
    if (fabs(quat[3]-1.0)> MAX_QUATERNION_COS_ERROR){
      BIASERR(" |quat[3]-1.0| ("<< fabs(quat[3]-1.0) <<") >> 0.0 ! "<<quat<<"R: "<<(*this)<<endl);
      BIASABORT;
      res=-1;
    }
    if(quat[3] > 1.0){
      quat[3] = 1.0;
    } else {
      quat[3] = -1.0;
    }
    }*/
  // normalize
  quat /= quat.NormL2();


#else
  // http://www.euclideanspace.com/maths/geometry/rotations/conversions/matrixToQuaternion/index.htm
  quat[3] = sqrt( max( 0.0, 1.0 + (*this)[0][0] + (*this)[1][1] + (*this)[2][2] ) ) / 2.0; 
  quat[0] = sqrt( max( 0.0, 1.0 + (*this)[0][0] - (*this)[1][1] - (*this)[2][2] ) ) / 2.0; 
  quat[1] = sqrt( max( 0.0, 1.0 - (*this)[0][0] + (*this)[1][1] - (*this)[2][2] ) ) / 2.0; 
  quat[2] = sqrt( max( 0.0, 1.0 - (*this)[0][0] - (*this)[1][1] + (*this)[2][2] ) ) / 2.0; 
  quat[0] = copysign( quat[0], (*this)[2][1] - (*this)[1][2] ) ;
  quat[1] = copysign( quat[1], (*this)[0][2] - (*this)[2][0] ) ;
  quat[2] = copysign( quat[2], (*this)[1][0] - (*this)[0][1] ) ;

 // normalize
  quat /= quat.NormL2();

  //BIASASSERT(Equal(quat.NormL2(), 1.0));

#endif
  return res;
}

int
RMatrixBase::GetRotationAxisAngle(Vector3<ROTATION_MATRIX_TYPE>& Axis, 
                                  ROTATION_MATRIX_TYPE& Angle) const
{
  // nicked from OpenSG
  Quaternion<ROTATION_MATRIX_TYPE> quat;
  if (GetQuaternion(quat)!=0)
    return -1;
  return quat.GetAxisAngle(Axis, Angle);
}

int RMatrixBase::GetRotationAxisAngle(Vector3<ROTATION_MATRIX_TYPE>& w) const
{
  double angle;
  // nicked from above
  Quaternion<ROTATION_MATRIX_TYPE> quat;
  if (GetQuaternion(quat)!=0)     return -1;
  quat.GetAxisAngle(w , angle);
  w.Normalize();
  //get a three-parameter representation by encoding the angle into the axis:
  w *= angle;
  return 0;
}


int
RMatrixBase::GetRotationAnglesXYZ(double& PhiX, double& PhiY, double& PhiZ)
{
  if (((*this)[0][2] == 1.0)||((*this)[0][2] == -1.0)) {
    BIASWARN("Rotation for Y-axis is +-PI/2, "
             "cannot decompose X-/Z-component!");
    return -1;
  }
  PhiY = asin((*this)[0][2]);
  if (isnan(PhiY) || isinf(PhiY)) return -1;
  // changed this from atan to atan2 (koeser)
  PhiX = atan2(-(*this)[1][2],(*this)[2][2]);
  if (isnan(PhiX) || isinf(PhiX)) return -2;
  PhiZ = atan2(-(*this)[0][1],(*this)[0][0]);
  if (isnan(PhiZ) || isinf(PhiZ)) return -3;
// #ifdef BIAS_DEBUG
//   RMatrixBase R;
//   R.SetXYZ(PhiX, PhiY, PhiZ);
//   if (R == (*this))
//     res = 0;
//   else
//     res = -1;
// #endif
  return 0;
}


int
RMatrixBase::GetRotationAnglesZYX(double& PhiX, double& PhiY, double& PhiZ)
{
  if (((*this)[0][2] == 1.0)||((*this)[0][2] == -1.0)) {
    BIASWARN("Rotation for Y-axis is +-PI/2, "
             "cannot decompose X-/Z-component!");
    return -1;
  }
  PhiY = asin(-(*this)[2][0]); 
  if (isnan(PhiY) || isinf(PhiY)) return -1;
  // changed this from atan to atan2 (koeser)
  PhiX = atan2((*this)[2][1],(*this)[2][2]);
  if (isnan(PhiX) || isinf(PhiX)) return -2;
  PhiZ = atan2((*this)[1][0],(*this)[0][0]);
  if (isnan(PhiZ) || isinf(PhiZ)) return -3;
// #ifdef BIAS_DEBUG
//   RMatrixBase R;
//   R.SetZYX(PhiX, PhiY, PhiZ);
//   if (R == (*this))
//     res = 0;
//   else
//     res = -1;
// #endif
  return 0;
}

RMatrixBase& RMatrixBase::operator= (const RMatrixBase &mat)
{
  Matrix<ROTATION_MATRIX_TYPE>::operator=(mat);
  return *this;
}

bool BIAS::operator==(const RMatrixBase& left, const RMatrixBase& right)
{
  bool res=true;
  for (register int i=0; i<9; i++){
    if (fabs(left.GetData()[i]-right.GetData()[i])>DBL_EPSILON){
      res=false;
      break;
    }
  }
  return res;
}



int RMatrixBase::SetFromQuaternion(const Quaternion<ROTATION_MATRIX_TYPE> &qu) 
{
  // fast version, see below for old code:
  const double norm = 1.0/qu.NormL2();
  const double q0 = qu[0]*norm;
  const double q1 = qu[1]*norm;
  const double q2 = qu[2]*norm;
  const double q3 = qu[3]*norm;

  double* pData = GetData();
  
  *pData++ = 1.0 - 2.0 * (q1 * q1 + q2 * q2);
  *pData++ =       2.0 * (q0 * q1 - q2 * q3);
  *pData++ =       2.0 * (q2 * q0 + q1 * q3);

  *pData++ =       2.0 * (q0 * q1 + q2 * q3);
  *pData++ = 1.0 - 2.0 * (q2 * q2 + q0 * q0);
  *pData++ =       2.0 * (q1 * q2 - q0 * q3);

  *pData++ =       2.0 * (q2 * q0 - q1 * q3);
  *pData++ =       2.0 * (q1 * q2 + q0 * q3);
  *pData++ = 1.0 - 2.0 * (q1 * q1 + q0 * q0);


#ifdef WANT_TO_COMPARE_WITH_OLD_CODE
  
  Quaternion<ROTATION_MATRIX_TYPE> q = qu; // de-const
  q.Normalize();
  // nicked from OpenSG
  RMatrixBase RCompare(MatrixZero);
  RCompare[0][0] = 1.0 - 2.0 * (q[1] * q[1] + q[2] * q[2]);
  RCompare[1][0] =       2.0 * (q[0] * q[1] + q[2] * q[3]);
  RCompare[2][0] =       2.0 * (q[2] * q[0] - q[1] * q[3]);

  RCompare[0][1] =       2.0 * (q[0] * q[1] - q[2] * q[3]);
  RCompare[1][1] = 1.0 - 2.0 * (q[2] * q[2] + q[0] * q[0]);
  RCompare[2][1] =       2.0 * (q[1] * q[2] + q[0] * q[3]);

  RCompare[0][2] =       2.0 * (q[2] * q[0] + q[1] * q[3]);
  RCompare[1][2] =       2.0 * (q[1] * q[2] - q[0] * q[3]);
  RCompare[2][2] = 1.0 - 2.0 * (q[1] * q[1] + q[0] * q[0]);

  for (unsigned int i=0; i<3; i++) {
    for (unsigned int j=0; j<3; j++) {
      if (!Equal((*this)[i][j], RCompare[i][j], 1e-12)) {
        cout<<"Matrices are not same !"<<*this<<" "<<RCompare<<endl;
        cout<<" at "<<i<<" "<<j<<" "<< ((*this)[i][j])-(RCompare[i][j])<<endl;
        BIASABORT;
      }
    }
  }
#endif

  return 0;
}


int RMatrixBase::Set(Quaternion<ROTATION_MATRIX_TYPE> &q)
{

  BIASERR("RMatrixBase::Set(Quaternion) fixed (tansposed compared to old)"
	  " Use SetFromQuaternion(), now");
  BIASABORT;
  return -1;
}


void  RMatrixBase::SetFromOrthogonalBasis(
  const BIAS::Vector3<ROTATION_MATRIX_TYPE> & xxx,
  const BIAS::Vector3<ROTATION_MATRIX_TYPE> & yyy,
  const BIAS::Vector3<ROTATION_MATRIX_TYPE> & zzz )
{
  SetFromRowVectors( xxx.GetNormalized(), 
                     yyy.GetNormalized(), 
                     zzz.GetNormalized() );
}


// JW
void RMatrixBase::SetFromHV(const BIAS::Vector3<ROTATION_MATRIX_TYPE> & xxx,
			    const BIAS::Vector3<ROTATION_MATRIX_TYPE> & yyy )
{
  // TODO, FIXME: 
  /// \todo check, wether xxx, yyy are linear independant.

  // normalize
  const BIAS::Vector3<ROTATION_MATRIX_TYPE> x0 ( xxx.GetNormalized() );
  
  // compute base z in rhs:
  const BIAS::Vector3<ROTATION_MATRIX_TYPE> 
    z0 ( xxx.CrossProduct(yyy).GetNormalized() );
  
  // compute y base orthogonal in rhs
  const BIAS::Vector3<ROTATION_MATRIX_TYPE> 
    y0 ( z0.CrossProduct(x0).GetNormalized() );
  
  (*this).SetFromColumnVectors(x0,  y0,  z0 );
}

int RMatrixBase::SetFromOriUpGL(BIAS::Vector3<ROTATION_MATRIX_TYPE> ori,
                                BIAS::Vector3<ROTATION_MATRIX_TYPE> up)
{
  //calculate orthogonal vectors
  BIAS::Vector3<double> right;
  ori.Normalize();
  ori.CrossProduct(up,right);
  right.Normalize();
  right.CrossProduct(ori,up);
  up.Normalize();
  ori*=-1.0; // we need the right handed system
  SetFromColumnVectors(right,up,ori);

  return 0;
}
