/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef _BIAS_IMAGE_DRAW_H_
#define _BIAS_IMAGE_DRAW_H_
#include "bias_config.h" 

#ifdef BIAS_HAVE_OPENCV
// for font names
#  include <cxcore.h> 
//#  include <cv.h> 
#endif

#include "Image.hh"
#include <Base/Math/Vector2.hh>
#include <Base/Math/Vector3.hh>

namespace BIAS {
    /** @class ImageDraw
    @ingroup g_image_other
    @brief drawing simple entities into the image like rectangles or lines


    As all fucntions are static they have to be called by
    eg. ImageDraw::RectangleCenter() and not with an object.
    */
    template <class StorageType=unsigned char>
    class BIASImageBase_EXPORT ImageDraw 
    {
        /** class for Drawing in Images
        @author fw */
    public:
        /* @brief returns value with good contrast to t */
        static StorageType GetContrastValue(StorageType t);

        ////////////////////////////////////////////////////////////
        /// rectangles
        ////////////////////////////////////////////////////////////

        ////////////////////////////////////
        // specifying rectangle corners

        /** @brief Draws the rectangle defined by upper left and bottom right
        corner with value[i] in channel i **/
        static int RectangleCorners(Image<StorageType>& im, unsigned int minx, 
            unsigned int miny, unsigned int maxx, 
            unsigned int maxy, 
            const StorageType value[]);

        /** @brief as above but fills all channels with value */
        static int RectangleCorners(Image<StorageType>& im, unsigned int minx, 
            unsigned int miny, unsigned int maxx, 
            unsigned int maxy, 
            const StorageType value);

        /** @brief decides which color to use */
        static int RectangleCorners(Image<StorageType>& im, unsigned int minx, 
            unsigned int miny, unsigned int maxx, 
            unsigned int maxy);

        /** @brief faster for one channel images */
        static int RectangleCornersGrey(Image<StorageType>& im, unsigned int minx, 
            unsigned int miny, unsigned int maxx, 
            unsigned int maxy, 
            const StorageType GreyValue);

        /** @brief faster for one channel images */
        static int RectangleCornersGreyFill(Image<StorageType>& im, 
            unsigned int minx, unsigned int miny, 
            unsigned int maxx, unsigned int maxy, 
            const StorageType GreyValue);
        //////////////////////////
        // specifying the rectangle center and size

        /** @brief Draws the rectangle around X, Y with Size and Value[i] 
        in channel i
        @author Felix Woelk **/
        static int RectangleCenter(Image<StorageType>& im,
            unsigned int x, unsigned int y, 
            unsigned int size, 
            const StorageType value[]);

        /** @brief fills the channels with value */
        static int RectangleCenterGrey(Image<StorageType>& im,
            unsigned int x, unsigned int y, 
            unsigned int size, 
            const StorageType value);

        static int RectangleCenterGreyFill(Image<StorageType>& im,
            unsigned int x, unsigned int y, 
            unsigned int size, 
            const StorageType value);

        /** convenience wrapper */
        inline 
            static int RectangleCenter(Image<StorageType>& im, unsigned int coo[2], 
            unsigned int size, 
            const StorageType value[])
        { return RectangleCenter(im, coo[0], coo[1], size, value); };

        /** @brief draws a rectangle of size around x and y, 
        decides which color to use */
        static int RectangleCenter(Image<StorageType>& im, unsigned int x, 
            unsigned int y, unsigned int size);

        /** convenience wrapper */
        static int RectangleCenter(Image<StorageType>& im, unsigned int coo[2], 
            unsigned int size);


        //////////////////////////////////////////////////////////////
        /// lines
        /////////////////////////////////////////////////////////////


        /** @brief Draws a line between the two points using the bresenham 
        algorithm uses value[i] in channel i **/
        static int Line(Image<StorageType>& im, const unsigned int start[2],
            const unsigned int end[2], 
            const StorageType value[]);
        /// differnet interface for the function above
        static int Line(Image<StorageType>& im, const int start[2],
            const int end[2], 
            const StorageType value[]);

        /** @brief uses value in every channel */
        static int Line(Image<StorageType>& im, const unsigned int start[2],
            const unsigned int end[2], 
            const StorageType value);

        /** @brief faster algorithm for 1 channel images */
        static int LineGrey(Image<StorageType>& im, const unsigned int start[2],
            const unsigned int end[2], 
            const StorageType value);

        /** @brief decides which color to use */
        static int Line(Image<StorageType>& im, const unsigned int start[2],
            const unsigned int end[2]);

        /** @brief faster algorithm for 1 channel images, 
        decides which color to use */
        static int LineGrey(Image<StorageType>& im, const unsigned int start[2],
            const unsigned int end[2]);

        //////////////////////////
        // just wrappers


        /** Draw a solid line with a specific thickness.
        Internally implemented as drawing multiple Bresenham lines, for now. 
        \bug This routine is a HACK and incorrect. A generela "solid fill by 4 given corners" algorithm has to be used or a Bresenham with eps=thckness/2.
        @param Value the line color.
        @param thickness of the line in pixel (perpendicualr to the direction). 
        @author Jan Woetzel 01/2005 
        DEPRECATED by JW 2005/08/09, replaced by OpenCV routine call. 
        */
        static int Line(Image<StorageType>& im, 
            const unsigned int StartX, 
            const unsigned int StartY, 
            const unsigned int EndX, 
            const unsigned int EndY, 
            const StorageType color[],
            const float thickness );


        /** convenience wrapper */
        static int Line(Image<StorageType>& im, const unsigned int StartX, 
            const unsigned int StartY, const unsigned int EndX, 
            const unsigned int EndY, 
            const StorageType Value[]);

        /** convenience wrapper */
        static int Line(Image<StorageType>& im, const unsigned int StartX, 
            const unsigned int StartY, const unsigned int EndX, 
            const unsigned int EndY, 
            const StorageType Value);

        /** convenience wrapper */
        static int Line(Image<StorageType>& im, const unsigned int StartX, 
            const unsigned int StartY, const unsigned int EndX, 
            const unsigned int EndY);

        /** convenience wrapper */
        static int LineGrey(Image<StorageType>& im, const unsigned int StartX, 
            const unsigned int StartY, const unsigned int EndX, 
            const unsigned int EndY, 
            const StorageType Value);

        /** convenience wrapper */
        inline static int Line(Image<StorageType>& im, unsigned int coo[4],
            const StorageType value[])
        { return Line(im, &coo[0], &coo[2], value); };

        /** @brief Convenience wrapper, draws a line even if start or end point are
        not positioned in the image.
        untested so far
        @author woelk 06/2005 */
        static int Line(Image<StorageType>& im, const double start[2], 
            const double end[2], 
            const StorageType value[]);

        ///////////////////////////
        // interpolated lines

        /** @brief draws an interpolated line 
        just a wrapper for the below
        @author Felix Woelk */
        static int InterpolatedLine(Image<StorageType>& im, 
            const int StartX, const int StartY,
            const int EndX, const int EndY,
            const StorageType Value[]);

        static int InterpolatedLineGrey(Image<StorageType>& im, 
            const int StartX, const int StartY,
            const int EndX, const int EndY,
            const StorageType Value);

        inline 
            static int InterpolatedLineGrey(Image<StorageType>& im, 
            unsigned int coo[4],
            const StorageType Value)
        { return InterpolatedLineGrey(im, coo[0], coo[1], coo[2], 
        coo[3], Value); };


        ////////////////////////////////////
        // arrows

        /**  @brief draws an arrow from start to end, the tips of the head are 
        length pixel back on the line and width pixel away from the line
        @author woelk 12/2003 */
        static int Arrow(Image<StorageType>& im, const unsigned start[2],
            const unsigned end[2], const unsigned length,
            const unsigned width, 
            const StorageType value[]);


        /////////////////////////////////////////
        // circles
        /** @brief draws a circular line, either using Value or 
        a good contrast value */  
        static int CircleCenter(Image<StorageType>& im, unsigned int CenterX,
            unsigned int CenterY, unsigned int Radius, 
            const StorageType Value[]=NULL);

        /** @brief draws a filled circle using Value */ 
        static int CircleCenterFilled(Image<StorageType>& im, unsigned int CenterX,
            unsigned int CenterY, unsigned int Radius, 
            const StorageType Value[]);


        //////////////////////////////////////////////////////////
        // ellipses
        //////////////////////////////////////////////////////////

        /** @brief draws an ellipse at center with half axes a and b
        @author woelk 05/2005 */
        static int Ellipse(Image<StorageType>& im, double center[2],
            double a[2], double b[2], 
            const StorageType Value[]);

        /** @TODO
        static int InterpolatedLineColor(Image<StorageType>& im, 
        const int StartX, const int StartY,
        const int EndX, const int EndY,
        StorageType Value[]); */




#ifdef BIAS_HAVE_OPENCV
        //////////////////////////////////////////////////////////
        // wrapper(s) for OpenCV draw functions (Jan Woetzel 2005)
        //////////////////////////////////////////////////////////

        // define some olors for easier usage: 
#define COLOR_RED    BIAS::Vector3<unsigned char>(255,0,0)
#define COLOR_GREEN  BIAS::Vector3<unsigned char>(0,255,0)
#define COLOR_BLUE   BIAS::Vector3<unsigned char>(0,0,255)
#define COLOR_WHITE  BIAS::Vector3<unsigned char>(255,255,255)
#define COLOR_GREY50 BIAS::Vector3<unsigned char>(128, 128, 128)
#define COLOR_BLACK  BIAS::Vector3<unsigned char>(0,0,0)

        /** @brief OpenCV: Draw Text into image 
        @param message text to display.
        @param posX (bottom) left corner of textbox, may be negative and outside image
        @param posY bottom (left) corner of textbox, may be negative and outside image
        @param colorRGB color to use
        @param fontface e.g. CV_FONT_HERSHEY_SIMPLEX, CV_FONT_HERSHEY_PLAIN, CV_FONT_HERSHEY_COMPLEX
        @author Jan Woetzel 2005 */
        static void Text(BIAS::Image<StorageType>& dstImg,
            const std::string & message,
            const int & posX=0,
            const int & posY=20,
            const BIAS::Vector3<StorageType> &colorRGB=Vector3<StorageType>(255,255,255),
            const int fontface=CV_FONT_HERSHEY_PLAIN,
            const double hscale=1.0,
            const double vscale=1.0,
            const double shear=0,
            const int thickness=1,
            const int linetype=8
            );

        /** OpenCV: Draws a circle 
        @author JW 2005 */
        static void Circle(BIAS::Image<StorageType>& dstImg,
            const int & centerX, 
            const int & centerY,
            const int & radius=3,
            const BIAS::Vector3<StorageType> &colorRGB=Vector3<StorageType>(255,255,255),                        
            const int & thickness=1, 
            const int & linetype=8, 
            const int & shift=0 
            );

        /** OpenCV: Draws a line segment connecting two points 
        @author JW 2005 */
        static void Line(BIAS::Image<StorageType>& dstImg,
            const int & pt1X,
            const int & pt1Y,                        
            const int & pt2X, 
            const int & pt2Y,
            const BIAS::Vector3<StorageType> &colorRGB=Vector3<StorageType>(255,255,255),                        
            const int thickness=1,
            const int linetype=8,
            const int shift=0 
            );

        /** @brief  OpenCV: Draws a simple, thick or filled rectangle @author JW 2005 
            @authior Jan Woetzel 2005 */
        static void Rectangle( 
            BIAS::Image<StorageType>& dstImg,
            const int & pt1X,
            const int & pt1Y,                        
            const int & pt2X, 
            const int & pt2Y,
            const BIAS::Vector3<StorageType> &colorRGB=Vector3<StorageType>(255,255,255),                        
            const int thickness=1,
            const int linetype=8,
            const int shift=0 
            );

        /** OpenCV: Draws simple or thick elliptic arc or fills ellipse sector 
            angles are given in degree 
            @author JW 2005  */
        static void Ellipse(
            BIAS::Image<StorageType>& dstImg,
            const int & centerX,
            const int & centerY,
            const int & axesX,
            const int & axesY,
            const double & angle,
            const double & start_angle=0, 
            const double & end_angle=360, 
            const BIAS::Vector3<StorageType> &colorRGB=Vector3<StorageType>(255,255,255),
            const int & thickness=1, 
            const int & linetype=8, 
            const int & shift=0 
            );

#endif

    }; // class ImageDraw

}  // namespace

#endif // _BIAS_IMAGE_DRAW_H_
