/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef _Matrix3x3_hh_
#define _Matrix3x3_hh_
#include "bias_config.h" 

#include <string>

#include <Base/Debug/Error.hh>
//#include <Base/Debug/Debug.hh>

#include "Matrix.hh"
#include <Base/Math/Matrix3x4.hh>
#include <Base/Math/Vector3.hh>

namespace BIAS {


  /** @class Matrix3x3
      @ingroup g_math
      @brief is a 'fixed size' quadratic matrix of dim. 3 x 3 which is 
      templated over the element-type.
      
      @author Jan Woetzel
      @status untested (04/17/2002)*/
  template<class T=double>
  class BIASMathBase_EXPORT Matrix3x3 : public Matrix<T> {

  public:
    /**	default constructor
        @status untested **/
    inline Matrix3x3<T>() : Matrix<T>(3, 3) {};

    // DEPRECATED due to instantiation problem JW
    //explicit Matrix3x3<T>(const char *s) : Matrix<T>(3, 3, s) {};
    /// JW
    explicit Matrix3x3<T>(const std::string & s);

    /** Set the matrix values JW */
    Matrix3x3<T>(
      const T a0, const T a1, const T a2, 
      const T a3, const T a4, const T a5,
      const T a6, const T a7, const T a8) ;


    /** @brief init matrix with identity or zero 
        @author koeser */
    explicit Matrix3x3<T>(const MatrixInitType& i) : Matrix<T>(3, 3, T(0)) {
      switch (i) {
      case MatrixZero:return;
      case MatrixIdentity:{ 
        this->v_[0] = this->v_[4] = this->v_[8] =T(1);
        return;
      }
      default: 
        BIASERR("undefined MatrixInitType"); 
        BIASABORT;
      }
    }

    /** @author Jan Woetzel
        @status untested (04/17/2002) **/
    Matrix3x3<T>(const Matrix3x3<T>& A) : Matrix<T>(A) {};

    /** cast constructor from Matrix */
    Matrix3x3<T>(const Matrix<T>& A);

    /** cast constructor from TNT::Matrix */
    Matrix3x3<T>(const TNT::Matrix<T>& A);

    /** destructor
        @status untested (04/17/2002)**/
    virtual ~Matrix3x3<T>();

    
    /** matrix - vector multiplicate this matrix with Vector3, 
        storing the result in destvec
        calculates:
        destvec = (this Matrix) * argvec
        @author Ingo Thomsen, Jan Woetzel
        @status untested (04/17/2002)**/
    inline void Mult(const Vector3<T> &argvec, Vector3<T> &destvec) const;

    /** multiplies matrix from left with transposed argvec, resulting
        in transposed destvec
        @author woelk 05/2003 */
    inline 
    void TransposedMult(const Vector3<T> &argvec, Vector3<T> &destvec) const;

    /** matrix-matrix multiplication with other Matrix3x3, 
        storing the result in destmat
        calculates:
        destmat = (this mat) * argmat
        @author Ingo Thomsen, Jan Woetzel
        @status untested (04/17/2002)**/
    inline void Mult(const Matrix3x3<T> & argmat,Matrix3x3<T> & destmat) const;
    /** @author woelk 06 2003 */
    inline void Mult(const Matrix3x4<T> & argmat,Matrix3x4<T> & destmat) const;

    /** scalar-matrix multiplication */
    inline void Scale(const T &scalar, Matrix3x3<T> & destmat) const;
   
    /** assignment operator
        calls corresponding operator from base class 
        @author Felix Woelk **/
    inline Matrix3x3<T>& operator=(const Matrix3x3<T>& mat);


    /**	tranpose this matrix "in place"
        example: 
        0 1 2  -->  0 3 6
        3 4 5  -->  1 4 7
        6 7 8  -->  2 5 8
        @author Jan Woetzel
        @status untested (04/17/2002)**/
    inline void TransposeIP();

    /** returns transposed matrix
        tested 12.06.2002
        @author Felix Woelk    */
    Matrix3x3<T> Transpose() const;

    /** sets this as transposed arg,  fw */
    inline void Transpose(const Matrix3x3<T>& arg); 

    /** set the elements of this matrix to the identity matrix
        (posisbly overriding the inherited method)
        @author Ingo Thomsen, Jan Woetzel
        @date 04/17/2002
        @status untested **/
    inline void SetIdentity();

    /// stl conform interface
    inline void clear() {
      this->SetZero();
    };


    /** sets the diagonalelements of this 3x3 Matrix
        rowwise with the values of the 9 (x1) vector
        @author Ingo Thomsen, Jan Woetzel
    **/
    void SetFromVector(const TNT::Vector<T> & vec);
    
    /** set this matrix from 3 vectors each representating a column
      	JW 07/2003 */
    void SetFromColumnVectors(const BIAS::Vector3<T> & v0,
                              const BIAS::Vector3<T> & v1,
                              const BIAS::Vector3<T> & v2);
                              
    /** set this matrix from 3 vectors, each representating a row
      	JW 07/2003 */                          
    void SetFromRowVectors(const BIAS::Vector3<T> & v0,
                           const BIAS::Vector3<T> & v1,
                           const BIAS::Vector3<T> & v2);
                           
    /** extract one row ('Zeile') from ths matrix (for convenience) JW */
    void GetRow(const unsigned int row, Vector3<T>& r) const;
    inline Vector3<T> GetRow(const unsigned int row) const
    { 
      Vector3<T> m; 
      GetRow(row, m); 
      return m; 
    }

    /** extract one column ('Spalte') from this matrix (for convenience) JW */
    BIAS::Vector3<T> GetColumn(const unsigned int col) const;


    /** calculate the determinante
        @author Ingo Thomsen
        @status alpha (04/17/2002)*/
    inline T det() const;

    /** just neccessary to avoid resizing of this 'fixed size' matrix because
        it is derived from the resizable Matrix
        should be removed when Matrix3x3 becomes a base class.
        (04/17/2002) Jan Woetzel
    **/
    Matrix3x3 & newsize(int rows, int cols);

    /** Sets matrix from vector as cross product matrix of this vector. */
    inline void SetAsCrossProductMatrix(const Vector3<T> &vec);

    /** Sets matrix from vector as cross product matrix of a vector. 
        with components (x, y, z) */
    inline void SetAsCrossProductMatrix(const T &x, const T &y, const T &z);

    /** Complete inversion 
        @author Jan-Friso Evers-Senne
        2002-08-27 */
    Matrix3x3<T> Invert() const;

    /** In place matrix conversion. 
        Returns -1 if determinant is zero, 0 on success. 
        @author woelk 07/2005 */
    int InvertIP();

    /** Matrix inversion: inverts this and stores resulty in argument inv.
        Returns -1 if determinant is zero, 0 on success.
        @author woelk 07/2005 */
    int Invert(Matrix3x3<T>& inv) const;
    
    /** return the trace of the matrix
        @author woelk 04/2006 */
    inline T Trace() const
    { const register T *p = this->GetData();	return p[0] + p[4] + p[8]; }

    /** returns the Determinant |A| of this
        @author grest  */
    T GetDeterminant() const;

    /** return biggest and smallest entry
        @author woelk 08/2004 */
    inline void MaxMin(T& max, T&min) const;

    /** reads the RC part of a BOG file 
        which is used by Daimler Chrysler 
        for storing camera orientation/rotation.
        @return 0 for success, negative for errror
        @author Jan Woetzel 03/2005 
        @todo what do the fields RotMatInRowColumn and ReferenceFrameCurrentCam field mean? transpose? JW
        */
    int LoadBogRC(const std::string & filename);

    /** reads the internal KMatrix part of a BOG file 
        @return 0 for success, negative for errror
        @author Jan Woetzel 03/2005 */
    int LoadBogK(const std::string & filename);


  }; // class Matrix3x3 --------------------------------------






  template <class T> inline void 
  Matrix3x3<T>::Mult(const Vector3<T> &argvec, Vector3<T> &destvec) const
  {
    // direct access to the matrix values using fixed indicies:
    // 0  1  2       0
    // 3  4  5   *   1
    // 6  7  8       2
    const register T *matP = this->GetData();	
    destvec[0] = argvec[0] * matP[0]
      + argvec[1] * matP[1]
      + argvec[2] * matP[2];
    destvec[1] = argvec[0] * matP[3]
      + argvec[1] * matP[4]
      + argvec[2] * matP[5];
    destvec[2] = argvec[0] * matP[6]
      + argvec[1] * matP[7]
      + argvec[2] * matP[8];
  }

  template <class T> inline void 
  Matrix3x3<T>::TransposedMult(const Vector3<T> &argvec, Vector3<T> &destvec) const
  {
    // direct access to the matrix values using fixed indicies:
    // 0  1  2       0
    // 3  4  5   *   1
    // 6  7  8       2    
    const register T *matP = this->GetData();	
    destvec[0] = argvec[0] * matP[0]
      + argvec[1] * matP[3]
      + argvec[2] * matP[6];
    destvec[1] = argvec[0] * matP[1]
      + argvec[1] * matP[4]
      + argvec[2] * matP[7];
    destvec[2] = argvec[0] * matP[2]
      + argvec[1] * matP[5]
      + argvec[2] * matP[8];
  }

  template <class T> inline void 
  Matrix3x3<T>::Mult(const Matrix3x3<T> & argmat, Matrix3x3<T> & destmat) const
  {
    //
    // direct access using fixed indicies with matrix order:
    //  0 1 2
    //  3 4 5
    //  6 7 8
    register T *destP 	= destmat.GetData(); 
    const register T *argP	= argmat.GetData(); 
    const register T *matP	= this->GetData(); 
    // does not work for in place multiplication
    BIASASSERT(destP!=argP);
    BIASASSERT(destP!=matP);

    destP[0] =
      matP[0] * argP[0] + matP[1] * argP[3] + matP[2] * argP[6];
    destP[1] =
      matP[0] * argP[1] + matP[1] * argP[4] + matP[2] * argP[7];
    destP[2] =
      matP[0] * argP[2] + matP[1] * argP[5] + matP[2] * argP[8];

    destP[3] =
      matP[3] * argP[0] + matP[4] * argP[3] + matP[5] * argP[6];
    destP[4] =
      matP[3] * argP[1] + matP[4] * argP[4] + matP[5] * argP[7];
    destP[5] =
      matP[3] * argP[2] + matP[4] * argP[5] + matP[5] * argP[8];

    destP[6] =
      matP[6] * argP[0] + matP[7] * argP[3] + matP[8] * argP[6];
    destP[7] =
      matP[6] * argP[1] + matP[7] * argP[4] + matP[8] * argP[7];
    destP[8] =
      matP[6] * argP[2] + matP[7] * argP[5] + matP[8] * argP[8];
  }

  template <class T> inline void 
  Matrix3x3<T>::Mult(const Matrix3x4<T> & argmat, Matrix3x4<T> & destmat) const
  {
    //
    // direct access using fixed indicies with matrix order:
    //  0 1 2
    //  3 4 5
    //  6 7 8
    // oder 3x4
    //  0 1 2  3
    //  4 5 6  7
    //  8 9 10 11
    T *destP 	= (T*) (destmat.GetData()); 
    const T *argP	= (T*)(argmat.GetData()); 
    const register T *matP	= this->GetData(); 
    // does not work for in place multiplication
    BIASASSERT(matP!=argP);
    BIASASSERT(matP!=destP);

    destP[0] =
      matP[0] * argP[0] + matP[1] * argP[4] + matP[2] * argP[8];
    destP[1] =
      matP[0] * argP[1] + matP[1] * argP[5] + matP[2] * argP[9];
    destP[2] =
      matP[0] * argP[2] + matP[1] * argP[6] + matP[2] * argP[10];
    destP[3] =
      matP[0] * argP[3] + matP[1] * argP[7] + matP[2] * argP[11];

    destP[4] =
      matP[3] * argP[0] + matP[4] * argP[4] + matP[5] * argP[8];
    destP[5] =
      matP[3] * argP[1] + matP[4] * argP[5] + matP[5] * argP[9];
    destP[6] =
      matP[3] * argP[2] + matP[4] * argP[6] + matP[5] * argP[10];
    destP[7] =
      matP[3] * argP[3] + matP[4] * argP[7] + matP[5] * argP[11];

    destP[8] =
      matP[6] * argP[0] + matP[7] * argP[4] + matP[8] * argP[8];
    destP[9] =
      matP[6] * argP[1] + matP[7] * argP[5] + matP[8] * argP[9];
    destP[10] =
      matP[6] * argP[2] + matP[7] * argP[6] + matP[8] * argP[10];
    destP[11] =
      matP[6] * argP[3] + matP[7] * argP[7] + matP[8] * argP[11];
  }


  template <class T> inline void 
  Matrix3x3<T>::Scale(const  T &scalar, Matrix3x3<T> & destmat) const
  {
    register T *dest = destmat.GetData();
    const register T *src	= this->GetData(); 
    for (int i=0; i<9; i++)
      dest[i] = scalar * src[i];
  }



  template <class T> inline void 
  Matrix3x3<T>::TransposeIP()
  {
    // swap the 1,2,5 with the 3,6,7
    // 0  1  2
    // 3  4  5
    // 6  7  8
    register T *dataP = this->GetData();	// pointer to the data array
    register T tmp;
    // swap 1,3:
    tmp = dataP[1];
    dataP[1] = dataP[3];
    dataP[3] = tmp;
    // swap 2,6:
    tmp = dataP[2];
    dataP[2] = dataP[6];
    dataP[6] = tmp;
    // swap 5,7:
    tmp = dataP[5];
    dataP[5] = dataP[7];
    dataP[7] = tmp;
  }
   
  template <class T> inline void 
  Matrix3x3<T>::Transpose(const Matrix3x3<T>& arg)
  {
    register T *argp = arg.GetData();
    this->v_[0] = argp[0]; this->v_[4] = argp[4]; this->v_[8] = argp[8];
    this->v_[1] = argp[3]; this->v_[2] = argp[6]; this->v_[3] = argp[1];
    this->v_[5] = argp[7]; this->v_[6] = argp[2]; this->v_[7] = argp[5];
  }


  template <class T> 
    inline 
    void  Matrix3x3<T>::SetIdentity()
  {
    register T* d = this->GetData(); // pointer to this matrix's data array
    d[0] = 1;
    d[1] = 0;
    d[2] = 0;
    d[3] = 0;
    d[4] = 1;
    d[5] = 0;
    d[6] = 0;
    d[7] = 0;
    d[8] = 1;
  }


  template <class T> inline T 
  Matrix3x3<T>::det() const
  {
    register T *dataP = this->GetData(); // pointer to the data array
    return (
            dataP[0] * dataP[4] * dataP[8]
            + dataP[1] * dataP[5] * dataP[6]
            + dataP[2] * dataP[3] * dataP[7]
            - dataP[6] * dataP[4] * dataP[2]
            - dataP[7] * dataP[5] * dataP[0]
            - dataP[8] * dataP[3] * dataP[1] );
  }


  template<> inline void 
  Matrix3x3<unsigned int>::SetAsCrossProductMatrix(const Vector3<unsigned int> & /*vec*/){
    BIASERR("not implemented because minus does not make sense for unsigned type");
    BIASBREAK;
  }

  template <class T> inline void 
  Matrix3x3<T>::SetAsCrossProductMatrix(const Vector3<T> &vec)
  {
    T *mat=this->GetData();
    mat[0]=mat[4]=mat[8]=0;
    mat[1]=-vec[2]; mat[2]=vec[1];
    mat[3]= vec[2]; mat[5]=-vec[0];
    mat[6]= -vec[1]; mat[7]=vec[0];
  }


  template<> inline 
  void Matrix3x3<unsigned int>::SetAsCrossProductMatrix(const unsigned int & /*x*/, const unsigned int & /*y*/, const unsigned int & /*z*/) {
    BIASERR("not implemented because minus does not make sense for unsigned type");
    BIASBREAK;
  }
  
  
  template <class T> inline   
  void Matrix3x3<T>::SetAsCrossProductMatrix(const T &x, const T &y, const T &z)
  {
    T *mat=this->GetData();
    mat[0]=mat[4]=mat[8]=0;
    mat[1]= -z; mat[2]= y;
    mat[3]=  z; mat[5]=-x;
    mat[6]= -y; mat[7]= x;
  }



  ///////////////////////// non class functions

  template <class T>
    inline std::istream& operator>>(std::istream &s, Matrix3x3<T> &A)
  {
    int M, N;

    s >> M >> N;

    if (M!=3 || N!=3) {
      s.setstate(std::ios::badbit | std::ios::failbit);
      return s;
    }

    if ( !(M == A.num_rows() && N == A.num_cols() ))
      A.newsize(M,N);

    for (int i=0; i<M; i++)
      for (int j=0; j<N; j++)
        s >>  A[i][j];

    return s;
  }

  template<class T>
  inline Matrix3x3<T>& Matrix3x3<T>::operator= (const Matrix3x3<T>& mat) 
  {
    Matrix<T>::operator=(mat);
    return *this;
  }

  template<class T>
  inline void Matrix3x3<T>::MaxMin(T& max, T& min) const 
  {
    register T *d=this->GetData();
    max = min = *d;

    if (d[1]>max) max=d[1];
    if (d[2]>max) max=d[2];
    if (d[3]>max) max=d[3];
    if (d[4]>max) max=d[4];
    if (d[5]>max) max=d[5];
    if (d[6]>max) max=d[6];
    if (d[7]>max) max=d[7];
    if (d[8]>max) max=d[8];

    if (d[1]<min) min=d[1];
    if (d[2]<min) min=d[2];
    if (d[3]<min) min=d[3];
    if (d[4]<min) min=d[4];
    if (d[5]<min) min=d[5];
    if (d[6]<min) min=d[6];
    if (d[7]<min) min=d[7];
    if (d[8]<min) min=d[8];
  }



} // end of namespace BIAS


#include "Operators.hh"

#endif

