/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __Pose_hh__
#define __Pose_hh__

#include <Base/Geometry/Quaternion.hh>
#include <Base/Geometry/HomgPoint3D.hh>
#include <Base/Geometry/HomgPoint3DCov.hh>
#include <Base/Math/Matrix4x4.hh>
#include <Base/Math/Matrix3x4.hh>
#include <Base/Geometry/RMatrixBase.hh>
#include <Geometry/CoordinateTransform3D.hh>
#include <Base/Geometry/PoseParametrization.hh>


#define POSE_TYPE double

namespace BIAS {

  /** @class Pose
      @brief Container for 3D pose, parametrized as euclidean center and
      quaternion orientation. 
      Special coordinate transformation that can also be interpreted as 
      measureable entity. Meaning: a (camera) pose is measured in the 
      coordinate system A. In A the pose will define the transformation from 
      coordinate in A to the pose (camera) coordinate system. This 
      transformation does not contain scale. Imagine A beeing embedded in 
      another coordinate system B. In order to find the pose in B the method
      TransformBy() can be used. The result will not be able to transform 
      coordinates
      from A into the pose system. This transform can be found using the 
      method CoordinateTransform3D::ConcatenateLocalTransform() 
      on a pose object.      
      
      @author woelk/koeser/bartczak 03/2006 */

  class BIASGeometry_EXPORT Pose : public CoordinateTransform3D
  {
  public:
    Pose();

    ~Pose();

    int Set(const PoseParametrization& pose);

    void Set(const Quaternion<POSE_TYPE>& Q,
             const Vector3<POSE_TYPE>& C);

    void Set(const Quaternion<POSE_TYPE>& Q, 
             const Vector3<POSE_TYPE>& C,
             const Matrix<POSE_TYPE>& cov);

    void SetC(const Vector3<POSE_TYPE>& C, 
              const Matrix3x3<POSE_TYPE>& cov);
    
    //    void SetC(const HomgPoint3D& C, const HomgPoint3DCov& cov);

    void SetQ(const Quaternion<POSE_TYPE>& Q, 
              const Matrix4x4<POSE_TYPE>& cov);

    /** resets covariance matrix to zero! */
    void SetQ(const Quaternion<POSE_TYPE>& Q);

    /** resets covariance matrix to zero */
    void SetC(const Vector3<POSE_TYPE>& C);

    PoseParametrization GetPoseParameters() const;

    int SetCovariance(const Matrix<POSE_TYPE>& c);

    /// Set internal covariance fron an Euclidian covariance matrix
    /// Elements 0,1,2 are C, Elements 3,4,5 are rotation angles
    int SetCovEuclidian(Matrix<POSE_TYPE> Cov);

    /// returns the 7x7 covariance matrix
    inline const Matrix<POSE_TYPE>& GetCovariance() const 
    { return Cov_; }

    Matrix3x3<POSE_TYPE> GetCovC() const;

    Matrix4x4<POSE_TYPE> GetCovQ() const;

    bool Load(const std::string& inputFile);

    friend std::istream& operator>>(std::istream &is, 
                                    Pose& p);
        
    const Pose& operator=(const Pose& p) {
      CoordinateTransform3D::operator=(p);
      Cov_ = p.GetCovariance();
      return (*this);
    }

    /** \brief Transforms the position and orientation stored in the pose
     * into another global coordinate system. The argument trsf is hereby
     * assumed to be the global system for the current global.
     * So that the result is : trsf*(this) = new local system.
     */
    void TransformToNewGlobalBy(const CoordinateTransform3D& trsf);

  protected:
    using CoordinateTransform3D::SetScale;

    // 7x7 covariance matrix
    Matrix<POSE_TYPE> Cov_;
  };


  inline std::istream& operator>>(std::istream &is, Pose& p){
    std::string temp;
    is>>temp>>p.Q_>>temp>>p.C_>>temp>>p.scale_;
    if(is && p.scale_!=1.0) {
      BIASERR("Loaded transformation is not a pose - resetting scale!\n");
      p.scale_ = 1.0;
    }
    BIASERR("writing of covariance matrix is missing");
    return is;
  }

}

#endif
