/** 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "Geometry/ProjectionParametersCylindric.hh" 
#include <Base/Common/CompareFloatingPoint.hh>

using namespace std;
using namespace BIAS;

void BIAS::ProjectionParametersCylindric::
Print() const
{
  cout << "ProjectionParametersCylindric::Print() :" << endl;
  cout << "- Pose = " << endl;
  Pose_.Talk();
  cout << "- focallenghtX_ = " << focallengthX_ << endl;
  cout << "- focallenghtPhi_ = " << focallengthPhi_ << endl;
  cout << "- principalX_ = " << principalX_ << endl;
  cout << "- principalY_ = " << principalY_ << endl;
  cout << "- xMin_ = " << xMin_ << endl;
  cout << "- xMax_ = " << xMax_ << endl;
  cout << "- phiMin_ = " << phiMin_ << endl;
  cout << "- phiMax_ = " << phiMax_ << endl;
  cout << "- width_ = " << width_ << endl;
  cout << "- height_ = " << height_ << endl << endl;
  cout << "- cylinder axis adaption rotation axis is "
       << RotAxis_[0] << "," << RotAxis_[1] << "," << RotAxis_[2]
       << ", rotation angle is " << 180.0*RotAngle_/M_PI << ", translation is "
       << Translate_[0] << "," << Translate_[1] << "," << Translate_[2]
       << "." << endl;

}

void BIAS::ProjectionParametersCylindric::
UpdateQuasiK_()
{
  //normal coordinates are mapped into image as follows: 
  //(x, phi) -> (  x * focallengthX_   + principalX_, 
  //             phi * focallengthPhi_ + principalY_)  
  focallengthX_ = double(width_)/(xMax_-xMin_);
  //xMin_ is mapped to first column
  principalX_ = (-1.0)*xMin_*focallengthX_;
  focallengthPhi_ = double(height_)/(phiMax_-phiMin_);
  //phiMin_ is mapped to first row
  principalY_ = (-1.0)*phiMin_*focallengthPhi_;
}


void BIAS::ProjectionParametersCylindric::
UpdateFromQuasiK_() 
{
  xMin_ = (-1.0)*principalX_/focallengthX_;
  xMax_ = double(width_)/focallengthX_ + xMin_;
  phiMin_ = (-1.0)*principalY_/focallengthPhi_;
  phiMax_ = double(height_)/focallengthPhi_ + phiMin_;
}


HomgPoint2D BIAS::ProjectionParametersCylindric::
ProjectLocal(const BIAS::Vector3<double> &point,
             bool IgnoreDistortion) const
{
//if (IgnoreDistortion) {
//  scale * point is intersection with "w=1" image plane
  HomgPoint2D res;
  if (Equal(point.NormL2(), 0.0)) {
    res.SetZero();
    return res;
  }

  // Translate and rotate point vector according to the cylinder's orientation
  Vector3<double> p;
  CylinderInv_.Mult(point - Translate_, p);

  // Calculate 3D intersection point of vector with cylinder hull
  if (Equal(p[1], 0.0) && Equal(p[2], 0.0))
    return HomgPoint2D(0, 0, 0);
  double scale = 1.0/sqrt(p[1]*p[1] + p[2]*p[2]);
  Vector2<double> xPhi;
  ProjectToNormalizedCoords_(scale*p, xPhi);

  // Calculate 2D coordinates of intersection point in cylinder hull image
  res[0] = xPhi[0]*focallengthX_+principalX_;
  res[1] = xPhi[1]*focallengthPhi_+principalY_;
  res[2] = 1.0;
  return res;
//}
//else {
//  BIASERR("Distortion is not implemented for cylindric projections.");
//  BIASABORT;
//}
//return HomgPoint2D(0, 0, 0);
}


HomgPoint3D ProjectionParametersCylindric::
UnProjectToImagePlane(const HomgPoint2D& pos,
                      const double& depth,
                      bool IgnoreDistortion) const 
{
  Vector3<double> p = UnProjectLocal(pos, IgnoreDistortion);
  double radius = sqrt(p[2]*p[2] + p[1]*p[1]);
  if (Equal(radius, 0.0))
    return HomgPoint3D(0, 0, 0, 0);
  p *= depth / radius;
  if (Equal(p.NormL2(), 0.0))
    return HomgPoint3D(0, 0, 0, 0);
  HomgPoint3D local(p);
  return Pose_.LocalToGlobal(local);
}


Vector3<double> BIAS::ProjectionParametersCylindric::
UnProjectLocal(const HomgPoint2D &pos, bool IgnoreDistortion) const
{
  // Normalize unprojected 3D point (resulting from 2D point pos) to length 1
  Vector3<double> res = UnProjectToCylinderPoint(pos, IgnoreDistortion);
  res.Normalize();
  return res;
}


Vector3<double> BIAS::ProjectionParametersCylindric::
UnProjectToPointLocal(const HomgPoint2D &pos, const double &depth,
                      bool IgnoreDistortion) const
{
  return (UnProjectLocal(pos, IgnoreDistortion) * depth);
}


Vector3<double> BIAS::ProjectionParametersCylindric::
UnProjectToCylinderPoint(const HomgPoint2D &pos,
                         bool IgnoreDistortion) const
{
  // Calculate 3D point on hull of cylinder with radius 1 relative to
  // the camera center, where 2D point given lies on cylinder hull image.
  Vector3<double> p;
  double phi;
  p[0] = (pos[0]-principalX_)/focallengthX_;
  phi = (pos[1]-principalY_)/focallengthPhi_;
  p[1] = sin(phi);
  p[2] = cos(phi);
  // Rotate resulting vector according to the cylinder's orientation
  Vector3<double> res;
  CylinderMod_.Mult(p, res);
  // Translate according to cylinder's translation
  res = res + Translate_;
  return res;
}


const bool BIAS::ProjectionParametersCylindric::
DoIntrinsicsDiffer(const ProjectionParametersBase* p) const
{
  if (ProjectionParametersBase::DoIntrinsicsDiffer(p)) {
    return true;
  }
  const ProjectionParametersCylindric* ppc = 
    dynamic_cast<const ProjectionParametersCylindric*>(p);
  if (ppc == NULL) {
    return true;
  }
  if (focallengthPhi_ != ppc->focallengthPhi_) {
    return true;
  }
  if (focallengthX_ != ppc->focallengthX_) {
    return true;
  }
  return false;
}


bool BIAS::ProjectionParametersCylindric::
DoesPointProjectIntoImageLocal(const BIAS::Vector3<double> &localP, 
                               HomgPoint2D &p, bool IgnoreDistortion) const
{
  p = ProjectLocal(localP, IgnoreDistortion);
  p.Homogenize(); //neccessary ?
  return (p[0] >= 0.0 && p[1] >= 0.0 && 
          p[0] <= (double)(width_-1) && p[1] <= (double)(height_-1));
}


void BIAS::ProjectionParametersCylindric::
SetCylinder(const Vector3<double> &rotAxis, const double alpha,
            const Vector3<double> &translate)
{
  RotAxis_   = rotAxis;
  RotAngle_  = alpha;
  Translate_ = translate;
  CylinderMod_ = RMatrix(RotAxis_,  RotAngle_);
  CylinderInv_ = RMatrix(RotAxis_, -RotAngle_);
}


void BIAS::ProjectionParametersCylindric::
SetCylinder(const ProjectionCylinderAxisEnum axisType,
            const Vector3<double> &translate)
{
  Vector3<double> w(0.0, 1.0, 0.0);
  if (axisType == AXIS_ORTHOGONAL) {
    SetCylinder(w, 0.0, translate);
  } else if (axisType == AXIS_PARALLEL) {
    SetCylinder(w, -0.5*M_PI, translate);
  } else {
    BIASERR("ProjectionParametersCylindric::SetCylinder(): Invalid axis type given.");
  }
}


/** @brief Takes the difference of the cylinder centers weighted by
           the cylinders' orientation difference to model the
           view difference. Minimal difference 0 is achived, if
           both cylinder coincide. Difference increases with cylinder
           center distance |M0-M1| and cylinder axis angle difference.
  */
double BIAS::ProjectionParametersCylindric::
ViewDifference(const ProjectionParametersBase* pPPB) const
{
  const ProjectionParametersCylindric *pPPC = 
    dynamic_cast<const ProjectionParametersCylindric* >(pPPB);
  // different projection model has a very different view
  if (pPPC == NULL) {
    BIASERR("View difference for different projection models requested !");
    return DBL_MAX;
  }
  Vector3<double> C0 = pPPC->GetC();
  Vector3<double> C1 = GetC();
  RMatrix R0 = pPPC->GetR();
  RMatrix R1 = GetR();
  Vector3<double> V0 = R0*pPPC->CylinderMod_*Vector3<double>(0,0,1);
  Vector3<double> V1 = R1*CylinderMod_*Vector3<double>(0,0,1);
  Vector3<double> M0 = C0 + V0 * 0.5*(pPPC->xMax_ - pPPC->xMin_);
  Vector3<double> M1 = C1 + V1 * 0.5*(xMax_ - xMin_);
  double phi = acos(V0.ScalarProduct(V1));
  return (M0-M1).NormL2()*(1.0 - fabs(sin(phi)));
}


#ifdef BIAS_HAVE_XML2
 
int ProjectionParametersCylindric::
XMLGetClassName(std::string &TopLevelTag, double &Version) const
{
  TopLevelTag = "ProjectionParametersCylindric";
  Version = 0.1;
  return 0;
}


int ProjectionParametersCylindric::
XMLOut(const xmlNodePtr Node, XMLIO &XMLObject) const
{
  xmlNodePtr theNode;
  string TopLevelName;
  double Version;
  ProjectionParametersBase::XMLGetClassName(TopLevelName, Version);
  theNode = XMLObject.addChildNode(Node, TopLevelName);
  XMLObject.addAttribute(theNode, "Version", Version);
  ProjectionParametersBase::XMLOut(theNode, XMLObject);

  // xmlNodePtr childNode;
  XMLObject.addAttribute(Node, "FocallengthX", focallengthX_);
  XMLObject.addAttribute(Node, "FocallengthPhi", focallengthPhi_);
  Vector<double> rotAxisVec(3);
  rotAxisVec[0] = RotAxis_[0];
  rotAxisVec[1] = RotAxis_[1];
  rotAxisVec[2] = RotAxis_[2];
  Vector<double> translateVec(3);
  translateVec[0] = Translate_[0];
  translateVec[1] = Translate_[1];
  translateVec[2] = Translate_[2];
  XMLObject.addAttribute(Node, "CylinderAxisRotAxis", rotAxisVec);
  XMLObject.addAttribute(Node, "CylinderAxisRotAngle", RotAngle_);
  XMLObject.addAttribute(Node, "CylinderAxisTranslation", translateVec);
  
  return 0; 
}


int ProjectionParametersCylindric::
XMLIn(const xmlNodePtr Node, XMLIO &XMLObject)
{
  string TopLevelName;
  double Version;
  ProjectionParametersBase::XMLGetClassName(TopLevelName, Version);

  xmlNodePtr childNode; 
  if ((childNode = XMLObject.getChild(Node, TopLevelName))==NULL) {
    BIASERR("Error in xml, no tag" << TopLevelName);
    return -1;
  }
  if (ProjectionParametersBase::XMLIn(childNode, XMLObject) != 0) return -1;

  focallengthX_ = XMLObject.getAttributeValueDouble(Node, "FocallengthX");
  focallengthPhi_ = XMLObject.getAttributeValueDouble(Node, "FocallengthPhi");
  Vector<double> rotAxisVec = XMLObject.getAttributeValueVecDbl(Node, "CylinderAxisRotAxis");
  Vector<double> translateVec = XMLObject.getAttributeValueVecDbl(Node, "CylinderAxisTranslation");
  RotAngle_ = XMLObject.getAttributeValueDouble(Node, "CylinderAxisRotAngle");
  RotAxis_[0] = rotAxisVec[0];
  RotAxis_[1] = rotAxisVec[1];
  RotAxis_[2] = rotAxisVec[2];
  Translate_[0] = translateVec[0];
  Translate_[1] = translateVec[1];
  Translate_[2] = translateVec[2];
  SetCylinder(RotAxis_, RotAngle_, Translate_);

  UpdateFromQuasiK_();
  return 0;
}



#endif // BIAS_HAVE_XML2
