/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "Geometry/ProjectionParametersGreatCircles.hh" 
#include <Base/Common/CompareFloatingPoint.hh>

using namespace BIAS;
using namespace std;

#define M_PI_TOL M_PI*1.1
#define M_PI_2_TOL M_PI_2*1.1


ProjectionParametersGreatCircles& ProjectionParametersGreatCircles::
operator=(const ProjectionParametersGreatCircles & P)
{
  ProjectionParametersBase::operator=(P);
  alphaMin_ = P.alphaMin_;
  betaMin_ = P.betaMin_;
  alphaRange_ = P.alphaRange_;
  betaRange_ = P.betaRange_;
  UpdateInternals_();      
  return *this;
}

void ProjectionParametersGreatCircles::
SetAngleRanges(const double alphaRange, 
               const double betaRange) 
{
  BIASASSERT(alphaRange>0.0);
  BIASASSERT(alphaRange<=2.0*M_PI_TOL);
  BIASASSERT(betaRange>0.0);
  BIASASSERT(betaRange<=2.0*M_PI_TOL);

  betaRange_ = betaRange;
  alphaRange_ = alphaRange;

  /** Corrects the possible alpha range and extends beta to wrap a whole
   * sphere -> we don't want to loose data
   */
  if( (alphaMin_+alphaRange_) > M_PI_2 ) {
//     cout<<"found alphaMin = "<<alphaMin_<<" and alphaRange ="
//         << alphaRange_<<" together = "<<(alphaMin_+alphaRange_)
//         <<" correcting...";
    alphaRange_ = M_PI_2 - alphaMin_;
    betaRange_ = 2.0*M_PI;
//     cout<<" -> "<<alphaRange_<<endl;
  }
  BIASASSERT(alphaRange_>0.0);

  UpdateInternals_();
}

void ProjectionParametersGreatCircles::
SetMinAngles(const double alphaMin, 
             const double betaMin) {
  BIASASSERT(alphaMin>-M_PI_2_TOL && alphaMin<M_PI_2_TOL);
  BIASASSERT(betaMin>-M_PI_TOL && betaMin<M_PI_TOL);
  alphaMin_ = alphaMin;
  /** Corrects the possible alpha range and extends beta to wrap a whole
   * sphere -> we don't want to loose data
   */
  if( (alphaMin_+alphaRange_) > M_PI_2 ) {
//     cout<<"found alphaMin = "<<alphaMin_<<" and alphaRange ="
//         << alphaRange_<<" together = "<<(alphaMin_+alphaRange_)
//         <<" correcting...";
    alphaRange_ = M_PI_2 - alphaMin_;
    betaRange_ = 2.0*M_PI;
//     cout<<" -> "<<alphaRange_<<endl;
  }
  BIASASSERT(alphaRange_>0.0);
  betaMin_ = betaMin;
  UpdateInternals_();
}

void ProjectionParametersGreatCircles::
UpdateInternals_() 
{
  betaPP_ = betaMin_ + M_PI;
  focalLengthAlpha_ = (width_-1)/alphaRange_;
  focalLengthBeta_ = (height_-1)/betaRange_;
  DetermineOffsetPose_(betaMin_, offsetPoseBeta_);
}

void ProjectionParametersGreatCircles::
GetCenteredAngleInterval(const HomgPoint3D& worldPoint,
                          const double halfAngleRange,
                          double& minAlpha, double& minBeta,
                          double& alphaRange, double& betaRange) 
{
  double minAlphaBak = alphaMin_;
  double minBetaBak = betaMin_;
  double alphaRangeBak = alphaRange_;
  double betaRangeBak = betaRange_;

  
  SetCenteredAngleInterval(worldPoint,
                           halfAngleRange);
  minAlpha = alphaMin_;
  minBeta = betaMin_;
  alphaRange = alphaRange_;
  betaRange = betaRange_;

  SetMinAngles(minAlphaBak, minBetaBak);
  SetAngleRanges(alphaRangeBak, betaRangeBak);      

}

int ProjectionParametersGreatCircles::
SetCenteredAngleInterval(const HomgPoint3D& worldPoint,
                         const double& halfAngleRange)
{
  double alpha0, beta0;
  Vector3<double> localPoint;
  Pose_.GlobalToLocal(worldPoint).GetEuclidian(localPoint);
  std::cout<<"centering around "<<localPoint<<std::endl;
  if(CamCoordVectorToUnshiftedAngles_(localPoint, alpha0, beta0)!=0)
    return -1;
  std::cout<<"found angles : alpha = "<<alpha0
           <<" beta = "<<beta0<<std::endl;
   
  double alphaMin = alpha0-halfAngleRange;
  double betaMin = beta0-halfAngleRange;
  double rangeAlpha = 2.0*halfAngleRange;
  if(alphaMin<-M_PI_2) {
    rangeAlpha -= (-M_PI_2-alphaMin);
    //    if(rangeAlpha>M_PI) rangeAlpha=M_PI;
    alphaMin = - M_PI_2;
  }
  if(betaMin<-M_PI) betaMin += 2.0*M_PI;
  std::cout<<"calculating minimal angles : alphaMin = "<<alphaMin
           <<" betaMin = "<<betaMin<<std::endl;
  SetMinAngles(alphaMin, betaMin);
  SetAngleRanges(rangeAlpha, 2.0*halfAngleRange);      
  return 0;
}
     

const bool ProjectionParametersGreatCircles::
DoIntrinsicsDiffer(const ProjectionParametersBase* p) const
{
  if(ProjectionParametersBase::DoIntrinsicsDiffer(p))
    return true;
  
  const ProjectionParametersGreatCircles* ppbc = 
    dynamic_cast<const ProjectionParametersGreatCircles*>(p);
  if(ppbc == NULL)
    return true;

  if(ppbc->focalLengthAlpha_ != focalLengthAlpha_)
    return true;
  if(ppbc->focalLengthBeta_ != focalLengthBeta_)
    return true;

  if(ppbc->alphaMin_ != alphaMin_)
    return true;
  if(ppbc->betaMin_ != betaMin_)
    return true;  

  return false;
}


bool ProjectionParametersGreatCircles::
DoesPointProjectIntoImageLocal(const Vector3<double>& localX, 
                               HomgPoint2D& x,                       
                               bool IgnoreDistortion) const
{
  if(localX[2] > 0.0) {
    x = ProjectLocal(localX, IgnoreDistortion);
    if(Equal(x[2], 0.0))
      return false;
    return (x[0]>=0.0 && x[1]>=0.0 && 
            x[0]<=(double)(width_-1) && x[1]<=(double)(height_-1) );
  } 
  return false;
}

//#define ex Vector3<double>(1, 0, 0)

int ProjectionParametersGreatCircles::
CamCoordVectorToUnshiftedAngles_(Vector3<double> vec, 
                                 double& alpha, double& beta) const
{
  if(Equal(vec.NormL2(), 0.0))
    return -1;

  // vec.Normalize();
  Vector3<double> yzProjection(0.0, vec[1], vec[2]);
  if(Equal(yzProjection.NormL2(), 0.0)) {//vector on baseline
    beta = 0;
    alpha = M_PI_2;
    if(vec[0]<0) alpha*=-1;
    return 0;
  }
  beta = atan2(vec[1], vec[2]);
    
  yzProjection.Normalize();
  Vector2<double> betaPlaneCoords(vec[0], 
                                  yzProjection[1]*vec[1] + 
                                  yzProjection[2]*vec[2]);
  
  alpha = atan2(betaPlaneCoords[0], betaPlaneCoords[1]);
    
  return 0;  
}

int ProjectionParametersGreatCircles::
CamCoordVectorToShiftedAngles_(Vector3<double> vec, 
                               double& alpha, double& beta) const
{
  if(Equal(vec.NormL2(), 0.0))
    return -1;
  
  //instead of using global to local:
//   vec.Set(vec[0], 
//           cos(betaPP_)*vec[1]+sin(betaPP_)*vec[2],
//           -1 * sin(betaPP_)*vec[1]+cos(betaPP_)*vec[2]);
  HomgPoint3D hVec(vec);
  offsetPoseBeta_.GlobalToLocal(hVec).GetEuclidian(vec);
  
  // vec.Normalize();
  Vector3<double> yzProjection(0.0, vec[1], vec[2]);
  if(Equal(yzProjection.NormL2(), 0.0)) {//vector on baseline
    beta = 0;
    alpha = M_PI_2;
    if(vec[0]<0) alpha*=-1;
    return 0;
  }
  beta = atan2(vec[1], vec[2]);
    
  yzProjection.Normalize();
  Vector2<double> betaPlaneCoords(vec[0], 
                                  yzProjection[1]*vec[1] + 
                                  yzProjection[2]*vec[2]);
  
  alpha = atan2(betaPlaneCoords[0], betaPlaneCoords[1]);
  
  alpha -= alphaMin_;
  beta  += M_PI;
  
  return 0;  
}

int ProjectionParametersGreatCircles::
ShiftedAnglesToCamCoordVector_(double alpha, double beta, 
                               Vector3<double>& vec) const
{

  alpha += alphaMin_;
  beta  -= M_PI;

//   cout<<"alpha to vec = "<<alpha<<endl;
//   cout<<"beta to vec = "<<beta<<endl;

  
  BIASASSERT(GreaterEqual(alpha, -M_PI_2_TOL));
  BIASASSERT(LessEqual(alpha, M_PI_2_TOL));
  BIASASSERT(GreaterEqual(beta, -M_PI_TOL));
  BIASASSERT(LessEqual(beta, M_PI_TOL));

  

  HomgPoint3D shiftedVector;
  shiftedVector.Set(sin(alpha), 
                    sin(beta)*cos(alpha), 
                    cos(beta)*cos(alpha), 1.0);
  shiftedVector = offsetPoseBeta_.LocalToGlobal(shiftedVector);
  shiftedVector.GetEuclidian(vec);

  return 0;
}

Vector2<double> ProjectionParametersGreatCircles::
ShiftedAnglesToImage(double alpha, double beta) const
{
  BIASASSERT(focalLengthAlpha_>0.0);
  BIASASSERT(focalLengthBeta_>0.0);

  return Vector2<double>(alpha*focalLengthAlpha_, beta*focalLengthBeta_);
}

int ProjectionParametersGreatCircles::
ImageToShiftedAngles(Vector2<double>& imgPoint,
                     double& alpha, double& beta) const
{
  BIASASSERT(focalLengthAlpha_>0.0);
  BIASASSERT(focalLengthBeta_>0.0);
  alpha = imgPoint[0]/focalLengthAlpha_;

  if(alpha < 0.0)
    return -1;
  if(alpha > alphaRange_)
    return -1;

  beta = imgPoint[1]/focalLengthBeta_;
  if(beta < 0)
    return -1;
  if(beta > betaRange_)
    return -1;

  return 0;
}

int ProjectionParametersGreatCircles::
DetermineOffsetPose_(double betaMin, 
                     Pose& localOffsetPose) const
{
  BIASASSERT(GreaterEqual(betaMin, -M_PI_TOL));
  BIASASSERT(LessEqual(betaMin, M_PI_TOL));

  double betaRotAngle = betaMin + M_PI;
  
  Quaternion<double>  betaQ;
  betaQ.SetValueAsAxisRad( 1.0, 0.0, 0.0, -betaRotAngle);
  
  localOffsetPose.Set(betaQ, Vector3<double>(0.0, 0.0, 0.0));
  
  return 0;
}


HomgPoint2D ProjectionParametersGreatCircles::
ProjectLocal(const Vector3<double>& point, bool IgnoreDistortion) const 
{
  double alpha, beta;
  if(CamCoordVectorToShiftedAngles_(point, alpha, beta) != 0) {
    return HomgPoint2D(0,0,0);
    }
  
  return HomgPoint2D(ShiftedAnglesToImage(alpha, beta)); 
  
}

HomgPoint3D ProjectionParametersGreatCircles::
UnProjectToImagePlane(const HomgPoint2D& pos,
                      const double& depth,
                      bool IgnoreDistortion) const 
{
  Vector3<double> p = UnProjectLocal(pos, IgnoreDistortion);
  double norm = p.NormL2();
  if (Equal(norm, 0.0))
    return HomgPoint3D(0,0,0,0);
  HomgPoint3D local(p * (depth/norm));
  local[3] = 1.0;
  return Pose_.LocalToGlobal(local);
}


Vector3<double>  ProjectionParametersGreatCircles::
UnProjectLocal(const HomgPoint2D& cPos, bool IgnoreDistortion) const {
  
  double alpha, beta;
  Vector2<double> m(cPos[0], cPos[1]);
  if(ImageToShiftedAngles(m, alpha, beta) != 0)
    return Vector3<double>(0,0,0);

//   cout<<"m = "<<m<<endl;
  Vector3<double> res;
  ShiftedAnglesToCamCoordVector_(alpha, beta, res);

  return res;
}

void ProjectionParametersGreatCircles::
CheckUnprojectionProjection() 
{
  HomgPoint2D source, reprojection;
  HomgPoint3D unProjection;
  bool consistent = true;
  for(unsigned int y=0; y<height_; y++) {
    for(unsigned int x=0; x<width_; x++) {
      source.Set(x,y,1.0);
      unProjection = HomgPoint3D(UnProjectToRay(source) + GetC());
      reprojection = Project(unProjection);
      if(!Equal((source-reprojection).NormL2(), 0.0)) {
        cout<<"inconsistancy "<<source<<"reprojected to "<<reprojection<<endl;
        consistent = false;
      }
    }
    if(consistent)
      cout<<"everything is conssistent!\n";
  }
}

void ProjectionParametersGreatCircles::
CheckUnprojectionProjectionLocal() 
{
  cout<<"LOCAL CONSISTANCY CHECK\n";
  HomgPoint2D source, reprojection;
  bool consistent = true;
  for(unsigned int y=0; y<height_; y++) {
    for(unsigned int x=0; x<width_; x++) {
      source.Set(x,y,1.0);
      reprojection = ProjectLocal(UnProjectLocal(source));
      if(! (Equal((source-reprojection).NormL2(), 0.0))) {
        cout<<"inconsistancy "<<source<<"reprojected to "
            <<reprojection<<" error = "<<(source-reprojection).NormL2()<<endl;
        consistent = false;
      }
    }
    if(consistent)
      cout<<"everything is conssistent!\n";
  }
}


#ifdef BIAS_HAVE_XML2
 
int ProjectionParametersGreatCircles::XMLGetClassName(std::string& TopLevelTag,
                                                     double& Version) const {
  TopLevelTag = "ProjectionParametersGreatCircles";
  Version = 0.1;
  return 0;
}

int ProjectionParametersGreatCircles::XMLOut(const xmlNodePtr Node, 
                                            XMLIO& XMLObject) const {
  xmlNodePtr theNode;
  string TopLevelName;
  double Version;
  ProjectionParametersBase::XMLGetClassName(TopLevelName, Version);
  theNode = XMLObject.addChildNode(Node, TopLevelName);
  XMLObject.addAttribute(theNode, "Version", Version);
  ProjectionParametersBase::XMLOut(theNode, XMLObject);

  XMLObject.addAttribute(Node,"AlphaOffset", alphaMin_);
  XMLObject.addAttribute(Node,"BetaOffset", betaMin_);
  XMLObject.addAttribute(Node,"AlphaFOV", alphaRange_);
  XMLObject.addAttribute(Node,"BetaFOV", betaRange_);
  return 0; 
 }

 
int ProjectionParametersGreatCircles::XMLIn(const xmlNodePtr Node, 
                                           XMLIO& XMLObject) {

  string TopLevelName;
  double Version;
  ProjectionParametersBase::XMLGetClassName(TopLevelName, Version);

  xmlNodePtr childNode; 
  if ((childNode = XMLObject.getChild(Node, TopLevelName))==NULL) {
    BIASERR("Error in xml, no tag" << TopLevelName);
    return -1;
  }
  if (ProjectionParametersBase::XMLIn(childNode, XMLObject)!=0) return -1;

  SetAngleRanges(XMLObject.getAttributeValueDouble(Node, "AlphaFOV"),
                 XMLObject.getAttributeValueDouble(Node, "BetaFOV"));
  SetMinAngles(XMLObject.getAttributeValueDouble(Node, "AlphaOffset"),
               XMLObject.getAttributeValueDouble(Node, "BetaOffset"));  
  UpdateInternals_();
  return 0;
}

#endif // BIAS_HAVE_XML2



