/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __BIAS_ProjectionParametersGreatCircles_hh__
#define __BIAS_ProjectionParametersGreatCircles_hh__
#include "bias_config.h" 

#include <Base/Common/BIASpragmaStart.hh>
#include <Geometry/ProjectionParametersBase.hh>
#include <Base/Geometry/KMatrix.hh>
#include <Geometry/PMatrix.hh>
#include <Base/Common/CompareFloatingPoint.hh>
#include <iostream>

namespace BIAS {

  /** @class ProjectionParametersGreatCircles
      @brief spherical camera that samples along big circles containig the H 
      vector.
      @author bartczak
  */
  class BIASGeometry_EXPORT ProjectionParametersGreatCircles
    : public  ProjectionParametersBase {
  public:

    ProjectionParametersGreatCircles()
      : ProjectionParametersBase(), 
        alphaMin_(-M_PI_2), betaMin_(-M_PI),
        betaRange_(2.0*M_PI), alphaRange_(M_PI)
    {      
      UpdateInternals_();
    };

    ProjectionParametersGreatCircles(const 
                                     ProjectionParametersGreatCircles & P){
      *this = P;
    };

    ProjectionParametersGreatCircles& 
    operator=(const ProjectionParametersGreatCircles & P);

    virtual ~ProjectionParametersGreatCircles(){};

    virtual inline void SetImageSize(const unsigned int w, 
                                     const unsigned int h) {
      ProjectionParametersBase::SetImageSize(w,h);
      UpdateInternals_();
    }
    
    /** An auto-adaption of the ranges is performed reltively
        to the minimal angles. Allowed value ranges are for 
        alphaRange : (0.0, 2*M_PI] and betaRange : (0.0, 2*M_PI].
    */
    virtual void SetAngleRanges(const double alphaRange, 
                                const double betaRange) ;

    
    virtual void SetMinAngles(const double alphaMin, 
                              const double betaMin) ;

    virtual const bool DoIntrinsicsDiffer(const ProjectionParametersBase* p) 
      const;


    bool DoesPointProjectIntoImageLocal(const Vector3<double>& localX, 
                                        HomgPoint2D& x,                       
                                        bool IgnoreDistortion = false) const;

    void CheckUnprojectionProjection();
    void CheckUnprojectionProjectionLocal();

    /** retrieves camera internals for configuration 
     *  that allows that image contains informations
     *  from the directions within the cone centered around the optical ray
     *  passing through worldPoint. The resulting viewing angles might be 
     *  larger then the opening angle halfAngleRange of the cone.
     */
    void GetCenteredAngleInterval(const HomgPoint3D& worldPoint,
                                   const double halfAngleRange,
                                   double& minAlpha, double& minBeta,
                                   double& alphaRange, double& betaRange);    

    /** Sets camera internals so that image contains informations
     *  from the directions within the cone centered around the optical ray
     *  passing through worldPoint. The resulting viewing angles might be 
     *  larger then the opening angle halfAngleRange of the cone.
     */
    int SetCenteredAngleInterval(const HomgPoint3D& worldPoint,
                                 const double& halfAngleRange);

    /** @brief calculates the projection of a point 
        in the camera coordinate system  
        to a pixel in the image plane of the camera
        with lens distortion
        In the simplest case perspective pinhole projection x = K * y  
        where y is the projection of X using  y = (R^T | -R^T C) X 
    */
    virtual HomgPoint2D ProjectLocal(const Vector3<double>& point,
                                     bool IgnoreDistortion = false) const;

    /** @brief map points from image onto unit diameter image plane in 3D.    
     * Chosen is the image plane with distance of one from the image center.
     */
    HomgPoint3D UnProjectToImagePlane(const HomgPoint2D& pos,
                                      const double& depth = 1.0,
                                      bool IgnoreDistortion = false) const;
   

    /** @brief calculates the viewing ray from the camera center (in the 
        camera coordinate system) which belongs to the image position pos
        with lens distortion
        e.g. (0,0,1) means optical axis
    */
    virtual Vector3<double> UnProjectLocal(const HomgPoint2D& pos,
                                           bool IgnoreDistortion = false)const;
      
 
    /** @brief covariant virtual copy constructor for use in Projection */
    virtual ProjectionParametersGreatCircles* Clone() const {
      return new ProjectionParametersGreatCircles(*this);
    } 

#ifdef BIAS_HAVE_XML2
    /** @brief specialization of XML block name function */
    virtual int XMLGetClassName(std::string& TopLevelTag,
                                double& Version) const;

    /** @brief specialization of XML write function */
    virtual int XMLOut(const xmlNodePtr Node, XMLIO& XMLObject) const;

    /** @brief specialization of XML read function */
    virtual int XMLIn(const xmlNodePtr Node, XMLIO& XMLObject); 
#endif

    friend std::ostream& operator<<(std::ostream &os, 
                                    ProjectionParametersGreatCircles& p);


    
  protected:
    /** Calculates the focallength and the beta principle point offset 
     *  rotation angle.
     */
    void UpdateInternals_();

    /** Takes angles from pp shifted frame and calcs cam coord vector
     */
    int ShiftedAnglesToCamCoordVector_(double alpha, 
                                       double beta, 
                                       Vector3<double>& vec) const;

    int CamCoordVectorToUnshiftedAngles_(Vector3<double> vec, 
                                         double& alpha, double& beta) const;

    int CamCoordVectorToShiftedAngles_(Vector3<double> vec, 
                                       double& alpha, double& beta) const;

    Vector2<double> ShiftedAnglesToImage(double alpha, double beta) const;
    int ImageToShiftedAngles(Vector2<double>& imgPoint,
                             double& alpha, double& beta) const;
      
    /** Calculates the coordinate transformation that embeds the 
     * internal shift frame into camera coordinates 
     */
    int DetermineOffsetPose_(double betaMin, 
                             Pose& localOffsetPose) const ;

    /** Pose that also holds the principle point offset for beta.
     *  Is not concatenated to the camera pose.
     */
    Pose offsetPoseBeta_;

    /** Calculates the angle values for world coord point without
     *  internal shift frame just in camera coordinates.
     */
    int CalculateUnbiasedAnglesLocal_(const Vector3<double>& point,
                                      double& alpha, double& beta) const;

    /** scale from angles to image*/
    double focalLengthAlpha_;
    /** scale from angles to image*/
    double focalLengthBeta_;

    //smallest angle within beta plane (x=0)
    double alphaMin_;
    //angle represented by y=0
    double betaMin_;
    //angle where the principle point lies in offset system
    double betaPP_;
    
    //range of angles within image, y coords
    double betaRange_;
    //range of angles within image, x-coords
    double alphaRange_;

  };

  inline std::ostream& operator<<(std::ostream &os, 
                                  ProjectionParametersGreatCircles& p) 
  {
    os <<"R:"<<p.GetR()<<"C:"<<p.GetC()<<std::endl;
    os<<"ImageSize: "<<p.width_<<" x "<<p.height_<<std::endl;
    os<<"AngleOffset: beta = "<<p.betaMin_<<", alpha = "
      <<p.alphaMin_<<std::endl;
    //operator<<(os, (ProjectionParametersBase&)p);
    os<<"angleRange: alpha = "<<p.alphaRange_;
    os <<", beta =  "<<p.betaRange_<<std::endl;
    os<<"pixel2AngleScale: alpha = "<<p.focalLengthAlpha_;
    os <<", beta = "<<p.focalLengthBeta_<<std::endl;

    os <<"offsetPoseBeta = "<<p.offsetPoseBeta_<<std::endl;    
    
    return os;
  }


 
} // end namespace 


#include <Base/Common/BIASpragmaEnd.hh>

#endif 

