/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "Geometry/ProjectionParametersPerspective.hh" 


using namespace BIAS;
using namespace std;

bool ProjectionParametersPerspective::
GetCutOutParameters(const Vector2<int>& centerPoint,
                    const unsigned int halfWidth,
                    const unsigned int halfHeight, 
                    ProjectionParametersPerspective& cutOutParams) const
{
  cutOutParams = (*this);
  Vector2<int> UL(centerPoint[0]-halfWidth, centerPoint[1]-halfHeight);
  Vector2<int> LR(centerPoint[0]+halfWidth, centerPoint[1]+halfHeight);
  cutOutParams.width_ = 2 * halfWidth + 1;
  cutOutParams.height_ = 2 * halfHeight + 1;
  
  cutOutParams.SetPrincipal(principalX_ - UL[0], principalY_ - UL[1]);

  return ((UL[0]>=0) && (UL[0]<(int)width_)  && 
          (UL[1]>=0) && (UL[1]<(int)height_) && 
          (LR[0]>=0) && (LR[0]<(int)width_)  && 
          (LR[1]>=0) && (LR[1]<(int)height_));
}

const bool ProjectionParametersPerspective::
DoIntrinsicsDiffer(const ProjectionParametersBase* p) const
{
  if(ProjectionParametersBase::DoIntrinsicsDiffer(p))
    return true;
  
  const ProjectionParametersPerspective* ppp = 
    dynamic_cast<const ProjectionParametersPerspective*>(p);
  if(ppp == NULL)
    return true;

  if(focallength_ != ppp->focallength_)
    return true;

  if(kc1_ != ppp->kc1_ || kc2_ != ppp->kc2_ || 
     kc3_ != ppp->kc3_ || kc4_ != ppp->kc4_)
    return true;

  return false;
}


bool ProjectionParametersPerspective::
DoesPointProjectIntoImageLocal(const Vector3<double>& localX, 
                               HomgPoint2D& x,                       
                               bool IgnoreDistortion) const
{
  if(localX[2] > 0.0) {
    x = ProjectLocal(localX, IgnoreDistortion);
    x.Homogenize(); //neccessary ?
    return (x[0]>=-0.5 && x[1]>=-0.5 && 
            x[0]<=(double)(width_)-0.5 && x[1]<=(double)(height_)-0.5 );
  } 
  return false;
}


HomgPoint2D ProjectionParametersPerspective::
ProjectLocal(const Vector3<double>& point, bool IgnoreDistortion) const 
{
  // make sure we dont project points behind the camera !
  if (point[2]<1e-20) return HomgPoint2D(0,0,0);
  // in-plane affine coordinate transformation with KMatrix
  HomgPoint2D pos2(point), pos;
  if (!IgnoreDistortion) DistortNormalized(pos2);
  K_.Mult(pos2, pos);
 
  //HomgPoint2D pos;
  //K_.Mult(point, pos);
  // radial distortion here 
  //if (!IgnoreDistortion) Distort(pos);
  return pos.Homogenize();
}

HomgPoint3D ProjectionParametersPerspective::
UnProjectToImagePlane(const HomgPoint2D& pos, const double& depth,
                      bool IgnoreDistortion) const 
{
  Vector3<double> p = UnProjectLocal(pos, IgnoreDistortion);
  if(Equal(p[2], 0.0))
    return HomgPoint3D(0,0,0,0);
  p*= depth / p[2];
  if(Equal(p.NormL2(), 0.0))
    return HomgPoint3D(0,0,0,0);
  HomgPoint3D local(p);
  return Pose_.LocalToGlobal(local);
}


Vector3<double>  ProjectionParametersPerspective::
UnProjectLocal(const HomgPoint2D& cPos, bool IgnoreDistortion) const {
  HomgPoint2D mpos(cPos); // de-const pos
  // radial undistortion here 
  if (!IgnoreDistortion) Undistort(mpos);
  // in-plane affine coordinate transformation with KMatrix
  Vector3<double> point;
  invK_.Mult(mpos, point);
  const double n = point.NormL2();
  if (n>0.0) point /= n;
  return point;
}


void ProjectionParametersPerspective::
DistortNormalized(HomgPoint2D& point2d) const 
{
  // do nothing in case of ideal lense
  if (kc1_==0 && kc2_==0 && kc3_==0 && kc4_==0) return;
  point2d.Homogenize();

  double x = point2d[0];
  double y = point2d[1];
  double xx = x*x;
  double yy = y*y;
  double rsquared = xx + yy;
    
  // radial distortion:
  double raddist = 1 + kc1_*rsquared + kc2_*rsquared*rsquared;
  // tangential distortion:
  if (kc3_!=0.0 || kc4_!=0.0){
    double tandistx=0.0, tandisty=0.0;
    double twoxy = 2*x*y;
    tandistx = kc3_*twoxy + kc4_ * (rsquared + 2*xx);
    tandisty = kc4_*twoxy + kc3_ * (rsquared + 2*yy);
    // apply distortion:
    point2d[0] = raddist * x + tandistx;
    point2d[1] = raddist * y + tandisty;
  } else {
    // apply distortion:
    point2d[0] = raddist * x;
    point2d[1] = raddist * y;
  }
}

void ProjectionParametersPerspective::
Distort(HomgPoint2D& point2d) const 
{
  // do nothing in case of ideal lense
  if (kc1_==0 && kc2_==0 && kc3_==0 && kc4_==0) return;

  // copy of distorted point
  double distx,disty;
  double x,y;
  double twoxy;
  
  double rsquared;
  double raddist;
  double tandistx, tandisty;
  
  point2d.Homogenize();
  

  // this is effectively: (x,y,1) =  K^-1 * point2D

  // Points must be Bougouet-normalized 
  // (subtract principal point, then divide by focal length)
  x = (point2d[0]-principalX_)/focallength_;
  y = (point2d[1]-principalY_)/(GetAspectratio()*focallength_);
      
  twoxy = 2 * x*y;
  rsquared = x*x + y*y;
    
  // radial distortion:
  raddist = 1 + kc1_*rsquared + kc2_*rsquared*rsquared;
  // tangential distortion:
  tandistx = kc3_*twoxy + kc4_ * (rsquared + 2*x*x);
  tandisty = kc4_*twoxy + kc3_ * (rsquared + 2*y*y);

  // apply distortion:
  distx = raddist * x + tandistx;
  disty = raddist * y + tandisty;
   
  // invert Bougouet-normalization
  point2d[0] = distx*focallength_ + principalX_;
  point2d[1] = disty*GetAspectratio()*focallength_ + principalY_;
}


void ProjectionParametersPerspective::
Undistort(HomgPoint2D& point2d) const {

  // do nothing in case of ideal lense
  if (kc1_==0 && kc2_==0 && kc3_==0 && kc4_==0) return;

// copy of distorted point
  double distx,disty;
  double x,y;
  double twoxy;
  
  double rsquared;
  double raddist;
  double tandistx, tandisty;
  
  point2d.Homogenize();

  // Points must be Bougouet-normalized 
  // (subtract principal point, then divide by focal length)
  x = point2d[0] = (point2d[0]-principalX_)/focallength_;
  y = point2d[1] = (point2d[1]-principalY_)/(GetAspectratio()*focallength_);
  
  // 5 iterations are enough
  for (unsigned int i=0;i<5;i++){
        
    twoxy = 2 * x*y;
    rsquared = x*x + y*y;
    
    // radial distortion:
    raddist = 1.0 + (kc1_ + kc2_*rsquared)*rsquared;
    // tangential distortion:
    tandistx = kc3_*twoxy + kc4_ * (rsquared + 2*x*x);
    tandisty = kc4_*twoxy + kc3_ * (rsquared + 2*y*y);
    
    // apply distortion:
    distx = raddist * x + tandistx;
    disty = raddist * y + tandisty;
    
    // subtract distortion from distorted coordinates
    x = point2d[0] - (distx-x);
    y = point2d[1] - (disty-y);

    // prevent NaN in next iteration step
    if (fabs(x)>1e10 || fabs(y)>1e10) {
      point2d[0] = point2d[1] = point2d[2] = 0;
      return;
    }
    
  }

  // invert Bougouet-normalization
  point2d[0] = x*focallength_ + principalX_;
  point2d[1] = y*GetAspectratio()*focallength_ + principalY_;
}

#ifdef BIAS_HAVE_XML2
 
int ProjectionParametersPerspective::
XMLGetClassName(std::string& TopLevelTag, double& Version) const {
  TopLevelTag = "ProjectionParametersPerspective";
  Version = 0.1;
  return 0;
}

int ProjectionParametersPerspective::
XMLOut(const xmlNodePtr Node, XMLIO& XMLObject) const {
  xmlNodePtr theNode;
  string TopLevelName;
  double Version;
  ProjectionParametersBase::XMLGetClassName(TopLevelName, Version);
  theNode = XMLObject.addChildNode(Node, TopLevelName);
  XMLObject.addAttribute(theNode, "Version", Version);
  ProjectionParametersBase::XMLOut(theNode, XMLObject);

  xmlNodePtr childNode;
  XMLObject.addAttribute(Node,"Focallength", focallength_);
  XMLObject.addAttribute(Node,"Skew", skew_);
  childNode = XMLObject.addChildNode(Node,"Distortion");
  XMLObject.addAttribute(childNode, "k1", kc1_);
  XMLObject.addAttribute(childNode, "k2", kc2_);
  XMLObject.addAttribute(childNode, "p1", kc3_);
  XMLObject.addAttribute(childNode, "p2", kc4_);
  return 0; 
 }

 
int ProjectionParametersPerspective::
XMLIn(const xmlNodePtr Node, XMLIO& XMLObject) {

  string TopLevelName;
  double Version;
  ProjectionParametersBase::XMLGetClassName(TopLevelName, Version);

  xmlNodePtr childNode; 
  if ((childNode = XMLObject.getChild(Node, TopLevelName))==NULL) {
    BIASERR("Error in xml, no tag" << TopLevelName);
    return -1;
  }
  if (ProjectionParametersBase::XMLIn(childNode, XMLObject)!=0) return -1;

  focallength_  = XMLObject.getAttributeValueDouble(Node, "Focallength");
  if (XMLObject.getAttributeByName(Node, "Skew")!=NULL)
    skew_  = XMLObject.getAttributeValueDouble(Node, "Skew");
  if ((childNode = XMLObject.getChild(Node, "Distortion"))!=NULL) {
    kc1_ = XMLObject.getAttributeValueDouble(childNode, "k1");
    kc2_ = XMLObject.getAttributeValueDouble(childNode, "k2");
    kc3_ = XMLObject.getAttributeValueDouble(childNode, "p1");
    kc4_ = XMLObject.getAttributeValueDouble(childNode, "p2");
  }
  ComputeK_();
  return 0;
}

#endif // BIAS_HAVE_XML2

void ProjectionParametersPerspective::
GetIdealK(KMatrix& K) const
{
  unsigned int imageWidth, imageHeight;
  GetImageSize(imageWidth, imageHeight);

  HomgPoint2D p(0, 0);
  Undistort(p);
  p = invK_ * p;
  p.Homogenize();
    
  double minX = p[0], maxX = p[0], minY = p[1], maxY = p[1];

  for (unsigned int x = 0; x < imageWidth; x++) {
    p = HomgPoint2D(x, 0);
    Undistort(p);
    p = invK_ * p;
    p.Homogenize();
    if (p[0] < minX) minX = p[0];
    if (p[0] > maxX) maxX = p[0];
    if (p[1] < minY) minY = p[1];
    if (p[1] > maxY) maxY = p[1];

    p = HomgPoint2D(x, imageHeight-1);
    Undistort(p);
    p = invK_ * p;
    p.Homogenize();
    if (p[0] < minX) minX = p[0];
    if (p[0] > maxX) maxX = p[0];
    if (p[1] < minY) minY = p[1];
    if (p[1] > maxY) maxY = p[1];
  }

  for (unsigned int y = 0; y < imageHeight; y++) {
    p = HomgPoint2D(0, y);
    Undistort(p);
    p = invK_ * p;
    p.Homogenize();
    if (p[0] < minX) minX = p[0];
    if (p[0] > maxX) maxX = p[0];
    if (p[1] < minY) minY = p[1];
    if (p[1] > maxY) maxY = p[1];

    p = HomgPoint2D(imageWidth-1, y);
    Undistort(p);
    p = invK_ * p;
    p.Homogenize();
    if (p[0] < minX) minX = p[0];
    if (p[0] > maxX) maxX = p[0];
    if (p[1] < minY) minY = p[1];
    if (p[1] > maxY) maxY = p[1];
  }

  double focalLengthX = idealImageWidth_ / (maxX - minX);
  double focalLengthY = idealImageHeight_ / (maxY - minY);
  double principalX = -focalLengthX * minX;
  double principalY = -focalLengthY * minY;

  K.SetIdentity();
  K[0][0] = focalLengthX;
  K[1][1] = focalLengthY;
  K[0][1] = skew_;
  K[0][2] = principalX;
  K[1][2] = principalY;
}

void ProjectionParametersPerspective::
SetIdealImageSize(unsigned int width, unsigned int height)
{
  idealImageWidth_ = width;
  idealImageHeight_ = height;
}

void ProjectionParametersPerspective::
GetIdealImageSize(unsigned int& width, unsigned int& height) const
{
  width = idealImageWidth_;
  height = idealImageHeight_;
}

void ProjectionParametersPerspective::
ComputeK_() {
  K_.SetIdentity();
  K_[0][0] = focallength_;
  K_[1][1] = aspectratio_ * focallength_;
  K_[0][1] = skew_;
  K_[0][2] = principalX_;
  K_[1][2] = principalY_;
  invK_ = K_.Invert();
}

double ProjectionParametersPerspective::
ViewDifference(const ProjectionParametersBase* pPPB) const {
  const ProjectionParametersPerspective *pPPP = 
     dynamic_cast<const ProjectionParametersPerspective* >(pPPB);
  // different projection model has a very different view
  if (pPPP==NULL) {
    BIASERR("View difference for different projection models requested !");
    return DBL_MAX;
  }

  Vector3<double> C0 = pPPP->GetC(), A0, A;
  
  RMatrix R = GetR(), R0 = pPPP->GetR();
  
  A[0] = R[0][2]; A[1] = R[1][2]; A[2] = R[2][2];
  A.Normalize();
  
  A0[0] = R0[0][2]; A0[1] = R0[1][2]; A0[2] = R0[2][2];
  A0.Normalize();
  
  // compute field of view, assuming const K !
  KMatrix K0 =  GetK();
  double f = 0.5*(K0[0][0] + K0[1][1]);
  double halfw = double(width_ + height_) / 4.0;
  double coshalffieldofview = cos(atan(halfw/f));
  
  double newness = (GetC()-C0).NormL2();
  newness += 1.0/(1.0-coshalffieldofview)*(1.0-fabs(A.ScalarProduct(A0)));
  return newness;
}


