#include "ImageCanvas.hh"

#include <Base/Image/ImageConvert.hh>
#include <Gui/StringConv.hh>

#ifdef   WIN32
#include <wx/image.h>
#endif //WIN32

#include <sstream>
#include <iomanip>

using namespace BIAS;
using namespace std;

#define SCROLLBAR_WIDTH 15

//-------------------------------------------------------
//-----------------------ZOOM FRAME----------------------

ZoomFrame::~ZoomFrame()
{}

ZoomFrame::ZoomFrame(wxWindow* parent, const wxString& title)
: wxFrame(parent, -1, title)
{
  Setup();
  Show();
}

void ZoomFrame::Setup()
{
  int widths[] = { -1, 280 };
  CreateStatusBar(2);
  SetStatusWidths(2, widths);

  QuitButton_ = new wxButton(this, MIA_ZoomQuitButton, wxT("Close"));

  Window_ = new ZoomImageCanvas(this, GetStatusBar(), 1);

  SLScale_ = new wxSlider(this, -1, 10, 1, 25, wxDefaultPosition,
    wxSize(-1, -1), wxSL_HORIZONTAL);

  wxBoxSizer *box = new wxBoxSizer(wxVERTICAL);
  wxBoxSizer *hbox = new wxBoxSizer(wxHORIZONTAL);

  box->Add(QuitButton_, 0, wxALL | wxGROW, 3);
  box->Add(Window_, 1, wxALL | wxGROW, 3);
  hbox->Add(new wxStaticText( this, -1, _T("Zoom") ), 0, wxALL, 5);
  hbox->Add(SLScale_, 1, wxALL | wxGROW, 5);
  box->Add(hbox, 0, wxGROW, 0);

  box->SetSizeHints(this);
  SetSizer(box);
}

void ZoomFrame::NewImage(BIAS::ImageBase& im,
             BIAS::Image<unsigned char>& ucim,
             const wxString& name)
{
  Window_->NewImage(im, ucim);
  SetStatusText(name, 0);
}

void ZoomFrame::NewCoo(unsigned x, unsigned y)
{ Window_->NewCoo(x, y); }



BEGIN_EVENT_TABLE (ZoomFrame, wxFrame)
EVT_BUTTON      (MIA_ZoomQuitButton, ZoomFrame::OnQuit)
EVT_SCROLL      (ZoomFrame::OnZoomFactor)
EVT_CLOSE       (ZoomFrame::OnClose)
//EVT_KEY_DOWN    (ZoomFrame::OnKeyPress)
END_EVENT_TABLE()

//void ZoomFrame::OnQuit(wxCommandEvent & event)
void ZoomFrame::OnQuit(wxCommandEvent &)
{
  Close(true);
}

//void ZoomFrame::OnClose(wxCloseEvent & event)
void ZoomFrame::OnClose(wxCloseEvent &)
{
  /*
  ((ImageCanvas*)GetParent())->SetZoomed(false);
  Destroy();
  */
  Hide();
}

void ZoomFrame::SetZoom(unsigned z)
{
  BIASASSERT(z>=1);
  BIASASSERT(z<=25);
  SLScale_->SetValue(z);
  Window_->NewZoom(SLScale_->GetValue());
}

//void ZoomFrame::OnZoomFactor(wxScrollEvent & event)
void ZoomFrame::OnZoomFactor(wxScrollEvent&)
{ Window_->NewZoom(SLScale_->GetValue()); }

// void ZoomFrame::OnKeyPress(wxKeyEvent &event)
// {
//   cerr <<"ZoomFrame::OnKeyPress\n";
//   event.Skip();
// }


//-------------------------------------------------------
//-----------------------ZOOM CANVAS---------------------

ZoomImageCanvas::ZoomImageCanvas(wxWindow* parent, wxStatusBar *sbar,
                 int sbarid, wxWindowID id, const wxPoint& pos,
                 const wxSize& size, long style,
                 const wxString& name)
                 : wxPanel(parent, id, pos, size, style, name)
{
  Bitmap_=NULL;

  wxbitmap_mutex_ = new pthread_mutex_t;
  origim_mutex_ = new pthread_mutex_t;
  pthread_mutex_init(wxbitmap_mutex_, NULL);
  pthread_mutex_init(origim_mutex_, NULL);

  ZoomFactor_=10;
  ZoomX_=0;
  ZoomY_=0;
  SBar_=sbar;
  SBarID_=sbarid;
  mousex_=mousey_=-1;
}

ZoomImageCanvas::~ZoomImageCanvas()
{
  if (wxbitmap_mutex_) delete wxbitmap_mutex_;
  if (origim_mutex_) delete origim_mutex_;
}

void ZoomImageCanvas::GetZoomImage()
{
#ifdef WIN32
  GetSize(&wwidth_, &wheight_);
  ZoomXPos_=wwidth_/2-ZoomFactor_/2;
  ZoomYPos_=wheight_/2-ZoomFactor_/2;
#endif //WIN32

  if (zim_.GetWidth()!=(unsigned)wwidth_ ||
    zim_.GetHeight()!=(unsigned)wheight_)
    zim_.Release();
  if (zim_.IsEmpty())
    zim_.Init(wwidth_, wheight_, 3);

  zim_.FillImageWithConstValue((unsigned char)0);
  unsigned char **ida = zim_.GetImageDataArray();
  unsigned char **sida= ucim_.GetImageDataArray();
  int xsize=wwidth_/ZoomFactor_/2+1;
  int ysize=wheight_/ZoomFactor_/2+1;
  register unsigned int xstart, ystart;
  register unsigned char res[3];
  unsigned width=ucim_.GetWidth(), height=ucim_.GetHeight();
  int mx, my;

  //cout << "wwidth/wheight: "<<wwidth_<<"x"<<wheight_<<endl;
  //cout << "width/height: "<<width<<"x"<<height<<endl;
  //cout << "x "<<(int)ZoomX_-(int)xsize<<" - "<<(int)ZoomX_+(int)xsize<<endl;
  //cout << "y "<<(int)ZoomY_-(int)ysize<<" - "<<(int)ZoomY_+(int)ysize<<endl;
  for (int x=(int)ZoomX_-(int)xsize; x<=(int)ZoomX_+(int)xsize; x++){
    for (int y=(int)ZoomY_-(int)ysize; y<=(int)ZoomY_+(int)ysize; y++){
      if (x<0 || y<0 || x>=(int)width || y>=(int)height){
        //cout <<"out of src range: "<< x << "x" <<y<<endl;
        res[0]=res[1]=res[2]=0;
      } else {
        mx=3*x;
        res[0]=sida[y][mx]; res[1]=sida[y][mx+1]; res[2]=sida[y][mx+2];
      }
      xstart = (x-(int)ZoomX_)*(int)ZoomFactor_+(int)ZoomXPos_;
      ystart = (y-(int)ZoomY_)*(int)ZoomFactor_+(int)ZoomYPos_;
      for (register int i=0; i<(int)ZoomFactor_; i++){
        for (register int j=0; j<(int)ZoomFactor_; j++){
          my=ystart+i; mx=(xstart+j);
          if (mx>=0 && my>=0 && mx<wwidth_ && my<wheight_){
            mx*=3;
            ida[my][mx] = res[0];
            ida[my][mx+1] = res[1];
            ida[my][mx+2] = res[2];
          }
        }
      }
    }
  }
  BIAS_MUTEX_LOCK(wxbitmap_mutex_);
  wximage_ = new wxImage(zim_.GetWidth(), zim_.GetHeight(),
    zim_.GetImageData(), true);
  BIAS_MUTEX_UNLOCK(wxbitmap_mutex_);
  if (mousex_>=0 && mousey_>=0){
    UpdateStatusBar((unsigned) mousex_, (unsigned) mousey_);
  }
}

void ZoomImageCanvas::UpdateStatusBar(unsigned mouseX, unsigned mouseY)
{
  if(SBar_ == NULL)
    return;
  if(Bitmap_==NULL)
    return;

  // compute original coordinate for zoomed canvas
  int ix=(int)mouseX-(int)ZoomXPos_;
  int iy=(int)mouseY-(int)ZoomYPos_;
  if (ix<0)
    ix=ix/(int)ZoomFactor_+(int)ZoomX_-1;
  else
    ix=ix/(int)ZoomFactor_+(int)ZoomX_;
  
  if (iy<0)
    iy=iy/(int)ZoomFactor_+(int)ZoomY_-1;
  else
    iy=iy/(int)ZoomFactor_+(int)ZoomY_;
  
  //BIAS_MUTEX_LOCK(wxbitmap_mutex_);
  BIAS_MUTEX_LOCK(origim_mutex_);
  
  {
    if(ix>=0 && ix<(int)origim_.GetWidth()
       && iy>=0 && iy<(int)origim_.GetHeight())
      {
        //BIAS_MUTEX_UNLOCK(wxbitmap_mutex_);

        ostringstream os;
        //display coordinate:
        os << "("<<setw(3)<<ix<<", "<<setw(3)<<iy<<") : (";

        // determine pixel type relying on storage type
        // dynamic_check to BIAS::Image<T> does not work because base image is 
        // loaded
        {
          for (unsigned int channel=0; 
               channel<origim_.GetChannelCount(); 
               channel++){
            if (channel!=0)
              os <<","; // no comma before first value
            BIAS::ImageBase::PrintPixelValue(origim_, ix,iy, channel, os ); 
          } 
        }
        os <<")";
        SBar_->SetStatusText(AsciiToWx(os.str()), SBarID_);
      };
  }
  BIAS_MUTEX_UNLOCK(origim_mutex_);
  BIAS_MUTEX_UNLOCK(wxbitmap_mutex_);
}

void ZoomImageCanvas::NewImage(ImageBase& im, Image<unsigned char>& ucim)
{
  BIAS_MUTEX_LOCK(origim_mutex_);
  origim_ = im;
  BIAS_MUTEX_UNLOCK(origim_mutex_);
  ucim_ = ucim;
  GetZoomImage();
#ifdef WIN32
  this->Refresh();
#else
  wxPaintEvent pe;
  wxPostEvent(this->GetEventHandler(),pe);
#endif
}

void ZoomImageCanvas::NewCoo(unsigned x, unsigned y)
{
  if (ZoomX_!=x || ZoomY_!=y){
    ZoomX_=x;
    ZoomY_=y;
    ZoomXPos_=wwidth_/2-ZoomFactor_/2;
    ZoomYPos_=wheight_/2-ZoomFactor_/2;
    GetZoomImage();
#ifdef WIN32
    this->Refresh();
#else
    wxPaintEvent pe;
    wxPostEvent(this->GetEventHandler(),pe);
#endif
  }
}

void ZoomImageCanvas::NewZoom(int zoom)
{
  if ((unsigned)zoom!=ZoomFactor_){
    ZoomFactor_=zoom;
    GetZoomImage();
#ifdef WIN32
    this->Refresh();
#else
    wxPaintEvent pe;
    wxPostEvent(this->GetEventHandler(),pe);
#endif
  }
}

BEGIN_EVENT_TABLE (ZoomImageCanvas, wxWindow)
EVT_PAINT       (ZoomImageCanvas::OnPaint)
EVT_MOTION      (ZoomImageCanvas::OnMouseMove)
EVT_RIGHT_DOWN  (ZoomImageCanvas::OnRightMouse)
EVT_LEFT_DOWN   (ZoomImageCanvas::OnLeftMouse)
EVT_CHAR        (ZoomImageCanvas::OnKey)
EVT_SIZE        (ZoomImageCanvas::OnResize)
EVT_LEAVE_WINDOW(ZoomImageCanvas::OnMouseLeave)
END_EVENT_TABLE()

void ZoomImageCanvas::OnRightMouse(wxMouseEvent &event)
{
  wxClientDC dc(this);
  PrepareDC(dc);
  wxPoint pos = event.GetPosition();
  int x = dc.DeviceToLogicalX( pos.x );
  int y = dc.DeviceToLogicalY( pos.y );
  int ix=(x-(int)(wwidth_>>1))/(int)ZoomFactor_+(int)ZoomX_;
  int iy=(y-(int)(wheight_>>1))/(int)ZoomFactor_+(int)ZoomY_;
  //cerr << "new coo: "<<ix<<", "<<iy; cerr.flush();
  NewCoo(ix, iy);
  //cerr << " .... finished\n";
}

//void ZoomImageCanvas::OnMouseLeave(wxMouseEvent &event)
void ZoomImageCanvas::OnMouseLeave(wxMouseEvent&)
{
  if (SBar_){ SBar_->SetStatusText(wxT(""), SBarID_); }
  mousex_=mousey_=-1;
}

void ZoomImageCanvas::OnResize(wxSizeEvent &event)
{
  GetSize(&wwidth_, &wheight_);
  ZoomXPos_=wwidth_/2-ZoomFactor_/2;
  ZoomYPos_=wheight_/2-ZoomFactor_/2;
  GetZoomImage();
  event.Skip();
#ifdef WIN32
  this->Refresh();
#else
  wxPaintEvent pe;
  wxPostEvent(this->GetEventHandler(),pe);
#endif
}

//void ZoomImageCanvas::OnPaint(wxPaintEvent & event)
void ZoomImageCanvas::OnPaint(wxPaintEvent & )
{
  //  wxPanel::OnPaint(event);

  wxPaintDC dc(this);
  PrepareDC(dc);
  BIAS_MUTEX_LOCK(wxbitmap_mutex_);
  if (wximage_ && wximage_->Ok()){
    if (Bitmap_ != NULL) delete Bitmap_;
    Bitmap_ = new wxBitmap(*wximage_);
  }
  if (Bitmap_ && Bitmap_->Ok() && dc.Ok()) {
    if (ClearDrawingContext_) dc.Clear();
    dc.DrawBitmap(*Bitmap_, 0, 0);
  }
  BIAS_MUTEX_UNLOCK(wxbitmap_mutex_);
}

void ZoomImageCanvas::OnMouseMove(wxMouseEvent &event)
{
  wxClientDC dc(this);
  PrepareDC(dc);
  wxPoint pos = event.GetPosition();
  mousex_ = dc.DeviceToLogicalX( pos.x );
  mousey_ = dc.DeviceToLogicalY( pos.y );
  UpdateStatusBar((unsigned) mousex_, (unsigned) mousey_);
}


void ZoomImageCanvas::OnLeftMouse(wxMouseEvent &event)
{
  // pass event through to ImageCanvas which is parent of parent
  if (GetParent()){
    //  if (GetParent() && GetParent()->GetParent()){
    wxClientDC dc(this);
    PrepareDC(dc);
    wxPoint pos = event.GetPosition();
    unsigned int x = dc.DeviceToLogicalX( pos.x );
    unsigned int y = dc.DeviceToLogicalY( pos.y );
    int ix=(int)x-(int)ZoomXPos_, iy=(int)y-(int)ZoomYPos_;
    if (ix<0) ix=ix/(int)ZoomFactor_+(int)ZoomX_-1;
    else ix=ix/(int)ZoomFactor_+(int)ZoomX_;
    if (iy<0) iy=iy/(int)ZoomFactor_+(int)ZoomY_-1;
    else iy=iy/(int)ZoomFactor_+(int)ZoomY_;

    wxMouseEvent ke=event;
    ke.m_x=ix;
    ke.m_y=iy;
    wxPostEvent(GetParent()->GetParent()->GetEventHandler(), ke);
    //cerr << "passing mouse event at ("<<ix<<", "<<iy
    //     <<") from ZoomImageCanvas to ImageCanvas\n";
  } else {
    BIASERR("no ImageCanvas belonging to ZoomImageCanvas\n");
  }
  event.Skip();
}


void ZoomImageCanvas::OnKey(wxKeyEvent &event)
{
  // pass event through to ImageCanvas which is parent of parent
  if (GetParent()){
    //  if (GetParent() && GetParent()->GetParent()){
    wxClientDC dc(this);
    PrepareDC(dc);
    wxPoint pos = event.GetPosition();
    unsigned int x = dc.DeviceToLogicalX( pos.x );
    unsigned int y = dc.DeviceToLogicalY( pos.y );
    int ix=(int)x-(int)ZoomXPos_, iy=(int)y-(int)ZoomYPos_;
    if (ix<0) ix=ix/(int)ZoomFactor_+(int)ZoomX_-1;
    else ix=ix/(int)ZoomFactor_+(int)ZoomX_;
    if (iy<0) iy=iy/(int)ZoomFactor_+(int)ZoomY_-1;
    else iy=iy/(int)ZoomFactor_+(int)ZoomY_;

    wxKeyEvent ke=event;
    ke.m_x=ix;
    ke.m_y=iy;
    ke.m_keyCode=event.GetKeyCode();
    wxPostEvent(GetParent()->GetParent()->GetEventHandler(), ke);
    // cerr << "passing KEY event at ("<<ix<<", "<<iy
    //     <<") from ZoomImageCanvas to ImageCanvas\n";
  } else {
    BIASERR("no ImageCanvas belonging to ZoomImageCanvas\n");
  }
  event.Skip();
}

//-------------------------------------------------------
//-------------------HISTOGRAMM FRAME--------------------
/** 
    @brief It's a Frame used to have a Histogramm Canvas, which
    describes a histogramm of current image
@status not working yet
@author amattal
@date 03/2007
*/

//Constructor
HistoFrame::HistoFrame(wxWindow* parent, const wxString& title)
  : wxFrame(parent, -1, title, wxPoint(10,10), wxSize(300,250))
{
  Setup();
  Show();
}

void HistoFrame::Setup(){
  int widths[] = {-1, 200};
  CreateStatusBar(2);
  SetStatusWidths(2, widths);
  
  //QuitButton_ = new wxButton(this, MIA_HistoQuitButton, wxT("Close"));
  Window_ = new HistoImageCanvas(this, GetStatusBar(), 1);
  
  wxBoxSizer *box = new wxBoxSizer(wxVERTICAL);
  
  //box->Add(QuitButton_, 0, wxALL | wxGROW, 3);
  box->Add(Window_, 1, wxALL | wxGROW, 3);
  
  box->SetSizeHints(this);
  SetSizer(box);
}

//Destructor
HistoFrame::~HistoFrame()
{
}

BEGIN_EVENT_TABLE (HistoFrame, wxWindow)
  EVT_CLOSE       (HistoFrame::OnClose)
  //EVT_BUTTON      (MIA_HistoQuitButton, HistoFrame::OnQuit)
  //EVT_MOTION      (HistoImageCanvas::OnMouseMove)
  //EVT_CHAR        (HistoFrame::OnKey)
  //EVT_SIZE        (HistoImageCanvas::OnResize)
  //EVT_LEAVE_WINDOW(HistoImageCanvas::OnMouseLeave)
END_EVENT_TABLE()

void HistoFrame::NewImage(BIAS::ImageBase& im,
                          const wxString& name){
  if(!(Window_->SameName((wxString)name, true))){
    Window_->NewImage(im);
    SetStatusText(name, 0);
  }
}

void HistoFrame::OnQuit(wxCommandEvent & event){
  Close(true);
}
void HistoFrame::OnClose(wxCloseEvent & event){
  Hide();
}

//-------------------------------------------------------
//------------------HISTOGRAMM CANVAS--------------------
/** 
    @brief It's a Canvas used to compute Histogrammfor current image
@status working
@author amattal
@date 03/2007
*/

HistoImageCanvas::HistoImageCanvas(wxWindow* parent, 
                                   wxStatusBar *sbar, int sbarid,
                                   wxWindowID id, 
                                   const wxPoint& pos,
                                   const wxSize& size,
                                   long style,
                                   const wxString& name)
  : wxPanel(parent, id, pos, size, style, name){
  histo_float_steps = 256;
  histo_type = -1;
  wxPaintEvent ev;
  OnPaint(ev);
}

HistoImageCanvas::~HistoImageCanvas(){
}

BEGIN_EVENT_TABLE (HistoImageCanvas, wxWindow)
  EVT_PAINT         (HistoImageCanvas::OnPaint)
  //EVT_MOTION      (HistoImageCanvas::OnMouseMove)
  //EVT_CHAR        (HistoFrame::OnKey)
  //EVT_SIZE        (HistoImageCanvas::OnResize)
  //EVT_LEAVE_WINDOW(HistoImageCanvas::OnMouseLeave)
END_EVENT_TABLE()

//CALCULATE HISTOGRAMM OF CURRENT IMAGE
void HistoImageCanvas::UpdateHistogramm(ImageBase image){

  histo_size = image.GetWidth()*image.GetHeight();

  //little pixelsum-check
  if(histo_size <= 0)
        BIASERR("No image for HISTOGRAMMS presented!");

  histo_type = -1;
  switch (image.GetStorageType()){
  case ImageBase::ST_unsignedchar:
    {          
      // !!! histogramm is computed ONLY FOR CHANNEL 0 at the moment !!!
      int channelcount = image.GetChannelCount();

      //debug-message
      cout << "UNSIGNED CHAR" << endl;
      histo_type = 0;

      //preparations
      histo_unsigned_char = new int[256];
      for (int i = 0; i < 256; i++){
        histo_unsigned_char[i] = 0;
      }
      
      //compute histogramm
      unsigned char* data = (unsigned char*)image.GetImageData();
      for (int i = 0; i < histo_size; i++){
        histo_unsigned_char[(int)data[i*channelcount]]++;
      }

      histo_percentage = new float[256];
      histo_percentage_biggest = 0.0;

      //filling percentages and giving debug info out
      float sum = 0.0;
      for (int i = 0; i < 256; i++){
        float perc = (float)histo_unsigned_char[i]/(float)histo_size;
#ifdef BIAS_DEBUG
        cout << i << ": " << histo_unsigned_char[i] << 
          " = " << 100*perc << "%" << endl;
#endif
        histo_percentage[i] = perc;
        sum += perc;
        if(perc > histo_percentage_biggest)
          histo_percentage_biggest = perc;
      }
#ifdef BIAS_DEBUG
      cout << sum*100 << "% total" << endl;
#endif

      break;
    } 
    
  case ImageBase::ST_float:
    {
      //debug-message
      cout << "FLOAT" << endl;

      //preparations
      histo_float = new Histo_Entry_Float[histo_float_steps];
      for (int i = 0; i < histo_float_steps; i++){
        histo_float[i].value_count = 0;
      }

      float* data = (float*)image.GetImageData();

      //get min and max-values in the image and compute scale-step
      float min =  100000.0;
      float max = -100000.0;
      
      for (int i = 0; i < histo_size; i++){
        if((float)data[i] > max){
          max = (float)data[i];
          int x = (int)(i/image.GetWidth());
          int y = i - x*image.GetWidth();
          cout << "max (" << y << ", " << x << ") value: " << max << endl;
        }
        if((float)data[i] < min)
          min = (float)data[i];
      }

      if(min >= 100000.0 || max <= -100000.0)
        BIASERR("Image has only pixels above 10^5 values or under -10^5");

      float step = (max-min)/histo_float_steps;

      histo_type = 1;
            
      //compute histogramm
      
      for (int i = 0; i < histo_size; i++){

        int index = 0;
        if(step != 0.0){
          index = (int)(((float)data[i]-min)/step);
        }

        //---COMPUTING HISTOGRAMM---
        //correcting slight errors while doing it with float-images
        if(index == histo_float_steps)
          index--;
        else if(index == -1)
          index = 0;
        //if the float-error is too big while counting histogramm
        else if(index < -1 || index > histo_float_steps){
          cout << "BAD pixel for histogramm: " << endl;
          cout << "data[i]: " << data[i] 
               << " min-data: " << min
               << " max-data: " << max
               << " index: " << index << endl;
          BIASERR("HISTOGRAMM IS WRONG, maybe TOO MANY STEPS/TOO WIDE RANGE");
        }
        //--------------------------

        histo_float[index].value_count++;
      }

      histo_percentage = new float[histo_float_steps];
      histo_percentage_biggest = 0.0;

      //filling percentages and giving debug info out
      float sum = 0.0;
      float begin = min;
      float end = min + step;
      for (int i = 0; i < histo_float_steps; i++){
        histo_float[i].value_begin = begin;
        histo_float[i].value_end   = end;
        
        if(begin <= 0.0 && end >= 0.0){
          histo_float_zero = i;
        }

        begin += step; end += step;
        
        float perc = (float)histo_float[i].value_count/(float)histo_size;
#ifdef BIAS_DEBUG
        cout << "values from " << histo_float[i].value_begin << " till " <<
          histo_float[i].value_end << ": " << histo_float[i].value_count <<
          " = " << 100*perc << "%" << endl;
#endif
        histo_percentage[i] = perc;
        sum += perc;

        if(perc > histo_percentage_biggest)
          histo_percentage_biggest = perc;
      }
      cout << sum*100 << "% total" << endl;
      cout << histo_float_steps << " steps in float-histogramm used" << endl;

      break;
    }
    
  default:
    break;
  }

  cout << "HISTOGRAMM UPDATED" << endl;

  wxPaintEvent ev;
  OnPaint(ev);
}

void HistoImageCanvas::OnPaint(wxPaintEvent& event){
  //prepare painter and get size of the window
  wxPaintDC dc(this);
  wxSize sz = GetClientSize();
  int wid = sz.x;
  int hei = sz.y;
  //cout << wid << "; " << hei << endl;
  
  //CLEAR
  dc.SetPen(*wxGREY_PEN);
  dc.SetBrush(*wxGREY_BRUSH);
 
  wxRect rectToDraw(0, 0, wid, hei);
  //wxRect rectToDraw(4, 4, wid-8, hei-8);
  dc.DrawRectangle(rectToDraw);

  dc.SetPen(*wxWHITE_PEN);
  dc.SetBrush(*wxWHITE_BRUSH);

  wxRect rectToDraw2(6, 6, wid-12, hei-12);
  //wxRect rectToDraw2(7, 7, wid-14, hei-14);
  dc.DrawRectangle(rectToDraw2);

  dc.SetPen(*wxLIGHT_GREY_PEN);
  dc.SetBrush(*wxLIGHT_GREY_BRUSH);
 
  wxRect rectToDraw3(8, 8, wid-16, hei-16);
  dc.DrawRectangle(rectToDraw3);
  //-----
  
  dc.SetPen(*wxBLACK_PEN);
  dc.SetBrush(*wxBLACK_BRUSH);
  
  int end = 0;
  if(histo_type == 0){
    end = 256;
    cout << "Painting Histogramm for unsigned-char image" << endl;
  }
  if(histo_type == 1){
    end = histo_float_steps;
    cout << "Painting Histogramm for float image" << endl;
  }
  //CRITICAL: if the picture is too wide (looks bad) or
  //          not as wide as 256 pixels -> no histo
  for(int i = 0; i < end; i++){
    // Our rectangle dimensions

    wxCoord w = 1, h = (int)(((float)(hei)-16.0)*
                             histo_percentage[i]/histo_percentage_biggest);

    //red mark for float-images, showing where value 0 is
    //DOESN't WORK PROPERLY YET - same image gives different 0-lines sometimes
    /**
    if(histo_type == 1 &&
       i == histo_float_zero){
      dc.SetPen(*wxRED_PEN);
      dc.SetBrush(*wxRED_BRUSH);

      wxRect rectToDrawRed1(i+8, 4, w, hei-8);
      dc.DrawRectangle(rectToDrawRed1);

      wxRect rectToDrawRed2(i+7,     3, w+2, 1);
      wxRect rectToDrawRed3(i+7, hei-3, w+2, 1);
      dc.DrawRectangle(rectToDrawRed2);
      dc.DrawRectangle(rectToDrawRed3);

      dc.SetPen(*wxBLACK_PEN);
      dc.SetBrush(*wxBLACK_BRUSH);
    }
    **/

    //cout << "h: " << h << endl;
    
    wxRect rectToDraw(i+8, hei-h-8, w, h);
    dc.DrawRectangle(rectToDraw);
  }

#ifdef BIAS_DEBUG
  cout << "biggest %: " << 100*histo_percentage_biggest << endl;
#endif
}

void HistoImageCanvas::NewImage(BIAS::ImageBase& im){
  UpdateHistogramm(im);
}
bool HistoImageCanvas::SameName(wxString name, bool settothenewone){
  if(histo_name != name){
    if(settothenewone)
      histo_name = name;
    return false;
  }
  return true;
}
void HistoImageCanvas::UpdateStatusBar(unsigned x, unsigned y){
}
void HistoImageCanvas::OnResize(wxSizeEvent & event){
}
void HistoImageCanvas::OnMouseMove(wxMouseEvent &event){
}
void HistoImageCanvas::OnMouseLeave(wxMouseEvent &event){
}

////////////////////////////////////////////////////////////////////////


ImageCanvas::ImageCanvas(wxWindow* parent, wxStatusBar* sbar, int sbarid,
             wxWindowID id, const wxPoint& pos, const wxSize& size,
             long style, const wxString& name)
             : wxScrolledWindow(parent, id, pos, size, style, name), origim_()
{
  Bitmap_=(wxBitmap *)NULL;
  wximage_=(wxImage *)NULL;
  SBar_=sbar;
  SBarID_=sbarid;

  wxbitmap_mutex_ = new pthread_mutex_t;
  origim_mutex_ = new pthread_mutex_t;
  pthread_mutex_init(wxbitmap_mutex_, NULL);
  pthread_mutex_init(origim_mutex_, NULL);

  Title_=name;
  mousex_=mousey_=-1;
  DefaultZoom_=5;
  ClearDrawingContext_=true;
  ShowScaled_ = false;

  //HINT: If it sparks while having a sequence of mixed images
  //      (unsigned char/float) then do like it with MyHistoWin_
  //      done is (set to NULL in the beginning and by every
  //      use of zoomcanvas check if MyZoomWin_ initialized is)
  MyZoomWin_=new ZoomFrame(this, Title_);
  MyZoomWin_->Show(false);

  MyHistoWin_=NULL;
}

ImageCanvas::~ImageCanvas()
{
  if (wximage_ != NULL) delete wximage_;
  if (Bitmap_ != NULL) delete Bitmap_;

  if (wxbitmap_mutex_) delete wxbitmap_mutex_;
  if (origim_mutex_) delete origim_mutex_;
}

void ImageCanvas::InitHisto(){
  MyHistoWin_=new HistoFrame(this, _T("HistoWindow"));
  MyHistoWin_->Show(true);
}

void ImageCanvas::Show(PyramidImage<unsigned char>& pim, string name)
{
  Image<unsigned char> im;
  pim.GetSingleImage(im);
  Show(im, name);
}


void ImageCanvas::Show(BIAS::Image<unsigned char>& im, string name)
{
  if (im.GetColorModel()!=ImageBase::CM_RGB){
    Image<unsigned char> tmpim;
    if (ImageConvert::Convert(im, tmpim, ImageBase::CM_RGB)!=0){
      BIASERR("error converting");
    }
    BIASDOUT(D_IC_IM,"Show(im,name) called Show(im,tmpim,name)");
    Show(im, tmpim, name);
  } else {
    BIASDOUT(D_IC_IM,"Show(im,name) called Show(im,im,name)");
    Show(im, im, name);
  }
}


void ImageCanvas::Show(BIAS::ImageBase& origim,
             BIAS::Image<unsigned char>& scaledrgbim,
             std::string name)
{
  BIAS_MUTEX_LOCK(origim_mutex_);
  BIASDOUT(D_IC_IM,"Adress of origim: "<<&origim);
  origim_=origim;
  BIAS_MUTEX_UNLOCK(origim_mutex_);
  rgbim_=scaledrgbim;
  BIASDOUT(D_IC_IM,"Adress of rgbim_: "<<&rgbim_);
  width_=origim.GetWidth();
  height_=origim.GetHeight();
  ImageName_=name;
  if(ShowScaled_) {
    // adjust virtual size and scrollbars to requested scaled image size
    SetVirtualSize(ScaledImageSizeX_, ScaledImageSizeY_);
    SetScrollbars(1, 1, ScaledImageSizeX_, ScaledImageSizeY_);
  } else {
    SetVirtualSize(width_, height_);
    SetScrollbars(1, 1, width_, height_);
    // workaround: the scrollbars do not dissapear when the size is equal
    // to the virtual size
    //SetSizeHints(-1, -1, width_+SCROLLBAR_WIDTH, height_+SCROLLBAR_WIDTH);
    //SetSizeHints(-1, -1, width_, height_);
  }
  BIAS_MUTEX_LOCK(wxbitmap_mutex_);
  if (wximage_ != NULL) delete wximage_;
  wximage_ =new wxImage(rgbim_.GetWidth(), rgbim_.GetHeight(),
    rgbim_.GetImageData(), true);
  BIAS_MUTEX_UNLOCK(wxbitmap_mutex_);
  if (MyZoomWin_->IsShown()) {
    //BIAS_MUTEX_LOCK(origim_mutex_);
    MyZoomWin_->NewImage(origim_, rgbim_, AsciiToWx(ImageName_));
    //BIAS_MUTEX_UNLOCK(origim_mutex_);
  }
  if(MyHistoWin_!=NULL){
	if (MyHistoWin_->IsShown()) {
		//BIAS_MUTEX_LOCK(origim_mutex_);
		MyHistoWin_->NewImage(origim_, AsciiToWx(ImageName_));
		//BIAS_MUTEX_UNLOCK(origim_mutex_);
	}
  }
  if (mousex_>=0 && mousey_>=0){
    UpdateStatusBar((unsigned)mousex_, (unsigned)mousey_);
  }
#ifdef WIN32
  this->Refresh();
#else
  wxPaintEvent pe;
  wxPostEvent(this->GetEventHandler(),pe);
#endif
}

void ImageCanvas::ShowHistogramm(ImageBase& image){
  if(MyHistoWin_==NULL)
	  InitHisto();
  
  if (!image.IsEmpty()){
    MyHistoWin_->NewImage(image, AsciiToWx(ImageName_));
    MyHistoWin_->Show(true);
  }
  else{
    BIASERR("Hey, no image saved in image!");
  }
}

void ImageCanvas::ShowHistogramm(){
  if(MyHistoWin_==NULL)
    InitHisto();
  
  if (!origim_.IsEmpty()){
    MyHistoWin_->NewImage(origim_, AsciiToWx(ImageName_));
    MyHistoWin_->Show(true);
  }
  else{
    BIASERR("Hey, no image saved in origim_!");
  }
}

void ImageCanvas::Zoom(unsigned x, unsigned y)
{
  //BIAS_MUTEX_LOCK(origim_mutex_);
  MyZoomWin_->NewImage(origim_, rgbim_, AsciiToWx(ImageName_));
  //BIAS_MUTEX_UNLOCK(origim_mutex_);
  MyZoomWin_->SetZoom(DefaultZoom_);
  MyZoomWin_->Show(true);

  if (ShowScaled_) {
    MyZoomWin_->NewCoo((int)((float)x * ((float)width_ / (float)ScaledImageSizeX_)),
      (int)((float)y *((float)height_/(float)ScaledImageSizeY_)));
  }
  else MyZoomWin_->NewCoo(x, y);
}

void ImageCanvas::UpdateStatusBar(unsigned x, unsigned y)
{

  // check if there is nothing to do:
  if (SBar_==NULL)
    return;
  if(Bitmap_==NULL)
    return;
  
  // TODO: do we really need these lockings?
  //BIAS_MUTEX_LOCK(wxbitmap_mutex_);
  
  BIASASSERT(SBar_   != NULL);
  BIASASSERT(Bitmap_ != NULL);
  
  if ((x>0) && (x<width_) && (y>0) & (y<height_)){
    
    ostringstream os;
    //displate coordinate:
    os << "("<<setw(3)<<x<<", "<<setw(3)<<y<<") : (";
    
    // determine pixel type relying on storage type
    // dynamic_check to BIAS::Image<T> does not work because base image is 
    // loaded
    
    BIAS_MUTEX_LOCK(origim_mutex_);
    {
      for (unsigned int channel=0; 
           channel<origim_.GetChannelCount(); 
           channel++){
        if (channel!=0)
          os <<","; // no comma before first value
        BIAS::ImageBase::PrintPixelValue(origim_, x,y, channel, os );
      }
    }
    BIAS_MUTEX_UNLOCK(origim_mutex_);
    os <<")";
    SBar_->SetStatusText(AsciiToWx(os.str()), SBarID_);
  }
  //BIAS_MUTEX_UNLOCK(wxbitmap_mutex_);
}

bool ImageCanvas::GetMousePos(int& x, int& y)
{
  x= mousex_; y=mousey_;
  return ((mousex_>=0) && (mousey_>=0));
}

bool ImageCanvas::HistogrammIsShown(){
  if(MyHistoWin_!=NULL && MyHistoWin_->IsShown())
    return true;
  else
    return false;
}

///////////////////////////////////////////////////////////

BEGIN_EVENT_TABLE (ImageCanvas, wxScrolledWindow)
EVT_PAINT       (ImageCanvas::OnPaint)
EVT_ERASE_BACKGROUND       (ImageCanvas::OnErase)
EVT_MOTION      (ImageCanvas::OnMouseMove)
EVT_RIGHT_DOWN  (ImageCanvas::OnRightMouseButton)
EVT_LEAVE_WINDOW(ImageCanvas::OnMouseLeave)
//EVT_SIZE        (ImageCanvas::OnSize)
//EVT_KEY_DOWN    (ImageCanvas::OnKeyPress)
END_EVENT_TABLE()

//void ImageCanvas::OnErase(wxEraseEvent& event)
void ImageCanvas::OnErase(wxEraseEvent& )
{
}


void ImageCanvas::OnPaint(wxPaintEvent & event)
{
  wxScrolledWindow::OnPaint(event);
  wxPaintDC dc(this);

  PrepareDC(dc);

  int CanvasWidth, CanvasHeight;
  GetSize(&CanvasWidth,&CanvasHeight);


  BIAS_MUTEX_LOCK(wxbitmap_mutex_);
  if (wximage_ && wximage_->Ok()){
    if (Bitmap_ != NULL) delete Bitmap_;
    if (ShowScaled_)  {
      wximage_->Rescale(ScaledImageSizeX_, ScaledImageSizeY_);
    }
    Bitmap_ = new wxBitmap( *wximage_ );
  }

  if (Bitmap_ && Bitmap_->Ok() && dc.Ok()) {
    // set ClearDrawingContext_ to fals to prevent flickering
    if (ClearDrawingContext_) dc.Clear();
    dc.DrawBitmap( *Bitmap_, 0, 0 );

  }
  BIAS_MUTEX_UNLOCK(wxbitmap_mutex_);

  //wxPostEvent(this->GetParent()->GetEventHandler(), event);
}

//void ImageCanvas::OnMouseLeave(wxMouseEvent &event)
void ImageCanvas::OnMouseLeave(wxMouseEvent &)
{
  if (SBar_) SBar_->SetStatusText(wxT(""), SBarID_);
  mousex_=mousey_=-1;
}

void ImageCanvas::OnMouseMove(wxMouseEvent &event)
{
  wxClientDC dc(this);
  PrepareDC(dc);
  if (GetParent()) GetParent()->PrepareDC(dc);
  wxPoint pos = event.GetPosition();
  mousex_ = dc.DeviceToLogicalX( pos.x );
  mousey_ = dc.DeviceToLogicalY( pos.y );

  if (ShowScaled_) {
    // if we are in scaled mode, correct coordinates by scale !
    double w = ScaledImageSizeX_, h = ScaledImageSizeY_;
    double ow = origim_.GetWidth(), oh = origim_.GetHeight();
    double ratiox = ow/w , ratioy = oh/h;
    mousex_ = (int)rint(mousex_ * ratiox);
    mousey_ = (int)rint(mousey_ * ratioy);
  }
  UpdateStatusBar((unsigned)mousex_, (unsigned)mousey_);
}

void ImageCanvas::OnRightMouseButton(wxMouseEvent &event)
{
  wxClientDC dc(this);
  PrepareDC(dc);
  wxPoint pos = event.GetPosition();
  unsigned x = (unsigned) dc.DeviceToLogicalX( pos.x );
  unsigned y = (unsigned) dc.DeviceToLogicalY( pos.y );
  Zoom(x, y);
}

void ImageCanvas::OnSize(wxSizeEvent& event)
{
  Fit();
  event.Skip();
}

// void ImageCanvas::OnKeyPress(wxKeyEvent &event)
// {
//   cerr <<"ImageCanvas::OnKeyPress\n";
//   event.Skip();
// }


//////////////////////////////////////////////////////////////////////////

namespace BIAS {
  enum ScaledImageCanvas_IDs {
    ID_Scale=1,
    ID_Offset,

    ID_Scale_L,
    ID_Scale_R,
    ID_Scale_Steps,

    ID_Offset_L,
    ID_Offset_R,
    ID_Offset_Steps,

    ID_Check_Auto_Sc_Of
  };
}

ScaledImageCanvas::ScaledImageCanvas(wxWindow* parent, wxStatusBar* sbar,
                   int sbarid, wxWindowID id,
                   const wxPoint& pos, const wxSize& size,
                   long style, const wxString& name)
                   : wxPanel(parent, id, pos, size, style, name), scaledf_()
{
  MiaScWi_ = new ImageCanvas(this, sbar, sbarid, -1, wxDefaultPosition,
    wxSize(320, 240), wxHSCROLL | wxVSCROLL, name);

  //count of steps in sliders
  slider_sc_digits = 100;
  slider_of_digits = 100;

  ///-------------SIZES---------------
  //Box for values of sliders/slider borders
  box_size.SetHeight(20);
  box_size.SetWidth(85);

  //Box for values of slider steps
  box_size_steps.SetHeight(20);
  box_size_steps.SetWidth(60);

  //Box for CheckBoxes
  box_size_checks.SetHeight(20);
  box_size_checks.SetWidth(150);
  ///---------------------------------

  ///---HISTO/AUTOSCALE CHECK BOXES---
  Check_Auto_Sc_Of_ = new wxCheckBox(this, ID_Check_Auto_Sc_Of, 
                                wxT("Auto(Scale/Offset)"), wxDefaultPosition,
                                box_size_checks);
  ///---------------------------------

  wxBoxSizer *hbox_scale = new wxBoxSizer(wxHORIZONTAL);
  wxBoxSizer *hbox_offset = new wxBoxSizer(wxHORIZONTAL);

  
  MinScale_ =    0.1;
  MaxScale_ =  100.0;
  MinOffSet_= -100.0;
  MaxOffSet_=  100.0;

  ///---SCALE SLIDER---
  int scale_value = (int)(slider_sc_digits/2);
  ScaleSL_ = new wxSlider(this, ID_Scale, scale_value, 0,
                          slider_sc_digits, wxDefaultPosition,
                          wxSize(-1, -1), wxSL_HORIZONTAL);
  scale_old = scale_value;

  double scal_v = ScaleSL_->GetValue();
  Scale_Text_ = new wxStaticText(this, -1, (wxString)((wxChar*)&scal_v), 
                                 wxDefaultPosition, box_size, wxSP_ARROW_KEYS);

  Scale_L_ = new wxTextCtrl(this, ID_Scale_L, wxT("0.1"), 
                            wxDefaultPosition, box_size, wxTE_PROCESS_ENTER);
  Scale_R_ = new wxTextCtrl(this, ID_Scale_R, wxT("100.0"), 
                            wxDefaultPosition, box_size, wxTE_PROCESS_ENTER);

  //constructing steps-text-window and putting current value in it
  ostringstream sc_stream; sc_stream << slider_sc_digits;
  wxString value_sc_steps = AsciiToWx(sc_stream.str());
  Scale_Steps_ = new wxSpinCtrl(this, ID_Scale_Steps, value_sc_steps,
                                wxDefaultPosition, box_size_steps,
                                wxSP_ARROW_KEYS, 1, 2000);
  
  hbox_scale->Add(new wxStaticText( this, -1, _T("Scale: ") ), 0, wxALL, 3);
  hbox_scale->Add(Scale_Text_);
  hbox_scale->Add(new wxStaticText( this, -1, _T(" [ ") ), 0, wxALL, 3);
  hbox_scale->Add(Scale_L_);
  hbox_scale->Add(ScaleSL_, 1, wxALL | wxGROW, 3);
  hbox_scale->Add(Scale_R_);
  hbox_scale->Add(new wxStaticText( this, -1, _T("] Steps: ") ), 0, wxALL, 3);
  hbox_scale->Add(Scale_Steps_);
  hbox_scale->Add(new wxStaticText( this, -1, _T(" | ") ), 0, wxALL, 3);

  wxStaticText* temp = new wxStaticText(this, -1, wxT(""), 
                                 wxDefaultPosition, box_size_checks, wxSP_ARROW_KEYS);
  hbox_scale->Add(temp);
  ///------------------

  ///---OFFSET SLIDER---
  int offset_value = (int)(slider_of_digits/2);
  OffsSL_ = new wxSlider(this, ID_Offset, offset_value, 0,
                         slider_of_digits, wxDefaultPosition, 
                         wxSize(-1, -1), wxSL_HORIZONTAL);
  offset_old = offset_value;

  double offs_v = OffsSL_->GetValue();
  Offset_Text_ = new wxStaticText(this, -1, (wxString)((wxChar*)&offs_v), 
                                 wxDefaultPosition, box_size, wxSP_ARROW_KEYS);

  Offset_L_ = new wxTextCtrl(this, ID_Offset_L, wxT("-100.0"),
                             wxDefaultPosition, box_size, wxTE_PROCESS_ENTER);
  Offset_R_ = new wxTextCtrl(this, ID_Offset_R, wxT("100.0"),
                             wxDefaultPosition, box_size, wxTE_PROCESS_ENTER);

  //constructing steps-text-window and putting current value in it
  ostringstream of_stream; of_stream << slider_of_digits;
  wxString value_of_steps = AsciiToWx(of_stream.str());
  Offset_Steps_ = new wxSpinCtrl(this, ID_Offset_Steps, value_of_steps,
                                 wxDefaultPosition, box_size_steps,
                                 wxSP_ARROW_KEYS, 1, 2000);

  hbox_offset->Add(new wxStaticText( this, -1, _T("Offset: ") ), 0, wxALL, 3);
  hbox_offset->Add(Offset_Text_);
  hbox_offset->Add(new wxStaticText( this, -1, _T(" [ ") ), 0, wxALL, 3);
  hbox_offset->Add(Offset_L_);
  hbox_offset->Add(OffsSL_, 1, wxALL | wxGROW, 3);
  hbox_offset->Add(Offset_R_);
  hbox_offset->Add(new wxStaticText( this, -1, _T("] Steps: ") ), 0, wxALL, 3);
  hbox_offset->Add(Offset_Steps_);
  hbox_offset->Add(new wxStaticText( this, -1, _T(" | ") ), 0, wxALL, 3);
  hbox_offset->Add(Check_Auto_Sc_Of_);
  ///-------------------

  wxBoxSizer *vbox = new wxBoxSizer(wxVERTICAL);

  vbox->Add(MiaScWi_, 1, wxALL | wxGROW, 0);
  vbox->Add(hbox_scale, 0, wxGROW, 0);
  vbox->Add(hbox_offset, 0, wxGROW, 0);
  vbox->SetSizeHints(this);
  SetSizer(vbox);

  p_StatBar=NULL;

  //Shut the AutoScale/Offset Box ON
  AutoadaptScale_ = true;
  Check_Auto_Sc_Of_->SetValue(true);

  
  //SetScaleRange(0.5, 50.0);
  //SetOffsetRange(-100.0, 100.0);
  SetRanges(0.5, 50.0, -100.0, 100.0);

  //Set that the ranges of sliders were:
  //scale: from 0.1*value to 10*value
  //offset:from value-100 to value+100
  //SetNiceRanges(10.0, 100.0);
}


ScaledImageCanvas::~ScaledImageCanvas()
{
  if (MiaScWi_) delete MiaScWi_;
  if (ScaleSL_) delete ScaleSL_;
  if (OffsSL_) delete OffsSL_;
}


void ScaledImageCanvas::SetStatusBarPtr(wxStatusBar* p_sbar){
  p_StatBar=p_sbar;
}

void ScaledImageCanvas::SetStatusText(const std::string & s){
  if(p_StatBar==NULL){
    wxMessageBox( AsciiToWx(string("p_StatBar NULL. " + s)) );
    return;
  }
  BIASASSERT(p_StatBar!=NULL);
  p_StatBar->SetStatusText(AsciiToWx(s));
}


void ScaledImageCanvas::Show(ImageBase& im, string name)
{
  ImageName_=name;
  switch (im.GetStorageType()) {
  case ImageBase::ST_float:
    origim_=im;
    break;
    //   case ImageBase::ST_unsignedshortint:
  default:
    ImageConvert::ConvertST(im, origim_, ImageBase::ST_float);
    break;
  }
  scaledf_=(Image<float>)origim_;
  if (AutoadaptScale_) {
    double scale=0, shift=0;
    scaledf_.CalcScaleShift(0.0, 255.0,scale, shift);
    //scaledf_.CalcScaleShift(0.0, 255.0,scale, shift);

    double a1, a2, b1, b2;
    if(scale < 0){
      a1 = scale*10; a2 = scale/10;
    }
    else if(scale > 0){
      a1 = scale/10; a2 = scale*10;
    }
    else{
      a1 = 0.0; a2 = 10.0;
    }

    if(shift < 0){
      b1 = shift*10; b2 = shift/10;
    }    
    else if(shift > 0){
      b1 = shift/10; b2 = shift*10;
    }     
    else{
      b1 = -50.0; b2 = 50.0;
    }

    SetRanges(a1, a2, b1, b2);

#ifdef BIAS_DEBUG
    cout << "Auto-Scale: " << scale << endl;
    cout << "Auto-Shift: " << shift << endl;
#endif    
    SetOffset(shift, false);
    SetScale(scale, false);
  }
  UpdateScaleOffset();
}


void ScaledImageCanvas::Show(PyramidImage<float>& pim, string name)
{
  Image<float> im;
  pim.GetSingleImage(im);
  Show(im, name);
}

int ScaledImageCanvas::SetScale(double c_scale)
{
  return SetScale(c_scale, true);
}

int ScaledImageCanvas::SetScale(double scale, bool update)
{
  if (scale<MinScale_){
    ScaleSL_->SetValue(0);
    cout << "invalid scale (too small) - clipping it to minimum" << endl;
  }
  else if (scale>MaxScale_){
    ScaleSL_->SetValue(slider_sc_digits);
    cout << "invalid scale (too big) - clipping it to maximum" << endl;
  }
  else{
    int value=(int)rint((scale-MinScale_)*(double)slider_sc_digits/
                        (MaxScale_-MinScale_));
    BIASASSERT(value>=0);
    //BIASASSERT(value<=1000);
    ScaleSL_->SetValue(value);
  }

  if(update)
    UpdateScaleOffset();
  return 0;
}

int ScaledImageCanvas::SetOffset(double offset)
{
  return SetOffset(offset, true);
}

int ScaledImageCanvas::SetOffset(double offset, bool update)
{
  if (offset < MinOffSet_){
    OffsSL_->SetValue(0);
    cout << "invalid offset (too small) - clipping it to minimum" << endl;
  }
  else if(offset > MaxOffSet_){
    OffsSL_->SetValue(slider_of_digits);
    cout << "invalid offset (too big) - clipping it to maximum" << endl;
  }
  else{
    int value=(int)rint((offset-MinOffSet_)*(double)slider_of_digits/
                        (MaxOffSet_-MinOffSet_));
    //BIASASSERT(value>=0);
    //BIASASSERT(value<=1000);
    OffsSL_->SetValue(value);
  }

  if(update)
    UpdateScaleOffset();
  return 0;
}

void ScaledImageCanvas::UpdateScaleOffset()
{
  if (!origim_.IsEmpty()){
    scaledf_=(Image<float>)origim_;
    double sc=1.0, ofs=0.0;

    //REAL VALUES OF SCALE AND OFFSET
    sc=((double)(ScaleSL_->GetValue())*(MaxScale_-MinScale_)/
      (double)slider_sc_digits)+MinScale_;

    ofs=((double)(OffsSL_->GetValue())*(MaxOffSet_-MinOffSet_)/
      (double)slider_of_digits)+MinOffSet_;

    //VALUES OF SLIDERS
#ifdef BIAS_DEBUG
    cout << "scale old: " << scale_old << endl;
#endif
    scale_old = ScaleSL_->GetValue();
#ifdef BIAS_DEBUG
    cout << "scale new: " << scale_old << endl;
#endif

#ifdef BIAS_DEBUG
    cout << "offset old: " << offset_old << endl;
#endif
    offset_old = OffsSL_->GetValue();
#ifdef BIAS_DEBUG
    cout << "offset new: " << offset_old << endl;
#endif

    ostringstream sc_stream; sc_stream << sc;
    ostringstream of_stream; of_stream << ofs;
    
    wxString value_sc = AsciiToWx(sc_stream.str());
    wxString value_of = AsciiToWx(of_stream.str());
    
    Scale_Text_->SetLabel(value_sc);
    Offset_Text_->SetLabel(value_of);
    
    //Set NEW VALUES
    ScaleSL_->SetValue(scale_old);
    OffsSL_->SetValue(offset_old);

    //cerr << "Scale: "<<sc<<"\tOffset: "<<ofs<<endl;
    stringstream ss;
    ss<<"scale "<<sc<<" offset: "<<ofs;
    //if (p_parent!=NULL) {
    //  p_parent->SetStatusText(ss.str().c_str() );
    //} else {
    cout<<ss.str()<<endl;
    //}
    
    scale_border_l_value = Scale_L_->GetValue();
    scale_border_r_value = Scale_R_->GetValue();
    offset_border_l_value = Offset_L_->GetValue();
    offset_border_r_value = Offset_R_->GetValue();
      
    scaledf_.ScaleShift(sc, ofs);
    scaledf_.BelowThresholdToValue(0.0, 0.0);
    scaledf_.AboveThresholdToValue(255.0, 255.0);

    float **sida=scaledf_.GetImageDataArray();
    float **ida=(float **)origim_.GetImageDataArray();
    unsigned w=origim_.GetWidth(), h=origim_.GetHeight();
    unsigned int channels =origim_.GetChannelCount();
    for (unsigned y=0; y<h; y++) {
      for (unsigned x=0; x<w; x++) {
        for (unsigned int c=0; c<channels; c++) {
          if (ida[y][channels*x+c]==0.0) sida[y][channels*x+c]=0.0;
        }
      }
    }
    if (ImageConvert::ConvertST(scaledf_, ucim_,
      ImageBase::ST_unsignedchar)!=0){
        BIASERR("error converting to unsigned char");
      }
      if (ucim_.GetColorModel()!=ImageBase::CM_RGB){
        ImageConvert::Convert(ucim_, rgbim_, ImageBase::CM_RGB);
      } else {
        rgbim_=ucim_;
      }
      MiaScWi_->Show(origim_, rgbim_, ImageName_);

      //BIASASSERT(this->SBar_ != NULL);
      ss.str("");
      ss<<"scale= ";
      ss<<sc;
      ss<<" offset=";
      ss<<ofs;
      //MiaScWi_->SBar_->SetStatusText(ss.str(), 0); // SBarID_ );
      //     cout<<"DBG "<<ss.str()<<endl;
      /// \todo: write ofsfet and scale to StatusText instead of cout (JW)
  }
}

void ScaledImageCanvas::UpdateScaleOffsetBorders()
{
  double scale = ((double)(ScaleSL_->GetValue())*(MaxScale_-MinScale_)/
                  (double)slider_sc_digits)+MinScale_;
  
  double offset = ((double)(OffsSL_->GetValue())*(MaxOffSet_-MinOffSet_)/
                   (double)slider_of_digits)+MinOffSet_;

  //CHECK IF VALUES IN BORDER-VALUE-TEXTS ARE CORRECT
  double scale_l = atof(WxToAscii(Scale_L_->GetValue()));
  double scale_r = atof(WxToAscii(Scale_R_->GetValue()));

  if(scale_l > scale_r){
    //that means thant Scale-right-border was changed
    if(scale_l_old == scale_l){
      ostringstream sc_stream;
      sc_stream << scale_l;
      Scale_R_->SetValue(AsciiToWx(sc_stream.str()));
      scale_r = scale_l;
    }
    //that means thant Scale-left-border was changed
    if(scale_r_old == scale_r){
      ostringstream sc_stream; 
      sc_stream << scale_r;
      Scale_L_->SetValue(AsciiToWx(sc_stream.str()));
      scale_l = scale_r;
    }
  }

  //UPDATE MIN_SCALE, MAX_SCALE
  MinScale_ = scale_l;
  MaxScale_ = scale_r;

  scale_l_old = scale_l;
  scale_r_old = scale_r;
  
  //CHECK IF VALUES IN BORDER-VALUE-TEXTS ARE CORRECT
  double offset_l = atof(WxToAscii(Offset_L_->GetValue()));
  double offset_r = atof(WxToAscii(Offset_R_->GetValue()));
  
  if(offset_l > offset_r){
    //that means thant Scale-right-border was changed
    if(offset_r_old != offset_r){
      ostringstream of_stream;
      offset_r = offset_l;
      of_stream << offset_r;
      Offset_R_->SetValue(AsciiToWx(of_stream.str()));
    }
    //that means thant Scale-left-border was changed
    else if(offset_l_old != offset_l){
      ostringstream of_stream;
      offset_l = offset_r;
      of_stream << offset_l;
      Offset_L_->SetValue(AsciiToWx(of_stream.str()));
    }
  }

  offset_l_old = offset_l;
  offset_r_old = offset_r;

  //UPDATE MIN_OFFSET, MAX_OFFSET
  MinOffSet_ = offset_l;
  MaxOffSet_ = offset_r;

  //UPDATE SCALE OFFSET SLIDERS (+STEPS)

  int scale_new =(int)rint((scale - MinScale_)*(double)slider_sc_digits/
                           (MaxScale_ - MinScale_));
  int offset_new = (int)rint((offset - MinOffSet_)*(double)slider_of_digits/
                             (MaxOffSet_ - MinOffSet_));

  if(scale_new > (int)slider_sc_digits)
    ScaleSL_->SetValue((int)slider_sc_digits);
  else if(scale_new < 0)
    ScaleSL_->SetValue((int)0);
  else
    ScaleSL_->SetValue((int)scale_new);

  if(offset_new > (int)slider_of_digits)
    OffsSL_->SetValue((int)slider_of_digits);
  else if(offset_new < 0)
    OffsSL_->SetValue((int)0);
  else
    OffsSL_->SetValue((int)offset_new);
}

ImageCanvas* ScaledImageCanvas::GetImageCanvas(){
  return MiaScWi_;
}

BIAS::ImageBase & ScaledImageCanvas::GetOrigimg()
{
  return origim_;
}


BEGIN_EVENT_TABLE (ScaledImageCanvas, wxPanel)
EVT_CHECKBOX(ID_Check_Auto_Sc_Of, ScaledImageCanvas::OnAutoScaleOffset)

EVT_TEXT_ENTER(ID_Scale_L, ScaledImageCanvas::OnScaleOffsetBorders)
EVT_TEXT_ENTER(ID_Scale_R, ScaledImageCanvas::OnScaleOffsetBorders)
EVT_SPINCTRL(ID_Scale_Steps, ScaledImageCanvas::OnScaleOffsetSteps)

EVT_TEXT_ENTER(ID_Offset_L, ScaledImageCanvas::OnScaleOffsetBorders)
EVT_TEXT_ENTER(ID_Offset_R, ScaledImageCanvas::OnScaleOffsetBorders)
EVT_SPINCTRL(ID_Offset_Steps, ScaledImageCanvas::OnScaleOffsetSteps)

EVT_SCROLL(ScaledImageCanvas::OnScaleOffset)
END_EVENT_TABLE()

void ScaledImageCanvas::SetRanges(double value_sc_double_l, 
                                  double value_sc_double_r,
                                  double value_of_double_l,
                                  double value_of_double_r){
    ostringstream sc_stream_l; sc_stream_l << value_sc_double_l;
    Scale_L_->SetValue(AsciiToWx(sc_stream_l.str()));
    ostringstream sc_stream_r; sc_stream_r << value_sc_double_r;
    Scale_R_->SetValue(AsciiToWx(sc_stream_r.str()));
    
    ostringstream of_stream_l; of_stream_l << value_of_double_l;
    Offset_L_->SetValue(AsciiToWx(of_stream_l.str()));
    ostringstream of_stream_r; of_stream_r << value_of_double_r;
    Offset_R_->SetValue(AsciiToWx(of_stream_r.str()));
    
    UpdateScaleOffsetBorders();
    UpdateScaleOffset();
}

void ScaledImageCanvas::SetNiceRanges(double scale_factor, 
                                      double offset_factor){
  //correct the real slider-values from the beginning
  ScaleSL_->SetValue((int)rint((double)(ScaleSL_->GetValue())));
  OffsSL_->SetValue((int)rint((double)(OffsSL_->GetValue())));

  if(scale_factor > 0.0 &&
     offset_factor > 0.0){
    //get real scale and offset values
    double value_sc_double = ((double)(ScaleSL_->GetValue())*
                              (MaxScale_-MinScale_)/
                              (double)slider_sc_digits)+MinScale_;
    double value_of_double = ((double)(OffsSL_->GetValue())*
                              (MaxOffSet_-MinOffSet_)/
                              (double)slider_of_digits)+MinOffSet_;
    
    //set borders of sliders as scale: (scale_value/scale_factor, 
    //                                  scale_value*scale_factor);
    //                         offset: (offset_value-offset_factor, 
    //                                  offset_value+offset_factor);

    double value_sc_double_l = value_sc_double/scale_factor;
    ostringstream sc_stream_l; sc_stream_l << value_sc_double_l;
    Scale_L_->SetValue(AsciiToWx(sc_stream_l.str()));
    double value_sc_double_r = value_sc_double*scale_factor;
    ostringstream sc_stream_r; sc_stream_r << value_sc_double_r;
    Scale_R_->SetValue(AsciiToWx(sc_stream_r.str()));
    
    double value_of_double_l = value_of_double-offset_factor;
    ostringstream of_stream_l; of_stream_l << value_of_double_l;
    Offset_L_->SetValue(AsciiToWx(of_stream_l.str()));
    double value_of_double_r = value_of_double+offset_factor;
    ostringstream of_stream_r; of_stream_r << value_of_double_r;
    Offset_R_->SetValue(AsciiToWx(of_stream_r.str()));
    
    UpdateScaleOffsetBorders();
    UpdateScaleOffset();
  }
}

void ScaledImageCanvas::OnScaleOffset(wxScrollEvent & )
{
  if(scale_old != ScaleSL_->GetValue() ||
     offset_old != OffsSL_->GetValue()){
    AutoadaptScale_ = false;
    Check_Auto_Sc_Of_->SetValue(false);
    UpdateScaleOffset();
  }
}

void ScaledImageCanvas::OnScaleOffsetBorders(wxCommandEvent & )
{
  if(scale_border_l_value != Scale_L_->GetValue() ||
     scale_border_r_value != Scale_R_->GetValue() ||
     offset_border_l_value != Offset_L_->GetValue() ||
     offset_border_r_value != Offset_R_->GetValue()){
    UpdateScaleOffsetBorders();
    /**
    cout << MinScale_ << " to " << MaxScale_ << " - Scale Bord" << endl;
    cout << MinOffSet_ << " to " << MaxOffSet_ << " - Offset Bord" << endl;
    **/
    UpdateScaleOffset();
    /**
    cout << MinScale_ << " to " << MaxScale_ << " - Scale Update" << endl;
    cout << MinOffSet_ << " to " << MaxOffSet_ << " - Offset Update" << endl;
    **/
  }
} 

void ScaledImageCanvas::OnScaleOffsetSteps(wxSpinEvent & )
{
  //UPDATE SLIDER-VALUES
  int value_sc_steps_new = Scale_Steps_->GetValue();
  int value_of_steps_new = Offset_Steps_->GetValue();

  double diff_sc = ((double)value_sc_steps_new)/((double)slider_sc_digits);
  if(diff_sc != 1.0){
    ScaleSL_->SetValue((int)rint((double)ScaleSL_->GetValue()*diff_sc));
    ScaleSL_->SetRange(0, value_sc_steps_new);
  }

  double diff_of = ((double)value_of_steps_new)/((double)slider_of_digits);
  if(diff_of != 1.0){
    OffsSL_->SetValue((int)rint((double)OffsSL_->GetValue()*diff_of));
    OffsSL_->SetRange(0, value_of_steps_new);
  }
  
  slider_sc_digits = value_sc_steps_new;
  slider_of_digits = value_of_steps_new;
  
  cout << endl;
  cout << slider_sc_digits << endl;
  cout << slider_of_digits << endl;

  UpdateScaleOffset();
} 

void ScaledImageCanvas::OnAutoScaleOffset(wxCommandEvent& /*event*/)
{
  if(origim_.GetStorageType() == ImageBase::ST_float) {
    AutoadaptScale_ = !AutoadaptScale_;
    if (AutoadaptScale_) {
      double scale=0, shift=0;
      scaledf_=(Image<float>)origim_;
      scaledf_.CalcScaleShift(0.0, 255.0,scale, shift);
#ifdef BIAS_DEBUG
      cout << "new scale  : " << scale/10 << " : " << scale*10 << endl;
      cout << "new offset : " << shift/10 << " : " << shift*10 << endl;
#endif

      double a1, a2, b1, b2;
      if(scale < 0){
        a1 = scale*10; a2 = scale/10;
      }
      else if(scale > 0){
        a1 = scale/10; a2 = scale*10;
      }
      else{
        a1 = 0.0; a2 = 10.0;
      }
      
      if(shift < 0){
        b1 = shift*10; b2 = shift/10;
      }    
      else if(shift > 0){
        b1 = shift/10; b2 = shift*10;
      }     
      else{
        b1 = -50.0; b2 = 50.0;
      }

      SetRanges(a1, a2, b1, b2);

      SetScale(scale, false);
      SetOffset(shift, false);
    }
    UpdateScaleOffset();
  }
}

void ScaledImageCanvas::OnHisto(wxCommandEvent& /*event*/)
{
}
