#ifndef __ImageCanvasCheckerboardExtract_hh__
#define __ImageCanvasCheckerboardExtract_hh__

#include "ImageCanvas.hh"
#include <Gui/StringConv.hh>
#include <Base/Geometry/HomgPoint2D.hh>

namespace BIAS {
  
  /** @class ICECCallBackInterface
      @ingroup g_gui
      @brief callback interface for ImageCanvasCheckerBoardExtract

      instructions:
      derive your class e.g. MyFrame from ICECallBackInterface
      and override the method CheckerboardSelected(...)to react to a finished 
      selection of a Checkerboard
      
      also declare a pointer to ImageCanvasCheckerBoardExtract, e.g.
      ImageCanvasCheckerBoardExtract *_ic;
      
      to select The Checkerboard corners then call:
      _ic->SelectCheckerBoardCorners(*this);
      where *this is a reference to a ICECCallBackInterface
      (if your class is derived from ICECCallBackInterface this is *this)
      
      @author ischiller 07/2007*/
  class BIASGui_EXPORT ICECCallbackInterface{
  public:
    ICECCallbackInterface();
    virtual ~ICECCallbackInterface();
    /// this function should be overloaded
    virtual void CheckerboardSelected(std::vector<BIAS::HomgPoint2D> &corners, 
                                      bool bOnlyFourPointsGiven);
  };

  /** @class ImageCanvasEnhanced
      @ingroup g_gui
      @brief provides functionality for selection of rectangle, point, ...
      
      

      @author woelk 08/2004 */
  class BIASGui_EXPORT ImageCanvasCheckerboardExtract : public ImageCanvas
  {
  public:
    ImageCanvasCheckerboardExtract(wxWindow* parent, wxStatusBar* sbar=NULL,
                                   int sbarid=0, wxWindowID id = -1, 
                                   const wxPoint& pos = wxDefaultPosition, 
                                   const wxSize& size = wxSize(640, 480), 
                                   long style = wxHSCROLL | wxVSCROLL, 
                                   const wxString& name = AsciiToWx("CheckerboardExtract"));
    ~ImageCanvasCheckerboardExtract();
    
    /** Call SelectCheckerboardCorners with an object derived from ICECallbackInterface.
        The function CheckerboardSelected() should be overloaded in this object.
        After rectangle selection is finished (4 clicks with left mouse,
        right button cancels) the function obj.CheckerboardSelected() is called
        with 4 coordinates.
        @author ischiller 07/2007 */
    void SelectCheckerboardCorners(ICECCallbackInterface& obj);
    void SetPatternParameters(unsigned int xCorners, unsigned int yCorners);
    void OnLeftMouseButton(wxMouseEvent &event);
    void OnMouseMove(wxMouseEvent &event);
    void OnRightMouseButton(wxMouseEvent &event);
    void OnPaint(wxPaintEvent& event);

  protected:
    int NumLeftClicks_;
    std::vector<BIAS::HomgPoint2D> checkerboardCorners_;
    BIAS::HomgPoint2D first_, second_, third_, forth_;
    bool SelectCheckerboard_;
    unsigned int xCorners_;
    unsigned int yCorners_;
   
    ICECCallbackInterface *CallbackObj_;
    
    void DrawCheckerboardCorners_(BIAS::HomgPoint2D first,
                                  BIAS::HomgPoint2D second, 
                                  BIAS::HomgPoint2D third,
                                  BIAS::HomgPoint2D forth);
    
    void RefreshDrawIm_();

    DECLARE_EVENT_TABLE()
  }; // class 


} // namespace



#endif // __ImageCanvasCheckerboardExtract_hh__
