#ifndef __ImgObjGL_h__
#define __ImgObjGL_h__

#include <bias_config.h>
#include <Base/Common/W32Compat.hh>

#ifndef BIAS_HAVE_OPENGL
#  error Please recompile BIAS with USE_OPENGL to use ImgObjGL
#endif // BIAS_HAVE_OPENGL

//#ifndef BIAS_HAVE_GLEW
//#  error Please recompile BIAS with USE_GLEW to use ImgObjGL
//#endif // BIAS_HAVE_GLEW


#include "biasgl.h"

// BIAS
#include <Base/Image/ImageBase.hh>
#include <Base/Image/Image.hh>

#define IMGOBJ_DEFAULT_MIPMAPPED false
#define IMGOBJ_IMMEDIATE_COPY_DEFAULT true


/** data holder describing OpenGL texture objects in addition to its unique texID.
Please note that querying glTexImage2D(GL_PROXY_TEXTURE_2D,... may be more reliable
See http://www.opengl.org/resources/faq/technical/texture.htm

\bug TODO copy constructor that duplicates the GL pixel data required for std::vector.resize() (major bug, do not resize with std::vector for now!)

@author Jan Woetzel  */
namespace BIAS {
  class BIASGui_EXPORT ImgObjGL {
  public: 
    ~ImgObjGL();

    ImgObjGL();

    //* ctor initializing members JW */
    ImgObjGL(
      const GLuint & id,
      const unsigned int & widthBrutto,
      const unsigned int & heightBrutto,
      const GLenum & target=GL_TEXTURE_2D
      );

    /// set members to defaults
    void InitMembers();

    /// convenience interace for glewInit() cwith checks and ifdefs
    static void InitGlew();

    /// frees the pointer members if we allocated them
    void Release();

    /// deletes the texture memory and frees texture in GL 
    void  DeleteGLTexture();


    /// @return true on classic GL_TEXTURE_2D target (instead of RECT or cubemap)
    bool IsTexture2D() const;

    /// @return true if RECT texture (instead of 2D pow2 or cubemap)
    bool IsRectangleTexture() const;

    /// @return true on cubemap texture
    bool IsCubemap() const;


    /**  creates an OpenGL texture from p_valueImage JW
    if ImmediateCopyData is false the memory is just allocated and not copied from c_img.
    JW */
    void CreateGLTexture(
      const BIAS::ImageBase & c_img, /*zeropad etc is currently missing*/
      const bool & forcePow2tex=true,
      const bool & flipY=true,
      const bool ImmediateCopyData=IMGOBJ_IMMEDIATE_COPY_DEFAULT );

    void CreateGLCubemap(const std::vector< BIAS::ImageBase * > cm,
      const bool ImmediateCopyData=IMGOBJ_IMMEDIATE_COPY_DEFAULT,
      const bool mipmapped=IMGOBJ_DEFAULT_MIPMAPPED );

    /// helper creating cubemap from disk files
    int CreateGLCubemap(
      const std::string & cm0PosX,
      const std::string & cm1NegX,
      const std::string & cm2PosY,
      const std::string & cm3NegY,
      const std::string & cm4PosZ,
      const std::string & cm5NegZ,
      const bool ImmediateCopyData=IMGOBJ_IMMEDIATE_COPY_DEFAULT,
      const bool mipmapped=IMGOBJ_DEFAULT_MIPMAPPED );

    /// helper creating cubemap from disk files
    int CreateGLCubemap(
      const std::vector< std::string > cmFilenames,
      const bool ImmediateCopyData=IMGOBJ_IMMEDIATE_COPY_DEFAULT,
      const bool mipmapped=IMGOBJ_DEFAULT_MIPMAPPED );


    // create an empty cubemap including allocating all mipmap texture levels
    // thus the complete mipmap chain for all faces and levels is created on request
    void CreateGLCubemapEmpty(const unsigned int w, const unsigned int h, 
      GLint gpu_internalFmt=GL_RGBA,
      const bool & mipmapped=false);


    /** set the textur retarget to rectangular or pow 2 texture JW */
    void SetTargetPOW2();
    void SetTarget2D();
    void SetTargetRECT();
    void SetTargetCUBE();

    /** get the GLenum pixel storage type for the corresponding BIAS enum @author Jan Woetzel */
    static GLenum GetGLStorageType(const BIAS::ImageBase::EStorageType & biasST,
      const unsigned int & ByteDepth );

    /** get the GLenum pixel src format for the corresponding BIAS enum ColroModel @author Jan Woetzel */
    static GLenum GetGLPixelFormat(const BIAS::ImageBase::EColorModel & biasCM,
      const unsigned int & channelcount );

    /// JW determin the GPU internal format the src data should be mapped to @author Jan Woetzel 
    static GLint GetGLInternalFormat(
      const BIAS::ImageBase::EStorageType & biasST,
      const BIAS::ImageBase::EColorModel & biasCM,
      const unsigned int & channelcount,
      const unsigned int & ByteDepth,
      const unsigned int & BitDepth );

    ///  interface for the above @author Jan Woetzel 
    static GLint GetGLInternalFormat(const BIAS::ImageBase & img);


  public:
    // OpenGL texture 
    GLuint id; // unique OpengL texture id
    GLenum target; // GL_TEXTURE_2D or GL_TEXTURE_RECTANGLE_NV

    // true if we "created" the above and need to free it in destructor. 
    bool p_id_SelfAllocated;

    GLenum minificationMode;
    GLenum magnificationMode;

    // original feed in size
    unsigned int widthOrig; // should match p_valueImage size if exists
    unsigned int heightOrig;

    // padded to next pow2 size:
    unsigned int widthBrutto;
    unsigned int heightBrutto;

    // image containing the original values
    BIAS::ImageBase * p_valueImage;

    // true if we "newed" the above and have to free it in destructor. 
    // false if we got a pointer which we do not need to clean up.
    bool p_valueImage_SelfAllocated;

  };
} // namespace BIAS

#endif

