#include <iostream>
#include <Base/Image/ImageIO.hh>
#include <Image/Camera.hh>
#include <Image/PlanarRectification.hh>
#include <Image/CylindricalRectification.hh>
#include <Image/SphericalRectification.hh>
#include <Image/RectificationBase.hh>
#include <Utils/Param.hh>
#include <Utils/ThreeDOut.hh>
#include <Base/Image/ImageConvert.hh>

using namespace std;
using namespace BIAS;



int main(int argc, char* argv[])
{
  Param params(true);
  bool* vrml = params.AddParamBool("writeVrml", 
                                   "write out vrml with rect geometry",
                                   false);

  bool* pointsonly = 
    params.AddParamBool("pointsonly", 
                        "dont use texture,but colored vertices",
                        false);

  bool* help = params.AddParamBool("help", "", false, 'h');
  enum {cylindrical=0, spherical, planar} rectTypeID;
  vector<string> rectificationTypes;
  rectificationTypes.push_back("cylindrical");//0
  rectificationTypes.push_back("spherical");  //1
  rectificationTypes.push_back("planar");     //2
  int* rectType = 
    params.AddParamEnum("type", 
                        "what kind of rectification shall be performed?:"
                        " [cylindrical], spherical, planar", 
                        rectificationTypes);

  double* resolution = 
    params.AddParamDouble("resolution", 
                          "scales the resolution of rectified images", 
                          0.8, 0.001);
  int fup = params.ParseCommandLine(argc, argv);
  
  if(argc<3 || *help) {
    cout<<"determines rectification and stores them in name\n";
    cout<<"\nusage:\n";
    cout<<argv[0]<<" [options] <origImgA> <origImgB>\n";
    params.Usage();
    return 0;
  }

  Camera<unsigned char> RGBcamA, camA, RGBcamB, camB;

  if(ImageIO::Load(argv[fup], camA)!=0) {
    cerr<<"unable to load first camera\n";
    return -1;
  }

  if(ImageIO::Load(argv[fup+1], camB)!=0) {
    cerr<<"unable to second first camera\n";
    return -1;
  }

  //   camA.ParseMetaData();
  //   camB.ParseMetaData();

  RectificationBase<unsigned char, unsigned char>* rectifier = NULL;
  RectificationViaProjectionMappingBase<unsigned char, 
    unsigned char>* rectifierPM = NULL;
  
    //choose different rectification at this point
  switch(*rectType) {
  case cylindrical : 
    rectifier = rectifierPM = 
      new CylindricalRectification<unsigned char, unsigned char>(1.3); 
    break;
  case spherical : 
    rectifier = rectifierPM = 
      new SphericalRectification<unsigned char, unsigned char>();
    break;
  case planar : 
    rectifier = rectifierPM = 
      new PlanarRectification<unsigned char, unsigned char>();
    break;
  default:
    cerr<<" unknown rectification type : "
        <<rectificationTypes[*rectType]<<endl;
    return -1;
    break;
  }
  

  rectifierPM->SetScale(*resolution);
  rectifierPM->SetSecondFill(0);
  if( rectifier->SetCameraA(camA) != 0 ) {
    cerr<<"invalid camera type for camera A!\n";
    delete rectifier;
    return -1;
  }

  if( rectifier->SetCameraB(camB) != 0 ) {
    cerr<<"invalid camera type for camera B!\n";
    delete rectifier;
    return -1;
  }
  
  ProjectionParametersBase* rectPPA; 
  ProjectionParametersBase* rectPPB;

  if(rectifierPM->GetClonesOfRectificationParameters(rectPPA, rectPPB) != 0) {
    cerr<<"rectification parameters could not be calculated!\n";
    delete rectifier;
    return -1;
  }
  
  cerr<<"Found rectification parameters! \n ";
  cerr<<"camera A rectification state:\n"<<*rectPPA<<endl<<endl;
  cerr<<"camera B rectification state:\n"<<*rectPPB<<endl<<endl;
  
  cerr<<"beginning rectification ... "<<flush;
  if(rectifier->Rectify()!=0) {
    cerr<<" failed \n";
    delete rectifier;
    delete rectPPA;
    delete rectPPB;
    return -1;
  }
  cerr<<" finished!\n";

  Camera<unsigned char> rectA, rectB, RGBrectA, RGBrectB;

  rectifier->GetRectifiedImageA(rectA);
  rectA.SetProj(Projection(*rectPPA));
  rectA.UpdateMetaData();

  rectifier->GetRectifiedImageB(rectB);
  rectB.SetProj(Projection(*rectPPB));
  rectB.UpdateMetaData();

  string path1, basename1, suffix1;
  string path2, basename2, suffix2;
  SplitName(argv[fup], path1, basename1, suffix1);
  SplitName(argv[fup+1], path2, basename2, suffix2);
  
  string firstName = basename1+"_rect"+suffix1;
  string secondName = basename2+"_rect"+suffix2;
  
  
  if(ImageIO::ExportImage(firstName, rectA)!=0) {
    cerr<<"writing first rectified image failed!\n";      
  }
  if(ImageIO::ExportImage(secondName, rectB)!=0) {
    cerr<<"writing first rectified image failed!\n";      
  }

  if(*vrml) {
    ThreeDOut tdo;
    tdo.AddProjection(camA.GetProj(), RGBAuc(0, 255, 0, 255));
    tdo.AddProjection(camB.GetProj(), RGBAuc(0, 255, 0, 255));

//     tdo.AddProjection(Projection(*rectPPA), RGBAuc(255, 0, 0, 255));
//     tdo.AddProjection(Projection(*rectPPB), RGBAuc(255, 0, 0, 255));
    cerr<<"generating vrmls with rectification data : \n";
    cerr<<"1. vrml containing only coordinate systems ... "<<flush;
    tdo.AddLine(rectPPA->GetC(), rectPPB->GetC());
    bool good = false;
    if(tdo.VRMLOut("rectCoordinates.wrl")!=0) {
      good = false;
    } else {
      good = true;
    }
    if(good) {
      cerr<<"finished";
    } else { 
      cerr<<"failure";
    }
    cerr<<endl;



    cerr<<"2. generating vrml containing projected images ... "<<flush;
    unsigned int width = 0, height = 0;
    rectPPA->GetImageSize(width, height);
    Image<float> depthmap(width, height, 1);
    depthmap.SetZero();
    
    double scale = 3.0;
    if (*pointsonly) {
      for (unsigned int y=0; y<height; y++) {
        for (unsigned int x=0; x<width; x++) {
          HomgPoint2D p2(x,y);
          HomgPoint3D p3 = rectPPA->UnProjectToImagePlane(p2, scale*0.99);
          if (p3.NormL2()<0.1) continue;
          unsigned char color = rectA.GetImageDataArray()[y][x];
          RGBAuc theColor(color,color,color,255);
          tdo.AddPoint(p3, theColor);
        }
      }
    } else {
      RGBrectA.Init(rectA.GetWidth(), rectA.GetHeight(), 3);
      ImageConvert::ToRGB(rectA, RGBrectA);
      TriangleMesh T;
      T.GenerateTexturedCamera(rectPPA, RGBrectA, scale*0.99) ;
      tdo.AddTriangleMesh(T, "textureA.png", "textureA.png");
    }
    
    if (*pointsonly) {
      rectPPB->GetImageSize(width, height);
      for (unsigned int y=0; y<height; y++) {
        for (unsigned int x=0; x<width; x++) {
          HomgPoint2D p2(x,y);
          HomgPoint3D p3 = rectPPB->UnProjectToImagePlane(p2, scale*1.01);
          if (p3.NormL2()<0.1) continue;
          unsigned char color = rectB.GetImageDataArray()[y][x];
          RGBAuc theColor(color,color,color,255);
          tdo.AddPoint(p3, theColor);
        }
      }
    } else {
      RGBrectB.Init(rectB.GetWidth(), rectB.GetHeight(), 3);
      ImageConvert::ToRGB(rectB, RGBrectB);
      TriangleMesh T;
      T.GenerateTexturedCamera(rectPPB, RGBrectB, scale*1.01) ;
      tdo.AddTriangleMesh(T, "textureB.png", "textureB.png");
    }
    Vector3<double> c= rectPPA->GetC() - rectPPB->GetC();

    cerr<<" writing ..."<<endl<<flush;
    tdo.VRMLOut("rectCoordinates_image.wrl");
    cerr<<"writing mipview3d-animfile cylindersweep.anim"<<endl;
    ofstream myanimfile("cylindersweep.anim");
    myanimfile <<"[0] :43 0 0 0 0 1 0 0 0"<<endl
               <<"[0] :43 1000 " <<c[0]<<" "<<c[1]<<" "<<c[2] 
               <<" 1 0 0 0"<<endl;
    cerr<<"3. generating wrl with orig cameras"<<endl<<flush;
    ThreeDOut tdo2;



    RGBcamA.Init(camA.GetWidth(), camA.GetHeight(), 3);
    ImageConvert::ToRGB(camA, RGBcamA);
    RGBcamB.Init(camB.GetWidth(), camB.GetHeight(), 3);
    ImageConvert::ToRGB(camB, RGBcamB);


    TriangleMesh T1, T2;
    T1.GenerateTexturedCamera(camA.GetProj().GetParameters(), RGBcamA, scale);
    T2.GenerateTexturedCamera(camB.GetProj().GetParameters(), RGBcamB, scale);
    tdo2.AddTriangleMesh(T1, "textureAorig.png", "textureAorig.png");
    tdo2.AddTriangleMesh(T2, "textureBorig.png", "textureBorig.png");
    cerr<<" writing ..."<<endl<<flush;
    tdo2.AddProjection(camA.GetProj(), RGBAuc(0, 255, 0, 255), 
                       scale*0.1, scale);
    tdo2.AddProjection(camB.GetProj(), RGBAuc(0, 255, 0, 255), 
                       scale*0.1, scale);
    tdo2.VRMLOut("orig.wrl");
  }


  delete rectPPA;
  delete rectPPB;
  delete rectifier;

  return 0;
}
