/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <Image/CameraParam.hh>
#include <Base/Image/ImageIO.hh>
#include <Image/CameraParam.hh>
#include <Image/SphericalUndistortion.hh>
#include <Image/UndistortionMapping.hh>
#include <sstream>
#include <iomanip>

#ifdef   WIN32
#include <Base/Common/getopt_W32.h>
#else  //not WIN32
#include <getopt.h>
#endif //WIN32

using namespace BIAS;
using namespace std;

void usage()
{
  cerr << " [-s XX] [-f XX] [-p XX] [-p XX] [-z XX] [-c XX] [-x XX] [-y XX] <image>\n";
  cerr << "-s XX Filename of the XML-camera calibration file " 
       << "(default is PtGrey-Omnitech)\n";
  cerr << "-f XX Field Of View in X-Direction in degree, default=100\n";
  cerr << "-z XX Zoom Factor for image size, default=1\n";
  cerr << "-x XX Set view center X, default=PrincipalX\n";
  cerr << "-y XX Set view center Y, default=PrincipalY\n";
  cerr << "-c XX draw circles XX is  const angle in deg, default=off\n";
  cerr << "-p XX Projection Type, default=0\n";
  cerr << "      Projection Types:\n";
  cerr << "        0: All Projections\n";
  cerr << "        1: Perspective projection      - p=k*tan(alpha)\n";
  cerr << "        2: Stereographic projection    - p=k*tan(alpha/2)\n";
  cerr << "        3: Equidistant projection      - p=k*alpha\n";
  cerr << "        4: sine-law projection         - p=k*sin(alpha)\n";
  cerr << "        5: equi-solid-angle projection - p=k*sin(alpha/2)\n";
  cerr << "        6: cylindrical projection\n";
}

int main (int argc, char *argv[])
{
  Image<unsigned char> source, sink;

  int projectionType = 0;
  float zoomFactor = 1;
  int circleAngle = 0;
  int viewCenterX = 0;
  int viewCenterY = 0;
  int FOVx = 100;
  string XMLFile="";

  SphericalUndistortion undist;
  UndistortionMapping<unsigned char, unsigned char> undistMap;

  ostringstream os;

  //  char **images=NULL;

  //  ifstream is;
  //char name[1024];

  if (argc <1) {
    usage();
    return 1;
  }

  //  images = argv;
  char c;
  while ((c = char(getopt(argc, argv, "s:f:z:c:p:x:y:"))) != EOF)
    {
      switch (c){
      case 's':
        if (optarg) {
          XMLFile = optarg;
        } else {
          usage();
          return 1;
        }
        break;
      case 'f':
        if (optarg) {
          FOVx = atoi(optarg);
        } else {
          usage();
          return 1;
        }
        break;
      case 'z':
        if (optarg) {
          zoomFactor = (float)atof(optarg);
        } else {
          usage();
          return 1;
        }
        break;
      case 'p':
        if (optarg) {
          projectionType = atoi(optarg);
        } else {
          usage();
          return 1;
        }
        break;
      case 'c':
        if (optarg) {
          circleAngle = atoi(optarg);
        } else {
          usage();
          return 1;
        }
        break;
      case 'x':
        if (optarg) {
          viewCenterX = atoi(optarg);
        } else {
          usage();
          return 1;
        }
        break;
      case 'y':
        if (optarg) {
          viewCenterY = atoi(optarg);
        } else {
          usage();
          return 1;
        }
        break;
      default:
        usage();
        return 1;
      }
    }

  if (argc<=optind) {
    BIASERR("no image file given!!");
    return -1;
    
  }
  if (ImageIO::Load(argv[optind], source)!=0){
    BIASERR("error loading image "<< argv[optind]);
    return -1;
  }
#ifdef BIAS_HAVE_XML2
  if (XMLFile=="") {
    undist.SetData_OMNITECH_ORIFL_190_3();
  } else {
    undist.Init(XMLFile);
  }
#else // BIAS_HAVE_XML2
    undist.SetData_OMNITECH_ORIFL_190_3();
#endif // BIAS_HAVE_XML2
  if (undist.IsSpherical()) {
    cout << XMLFile << " used for initialization." << endl;
    undistMap.SetCameraParam(undist);
  } else {
    cout << "Error: could not read camera data file or no sphere lens!" 
         << endl;
    BIASABORT;
  }

  bool testProjection=false;
  if (testProjection) {
    double centerX, centerY, radius;
    undist.GetSphereCenter(centerX, centerY, radius);
    cout <<"centerx="<<centerX<<";centery="<<centerY<<";radius="<<radius<<endl; 
    double theta, phi;
    HomgPoint2D hp(298,388);
    undist.TransfCoordImage2Sphere (hp, theta, phi);
    cout << "CoordImage2Sphere" << endl;
    cout <<"icX="<<hp[0]<<";icY="<<hp[1]<<";theta="<<theta<<";phi="<<phi<<endl;
    undist.TransfCoordSphere2Image (theta, phi, hp[0], hp[1]);
    cout << "CoordSphere2Image" << endl;
    cout <<"icX="<<hp[0]<<";icY="<<hp[1]<<";theta="<<theta<<";phi="<<phi<<endl;
    HomgPoint2D hpp;
    KMatrix k;
    k.SetIdentity();
    undist.ProjectCoord(SphericalUndistortion::EQUIDISTANT, k, hp, hpp);
    cout << "ProjectCoord" << endl;
    cout <<"icX="<<hp[0]<<";icY="<<hp[1]<<";projX="<<hpp[0]<<";projY="<<hpp[1]<<endl;

    return 0;
  }
    
  if (circleAngle) {
    undist.DrawCircles(source, circleAngle);
    ImageIO::ExportImage("circled.ppm", source, ImageIO::FF_ppm);
    cout << " created: "<<"circled.ppm"<<endl;
  }

  undist.CorrectVignette(source);

  os.str("");
  KMatrix K;
  if (projectionType==6) {
    undist.ProjectImageCylindric(source, sink, viewCenterX, 
                                 viewCenterY, zoomFactor);
    os << "projection"<< (projectionType-1) <<".ppm";
    if (ImageIO::ExportImage(os.str(), sink, ImageIO::FF_ppm)!=0){
      BIASERR("error writing image "<<os.str());
      return -2;
    }
    cout << " created: "<<os.str()<<endl;    
  } else {
    if (!projectionType) {
      int projType;
      for (projType=0; projType<5; projType++){
        undistMap.ProjectSphereImage(source, sink, 
                                     source.GetWidth(), source.GetHeight(), 
                                     FOVx, 0, viewCenterX, viewCenterY, 
                                     projType, zoomFactor, K);
        os.str("");
        os << "projection" << projType << ".ppm";
        if (ImageIO::ExportImage(os.str(), sink, ImageIO::FF_ppm)!=0){
          BIASERR("error writing image "<<os.str());
          return -2;
        }
        cout << " created: "<<os.str()<<endl;
      }
    } else {
      undistMap.ProjectSphereImage(source, sink, 
                                   source.GetWidth(), source.GetHeight(), 
                                   FOVx, 0, viewCenterX, viewCenterY, 
                                   projectionType-1, zoomFactor, K);
      os.str("");
      os << "projection" << (projectionType-1) << "UM.ppm";
      if (ImageIO::ExportImage(os.str(), sink, ImageIO::FF_ppm)!=0){
        BIASERR("error writing image "<<os.str());
        return -2;
      }
        cout << " created: "<<os.str()<<endl;
    }
  }
    
  return 0;
}
