/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef __BIAS_PMDIMAGEPROC_H__
#define __BIAS_PMDIMAGEPROC_H__

#include <bias_config.h>
#include <Base/Image/Image.hh>

#define __USE_BIAS__
#include <Base/Image/PMDImageIO.hh>  // only after #define __USE_BIAS__ 
#include <Geometry/Projection.hh>
#include <Base/Geometry/KMatrix.hh>
#include <Geometry/HMatrix.hh>
#include <Geometry/Pose.hh>
#include <Base/Common/CompareFloatingPoint.hh>

#ifndef BIAS_HAVE_TIFF
#  error Please recompile BIAS with USE_TIFF
#endif // BIAS_HAVE_TIFF

namespace BIAS {


  class BIASImage_EXPORT PMDImageProc
  {
  public:
    /** @brief Write a PMD-data 2d/3d-tiff image to disk
        @param FileName Output filename, should have the extension .tif
        @param img the 2d-image data to save
        @param depthimg the depth data to save
        @param MetaData2D the Metadata of the 2D-image to save
        @param MetaDataDepth the Metadata of the depth image to use for padding
        @author Birger Streckel **/
    static int Write(const std::string& FileName, 
                     Image<unsigned char> &Img2D, 
                     Image<float> &ImgDepth, 
                     Image<float> &ImgModCoeff, 
                     PMDImageMetaData &MetaData2D,
                     PMDImageMetaData &MetaDataDepth);
    
    /** @brief Read a PMD-data 2d/3d-image from disk, automatically detecting 
        TIFF or XML file format and calling the appropriate subroutine.
        @param FileName complete name of file to be loaded
        @param img to receive the 2d-image data from the file
        @param depthimg to receive the depth-image data from the file
        @param MetaData to receive the Metadata from the file
        @author Birger Streckel **/
    static int Read(const std::string& FileName, 
                    Image<unsigned char> &Img2D, 
                    Image<float> &ImgDepth, 
                    Image<float> &ImgModCoeff, 
                    PMDImageMetaData &MetaData2D, 
                    PMDImageMetaData &MetaDataDepth);
    /** @brief Read a PMD-data 2d/3d-image from TIFF-file.
        @author Birger Streckel
    */
    static int ReadFromTIFF(const std::string& FileName, 
                            Image<unsigned char> &Img2D, 
                            Image<float> &ImgDepth, 
                            Image<float> &ImgModCoeff, 
                            PMDImageMetaData &MetaData2D, 
                            PMDImageMetaData &MetaDataDepth);

#ifdef BIAS_HAVE_XML2
    /** @brief Read a PMD-data 2d/3d-image from PMDtec's XMK format.
        The 1st image is read if the file contains a sequence.
        @author Birger Streckel
    */
    static int ReadFromXML(const std::string& FileName, 
                           Image<unsigned char> &Img2D, 
                           Image<float> &ImgDepth, 
                           Image<float> &ImgModCoeff, 
                           PMDImageMetaData &MetaData2D, 
                           PMDImageMetaData &MetaDataDepth);

    static int ReadFromXML(const std::string& FileName, 
                           std::vector<Image<unsigned char> > &Imgs2D, 
                           std::vector<Image<float> > &ImgDepth, 
                           std::vector<Image<float> > &ImgModCoeff, 
                           std::vector<PMDImageMetaData> &MetaData2D, 
                           std::vector<PMDImageMetaData> &MetaDataDepth);

#endif

    /** @brief Read PMD-image from 3 space separated text files.
        @author Birger Streckel
    */
    static int ReadFromTXT(int width, int height,
                           const std::string& FileName2D, 
                           const std::string& FileNameDepth, 
                           const std::string& FileNameModCoeff, 
                           Image<unsigned char> &Img2D, 
                           Image<float> &ImgDepth, 
                           Image<float> &ImgModCoeff, 
                           PMDImageMetaData &MetaDataDepth);
    
    /** @brief convert a depth image such that the depth and the grey value 
        at the same pixel position belong to the same ray. The transformation 
        information is taken from the metadata. MetaDataDepth is changed
        for the converted depth image and is identical to MetaData2D afterwards
        @param ImgDepth is the image to convert. This image has to contain 
        depth information and is changed in size and contents.
        @param MetaData2D this is the metadata for the 2D-image.
        @param MetaDataDepth the metadata for the depth image. Is adapted 
        to reflect the image transformation.
        @author Birger Streckel **/
    static int FitDepthTo2DImage(Image<float> &ImgDepth, 
                                 const PMDImageMetaData &MetaData2D, 
                                 PMDImageMetaData &MetaDataDepth);

    static int FitDepthTo2DImage(Image<float> &ImageDepth,
                                 const BIAS::Projection &proj2D,
                                 const BIAS::Projection &projDepth);


    static int FitModCoeffTo2DImage(Image<float> &ImgModCoeff, 
                                    const Image<float> &ImgDepth, 
                                    const PMDImageMetaData &MetaData2D, 
                                    PMDImageMetaData &MetaDataDepth);

    static KMatrix KFromMeta(const PMDImageMetaData &MetaData);
    static Pose PoseFromMeta(const PMDImageMetaData &MetaData);
    static PMDImageMetaData MetaFromPoseAndK(const Pose &Po,
                                             const KMatrix &K, int width, 
                                             int height, int channels); 

    /** Give the standard deviation for a pixel in the depth image.
     *  Measurement precision polynome was taken from:
     *  M.Stommel K.-D.Kuhnert "Fusion of Stereo-Camera and PMD-Camera
     *  Data for Real-Time Suited Precise 3D Environment Reconstruction"
     *  @returns the sigma of the pixel depth depending on the precision
     *    formula and the surrounding pixels.
     */
    static double CalcSigmaDepth(const Image<float> &ImgDepth, 
                                 const unsigned int x, 
                                 const unsigned int y,
                                 const unsigned int HalfWinSize);


    /** @brief This function denoises a PMD-Depth image by Median and 
     *  variance filtering. Sets noise pixels to zero.
     */
    static int DenoiseDepthImage(Image<float> &ImgDepth);

    /** @brief This function adds noise to a depth image.
     * The intention is to add noise to a synthetically 
     * created depth image. The noise sigma is similar to
     * that of a good conditioned PMD-measurement.
     */
    static int AddNoiseToDepthImage(Image<float> &ImgDepth);


    /** @brief This function scales the depth image with the given 
     * scale factor.
     */
    static void ScaleDepthImage(Image<float> &ImgDepth, double scaleFactor);

    /* @brief PMD-Camera Depth calibration 
       Uses the results from: M.Lindner, A.Kolb
       "Lateral and Depth Calibration of PMD-Distance Sensors"
       @param uncalibrated depth value
       @returns calibrated depth
    */
    static double CalibrateDepthValue(double depth);

    /*
      @brief Prepare PMD-Depth calibration with polynomial
      Uses the results from your own measurements
      if left empty predefined measurements (ischiller) are used
      @param vector with control points
      @param vector with measurement points
      @author ischiller
    */
    static int PreparePolyDepthCalibration(std::vector<double> Measurements,
                                           std::vector<double> Values);
    
    /* 
       @brief Get calibrated depth from polynomial
       @warning PreparePolyDepthCalibration has to be called first
       @param The current uncalibrated depth
       @return The calibrated depth value after polanomial computation
       @author ischiller       
    */
    static double CalibrateDepthValuePoly(double depth);

  protected:

    /** @brief These methods transform the depth image to fit the 
     *  2D-image. After transformation the Depth for every 2D-point
     *  is at the same pixel position in the depth image. There are 
     *  3 steps of transformaiton complexity: Only K, Only Rot+K, Free.
     *  @author Birger Streckel
     */
    static int FitDepthTo2DImageKOnly_(Image<float> &ImgToModify,
                                       const PMDImageMetaData &MetaData2D, 
                                       PMDImageMetaData &MetaDataDepth);
    static int FitDepthTo2DImageHomogr_(Image<float> &ImgToModify,
                                        const PMDImageMetaData &MetaData2D, 
                                        PMDImageMetaData &MetaDataDepth);
    static int FitDepthTo2DImageFree_(Image<float> &ImgToModify,
                                      const Image<float> &ImgDepth, 
                                      const PMDImageMetaData &MetaData2D, 
                                      PMDImageMetaData &MetaDataDepth);

    /** @brief Deletes (sets to 0) pixels in PMD-Image with a local 
     *  variance > sigmaThresh. Eleminates depth noise.
     *  @author grest
     */
    static void DeleteHighVarianceValues_(BIAS::Image<float> &img,
                                          double sigmaThresh);

    /** @brief Create a H-Matrix that converts the depth image to the 2D view
     */
    static HMatrix HFromMeta_(const PMDImageMetaData &MetaData2D, 
                              const PMDImageMetaData &MetaDataDepth);

    
    static uint16 DepthConvert_(float depth);
    static float DepthConvert_(uint16 depth);
    static uint16 ModCoeffConvert_(float modcoeff);
    static float ModCoeffConvert_(uint16 modcoeff);
    static uint16 Convert_(float val, float lower, float upper);
    static float Convert_(uint16 val, float lower, float upper);

    static int PostprocessData_(uint8* data2d, uint16* dataDepth, 
                                uint16* dataModCoeff,                               
                                Image<unsigned char> &Img2D, 
                                Image<float> &ImgDepth, 
                                Image<float> &ImgModCoeff, 
                                PMDImageMetaData &MetaData2D, 
                                PMDImageMetaData &MetaDataDepth);

    static void Grey16to8Bit_(uint16* src, uint8* &dst, unsigned int size);

    static const double CalibrationLUT[][2];
    static bool bLeastSquaresPrepared_;
    //matrix for least squares computation (measured percentages)
    static BIAS::Matrix<double>  A_;
    //vector  with measurement points
    static BIAS::Vector<double>  Y_; 
    //result vector of least squares polynomial fitting
    static BIAS::Vector<double>  X_; 
    // the predefined depth poly values [controlpoint][value]
    static const double DepthCaliPredefinedMeasurements_[][2];

  };


}
#endif
