/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <Image/RectificationViaProjectionMappingBase.hh>

#include <Image/RectificationViaProjectionMappingBase.hh>
#include <Image/ProjectionMapping.hh>
#include <Geometry/Triangulation.hh>
#include <limits>

// enable this for gradient based alignemnt of grey values along epipolar lines
//#define SUBPIXEL_REFINEMENT
#ifdef SUBPIXEL_REFINEMENT
# include <Matcher2D/TrackerBaseWeighted1D.hh>
# include <Base/Image/ImageConvert.hh>
# include <Filter/GradientSimple.hh>
# include <Filter/Gauss.hh>
# include <Base/Image/ImageIO.hh>
#endif

#define PARENT RectificationBase<InputStorageType, OutputStorageType>

using namespace BIAS;
using namespace std;

template <class InputStorageType, class OutputStorageType>
BIAS::RectificationViaProjectionMappingBase<InputStorageType, 
                                            OutputStorageType>::
~RectificationViaProjectionMappingBase()
{
  delete rectPPA_;
  delete rectPPB_;
}


template <class InputStorageType, class OutputStorageType>
BIAS::RectificationViaProjectionMappingBase<InputStorageType, 
                                            OutputStorageType>::
RectificationViaProjectionMappingBase(const double& scale) : 
  rectPPA_(NULL), rectPPB_(NULL), scale_(scale),
  rectificationParamsAreValid_(false),
  interpolationType_(MapTrilinear)
{ 
  secondFill_ = std::numeric_limits<OutputStorageType>::max();
}

template <class InputStorageType, class OutputStorageType>
int BIAS::RectificationViaProjectionMappingBase<InputStorageType, 
                                                OutputStorageType>::
SetCameraA(BIAS::Camera<InputStorageType>& cam)
{
  rectificationParamsAreValid_ = false;
  return PARENT::SetCameraA(cam);
}

template <class InputStorageType, class OutputStorageType>
int BIAS::RectificationViaProjectionMappingBase<InputStorageType, 
                                                OutputStorageType>::
SetCameraA(const BIAS::Image<InputStorageType>& img,
           const BIAS::ProjectionParametersBase* proj)
{
  rectificationParamsAreValid_ = false;
  return PARENT::SetCameraA(img, proj);
}


template <class InputStorageType, class OutputStorageType>
int BIAS::RectificationViaProjectionMappingBase<InputStorageType, 
                                                OutputStorageType>::
SetCameraB(BIAS::Camera<InputStorageType>& cam)
{
  rectificationParamsAreValid_ = false;
  return PARENT::SetCameraB(cam);
}


template <class InputStorageType, class OutputStorageType>
int BIAS::RectificationViaProjectionMappingBase<InputStorageType, 
                                                OutputStorageType>::
 SetCameraB(const BIAS::Image<InputStorageType>& img,
            const BIAS::ProjectionParametersBase* proj)
{
  rectificationParamsAreValid_ = false;
  return PARENT::SetCameraB(img, proj);
}


template <class InputStorageType, class OutputStorageType>
void BIAS::RectificationViaProjectionMappingBase<InputStorageType, 
                                                 OutputStorageType>::
SetInterpolationMethod(InterpolationMethod interpolationType) 
{
  interpolationType_ = interpolationType;
}

template <class InputStorageType, class OutputStorageType>
bool BIAS::RectificationViaProjectionMappingBase<InputStorageType, 
                                                OutputStorageType>::
AreRectificationParametersValid_()
{
  if (!rectificationParamsAreValid_) {
    if (DetermineScaledRectificationParameters_() == 0) {
      rectificationParamsAreValid_ = true;
    }
  }
    
  return rectificationParamsAreValid_;
}

template <class InputStorageType, class OutputStorageType>
int BIAS::RectificationViaProjectionMappingBase<InputStorageType, 
                                                OutputStorageType>::
DetermineScaledRectificationParameters_() {
  int res = DetermineRectificationParameters_();
  if (res==0) {
    BIASASSERT(rectPPA_!=NULL);
    rectPPA_->Rescale(float(1.0/scale_));
    BIASASSERT(rectPPB_!=NULL);
    rectPPB_->Rescale(float(1.0/scale_));
  }
  return res;
}

template <class InputStorageType, class OutputStorageType>
int BIAS::RectificationViaProjectionMappingBase<InputStorageType, 
                                                OutputStorageType>::
GetClonesOfRectificationParameters(ProjectionParametersBase*& rectPPA,
                                   ProjectionParametersBase*& rectPPB)
{
  rectPPA = NULL;
  rectPPB = NULL;

  if (!AreRectificationParametersValid_()) {
    return -1;
  }

  if (rectPPA_ == NULL || rectPPB_ == NULL) {
    return -1;
  }
  
  rectPPA = rectPPA_->Clone();
  rectPPB = rectPPB_->Clone();
  return 0;
}

template <class InputStorageType, class OutputStorageType> 
int BIAS::RectificationViaProjectionMappingBase<InputStorageType, 
                                                OutputStorageType>::
Rectify()
{
  if (!AreRectificationParametersValid_()) {
    return -1;
  }

  unsigned int width, height;

  rectPPA_->GetImageSize(width, height);
  if (!PARENT::rectImageA_.IsEmpty()) PARENT::rectImageA_.Release();
  PARENT::rectImageA_.Init(width, height, 
                                PARENT::imageA_.GetChannelCount());
  RectificationBase<InputStorageType, OutputStorageType>::
    rectImageA_.SetZero();

  rectPPB_->GetImageSize(width, height);
  if (!PARENT::rectImageB_.IsEmpty()) PARENT::rectImageB_.Release();
  PARENT::rectImageB_.Init(width, height, 
                                 PARENT::imageB_.GetChannelCount());
  
  OutputStorageType* fill = new OutputStorageType[PARENT::
                                                  imageB_.GetChannelCount()];
  for (unsigned int i = 0; i < PARENT::imageB_.GetChannelCount(); i++) {
    fill[i] = secondFill_;
  }
  PARENT::rectImageB_.FillImageWithConstValue(*fill);
  delete fill;

  ProjectionMapping<InputStorageType, OutputStorageType> rectMapper;
  
  rectMapper.SetSourceCam(Projection(*PARENT::ppBA_));
  rectMapper.SetSinkCam(Projection(*rectPPA_));
  if (rectMapper.Map(PARENT::imageA_, 
                     PARENT::rectImageA_, 
                     interpolationType_) < 0) {
    BIASERR("rectification of image A failed");
    return -1;
  }  

  rectMapper.SetSourceCam(Projection(*PARENT::ppBB_));
  rectMapper.SetSinkCam(Projection(*rectPPB_));
  if (rectMapper.Map(PARENT::imageB_, 
                     PARENT::rectImageB_, 
                     interpolationType_) < 0) {
    BIASERR("rectification of image B failed");
    return -1;
  }  

  return 0;
}



template <class InputStorageType, class OutputStorageType>
int BIAS::RectificationViaProjectionMappingBase<InputStorageType, 
                                                OutputStorageType>:: 
Disp2Depth(const Image<float>& DisparityMap, 
           Image<float>& DepthMap,
           unsigned int border)
{
  if (!AreRectificationParametersValid_()) {
    return -1;
  }

  unsigned int width, height;
  PARENT::ppBA_->GetImageSize(width, height);
    
  if(!DepthMap.IsEmpty()) DepthMap.Release();
  DepthMap.Init(width, height);
  DepthMap.SetZero();
  
  float** DepthMapPel = DepthMap.GetImageDataArray();
  //  float** DispMapPel = DisparityMap.GetImageDataArray();
  //  unsigned int dispMapWidth = DisparityMap.GetWidth();
  //  unsigned int dispMapHeight = DisparityMap.GetHeight();

  float disparityThreshold = 0.0001f + (float)DisparityMap.GetMinPixelValue(); 

  //allows acces from Orig A to Rect A via GetSourceCoords
  ProjectionMapping<InputStorageType, OutputStorageType> relativeOrigAToRectA;
  //allows acces from Rect B to Orig B via GetSourceCoords
  ProjectionMapping<InputStorageType, OutputStorageType> relativeRectBToOrigB;

  relativeOrigAToRectA.SetSinkCam(Projection(*PARENT::ppBA_));
  relativeOrigAToRectA.SetSourceCam(Projection(*rectPPA_));
  relativeRectBToOrigB.SetSinkCam(Projection(*rectPPB_));
  relativeRectBToOrigB.SetSourceCam(Projection(*PARENT::ppBB_));

#ifdef SUBPIXEL_REFINEMENT

  // want to track in float to avoid messing around with templates
  // compute blurred texture and gradient images in float:
  Gauss<OutputStorageType, float> Blurrer;
  Blurrer.SetSigma(0.1);
  Image<float> rectAf, gradax, graday;
  Blurrer.Filter(this->rectImageA_, rectAf);
  Image<float> rectBf, gradbx, gradby;
  Blurrer.Filter(this->rectImageB_, rectBf);
#ifdef BIAS_DEBUG_KK
  ImageIO::Write("rectAf.mip", rectAf);
  ImageIO::Write("rectBf.mip", rectBf);
#endif
  // assume that images are already blurred
  GradientSimple<float, float> MyGradient;
  MyGradient.Filter(rectAf, gradax, graday);
  MyGradient.Filter(rectBf, gradbx, gradby);
#ifdef BIAS_DEBUG_KK
  ImageIO::Write("gradax.mip", gradax);
  ImageIO::Write("gradbx.mip", gradbx);
#endif

  // generate tracker which searches only for dx
  TrackerBaseWeighted1D<float> SubPixelRefiner;
  //SubPixelRefiner.SetDebugLevel(D_TRACKERB_KLT);
  SubPixelRefiner.SetHalfWinSize(7);
  SubPixelRefiner.SetMaxIterations(25);
  SubPixelRefiner.SetMaxError(0.05);
  SubPixelRefiner.SetMaxResiduumMAD(1e10);
  SubPixelRefiner.SetRejectionType(RT_MAD);
  SubPixelRefiner.SetAffineBrightnessInvariance(true);
  SubPixelRefiner.Init(rectAf, rectBf, gradax, graday, gradbx, gradby);

  // some variables which we need but which we wont evaluate ...
  KLT_TYPE kltsteperror = 0;
  int kltiterations = 0;
  KLT_TYPE residuumMAD = 0;
  KLT_TYPE residuumMSD = 0;
  Matrix<KLT_TYPE> cov(2,2,MatrixZero);
#endif
                      

  HomgPoint2D depthMapPoint, dispMapPoint;
  HomgPoint2D correspondenceInRectB, correspondenceInOrigB;
  HomgPoint3D point3D;
  
  Triangulation triangulator;
  double disparity = 0;

  depthMapPoint[2] = 1.0;
  dispMapPoint[2] = 1.0;  
  for(unsigned int y=border; y<height-border; y++) {
    for(unsigned int x=border; x<width-border; x++) {

      depthMapPoint[0] = x;
      depthMapPoint[1] = y;
    
      bool valid = true;
      if(relativeOrigAToRectA.GetSourceCoordinates(depthMapPoint, 
                                                    dispMapPoint)==0) {
        //have rect A coords get me the disp map value
        disparity = 
          DisparityMap.BilinearInterpolationGreyMinVal(dispMapPoint[0],
                                                       dispMapPoint[1],
                                                       disparityThreshold);
        if(disparity>disparityThreshold) { //valid disparity
          correspondenceInRectB.Set(dispMapPoint[0]-disparity,
                                    dispMapPoint[1], 1.0);
// dont let the refiner compute an offset larger than this
#define MAX_OFFSET_REFINEMENT 0.6
#ifdef SUBPIXEL_REFINEMENT
          HomgPoint2D refinedPoint = correspondenceInRectB;

          // track only those points which are triangulated






          // Instead of using a constant gaussian weighted window for every
          // correspondence, it would be much more intelligent to use
          // something like 1/(0.5+fabs(disparity difference)) as a weight for 
          // each pixel in the neighborhood. This automatically decreases
          // the window at non-fronto-parallel structures or depth 
          // discontinuities.
          // simply add a SetWeight method to Tracker1D






          int refinementresult = 
            SubPixelRefiner.Track(dispMapPoint, correspondenceInRectB,
                                  refinedPoint,
                                  kltsteperror, kltiterations,residuumMAD,
                                  residuumMSD, cov);

          // did the refinement succeed at all ?
          if (refinementresult==0) {
            // fine, check if the step is consistent with our guess or
            // whether it is too large (violates smoothness in neighborhood)
            double diff = refinedPoint[0]-correspondenceInRectB[0];
            // make sure y does not change
            BIASASSERT(Equal(refinedPoint[1], correspondenceInRectB[1]));
            if (fabs(diff)<MAX_OFFSET_REFINEMENT) {
              // fine! set this as the new correspondence for the point
              correspondenceInRectB = refinedPoint;
              //cout<<"+("<<diff<<") ";
            } else {
              // step to big, presumably other local minimum
              //cout<<"-("<<diff<<") ";
            }
          } else {
            // klt complained, presumably no structure or too many iterations
            //cout<<"!"<<refinementresult<<"!";
          }
#endif

          if(relativeRectBToOrigB.
             GetSourceCoordinates(correspondenceInRectB, 
                                  correspondenceInOrigB)==0){

            //have got correspondance in orig B!
            if(triangulator.Triangulate(PARENT::ppBA_, 
                                        PARENT::ppBB_,
                                        depthMapPoint, correspondenceInOrigB, 
                                        point3D)==0) {
              //ok triangulation was good calc depth
              Vector3<double> euc;
              point3D.GetEuclidian(euc);
              Vector3<double> diff = euc - PARENT::ppBA_->GetC();

              BIASASSERT(diff.NormL2()>1e-5);
              DepthMapPel[y][x] = (float)diff.NormL2();
              valid = true;
            }
          } //no correspondance in orig B :(
          else valid = false;
        }//disprity is too small:(
        else valid = false;
        
      } //end of check if rect coord available
      else valid = false;
      
      if(!valid) //ok did we get valid depth ?
        DepthMapPel[y][x] = UNDEFINED_DEPTH;

    }//end of for x 
#ifdef SUBPIXEL_REFINEMENT
    // a new line is completed
    cout<<"S"<<flush;
#endif
  }//end of for y

  return 0;  
}


template class RectificationViaProjectionMappingBase<unsigned char, unsigned char>;
template class RectificationViaProjectionMappingBase<float, float>;
template class RectificationViaProjectionMappingBase<unsigned char, float>;
template class RectificationViaProjectionMappingBase<float, unsigned char>;

#if defined(BUILD_IMAGE_CHAR)
template class RectificationViaProjectionMappingBase<unsigned char, char>;
template class RectificationViaProjectionMappingBase<char, char>;
#endif

#if defined(BUILD_IMAGE_USHORT)
template class RectificationViaProjectionMappingBase<unsigned short,unsigned short>;
#endif

#if defined(BUILD_IMAGE_SHORT)
template class RectificationViaProjectionMappingBase<short, short>;
#endif

#if defined(BUILD_IMAGE_SHORT)&&defined(BUILD_IMAGE_USHORT)
template class RectificationViaProjectionMappingBase<unsigned short, short>;
#endif

#if defined(BUILD_IMAGE_INT)
template class RectificationViaProjectionMappingBase<int,int>;
#endif

#if defined(BUILD_IMAGE_USHORT)
template class RectificationViaProjectionMappingBase<unsigned short,float>;
#endif

#if defined(BUILD_IMAGE_USHORT) && defined(BUILD_IMAGE_INT)
template class RectificationViaProjectionMappingBase<unsigned short,int>;
#endif

#if defined(BUILD_IMAGE_DOUBLE)
template class RectificationViaProjectionMappingBase<double,double>;
#endif
