/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <Base/Common/BIASpragmaStart.hh>

#ifndef __RectificationViaProjectionMappingBase_hh__
#define __RectificationViaProjectionMappingBase_hh__

#include <Image/RectificationBase.hh>
#include <Image/BackwardMapping.hh>

namespace BIAS {
  
  /** Base class for rectification implementations that make use of
   *  projections to represent rectified state and where the projection-centers
   *  in rectified and original state are the same.
   *  Implementation of child classes only have to supply the rectified params
   *  calculation. Aftwerwards they can make use of the ProjectionMapping 
   *  facilities that are implemented in this class to rectify images and
   *  evaluate disparity maps into depth maps. Be warned: ProjectionMapping
   *  is a very expansive implementation, yet takles nonlinearities in 
   *  coordinate transformation (i.e. distortions).
   * 
   *  \author bartczak 11/2006
   */
  template <class InputStorageType, class OutputStorageType> 
  class BIASImage_EXPORT RectificationViaProjectionMappingBase 
    : public RectificationBase<InputStorageType, OutputStorageType>
  {
  public:
    
    virtual ~RectificationViaProjectionMappingBase();
    RectificationViaProjectionMappingBase(const double& Scale = 1.0);
    inline void SetScale(const double& scale) { scale_ = scale; }

    virtual int SetCameraA(BIAS::Camera<InputStorageType>& cam);
    virtual int SetCameraA(const BIAS::Image<InputStorageType>& img,
                           const BIAS::ProjectionParametersBase* proj);
    virtual int SetCameraB(BIAS::Camera<InputStorageType>& cam);
    virtual int SetCameraB(const BIAS::Image<InputStorageType>& img,
                           const BIAS::ProjectionParametersBase* proj);

    virtual int Rectify();
    virtual int Disp2Depth(const Image<float>& DisparityMap, 
                           Image<float>& DepthMap,
                           unsigned int border);
    
    /** Tries to calculate the rectification parameters and returns clones
     *  of internal members. If just one rectification parameter could not
     *  be determined the returned pointer will be NULL and the return value is
     *  less 0.
     */
    virtual int 
    GetClonesOfRectificationParameters(ProjectionParametersBase*& rectPPA,
                                       ProjectionParametersBase*& rectPPB);

    /** Determines the type of interpolation used when rectifying.
     *  Types are taken from BackwardMapping.hh
     */
    void SetInterpolationMethod(InterpolationMethod interpolationType);
    

    /** fill color for second image (first is black) */
    void SetSecondFill(OutputStorageType d) {
      secondFill_ = d;
    }

  protected:

    /** Calculating the projection parameters for specific rectification.
     * Is called by Rectify implementation. Must return 0 upon success, every
     * other code is handled as failure.
     */
    virtual int DetermineRectificationParameters_() = 0;

    /** this computes ideal parameters and applies the scale */
    int DetermineScaledRectificationParameters_();

    /** Projection parameters of the rectified image A */
    ProjectionParametersBase* rectPPA_;

    /** Projection parameters of the rectified image B */
    ProjectionParametersBase* rectPPB_;

    /// scale computed rectification resolution by this amount 
    double scale_;
  private:
        /** Checks for rectificationParamsAreValid_ if not true will
     *  try to succsessfully process DetermineRectificationParameters_().
     *  If this is the case it will set rectificationParamsAreValid_ to true
     *  and return this value. Otherwise will return false.
     */
    bool AreRectificationParametersValid_();

    /** Flag is true when DetermineRectificationParameters_() was succesfull
     * and no setter was called afterwards.
     */
    bool rectificationParamsAreValid_;

    InterpolationMethod interpolationType_;
    OutputStorageType secondFill_;
  };
    
}

#endif

#include <Base/Common/BIASpragmaEnd.hh>
