/* 
   This file is part of the BIAS library (Basic ImageAlgorithmS).

   Copyright (C) 2003, 2004    (see file CONTACTS for details)
   Multimediale Systeme der Informationsverarbeitung
   Institut fuer Informatik
   Christian-Albrechts-Universitaet Kiel
   
   
   BIAS is free software; you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation; either version 2.1 of the License, or
   (at your option) any later version.
   
   BIAS is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with BIAS; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <Image/SphericalRectification.hh>
#include <Geometry/ProjectionParametersGreatCircles.hh>

using namespace BIAS;

#define SUPERPARENT RectificationBase<InputStorageType, OutputStorageType>
#define PARENT RectificationViaProjectionMappingBase<InputStorageType, OutputStorageType>


template <class InputStorageType, class OutputStorageType>
BIAS::SphericalRectification<InputStorageType, OutputStorageType>::
SphericalRectification()
{ }

template <class InputStorageType, class OutputStorageType>
BIAS::SphericalRectification<InputStorageType, OutputStorageType>::
~SphericalRectification() 
{ }


template <class InputStorageType, class OutputStorageType>
bool BIAS::SphericalRectification<InputStorageType, OutputStorageType>::
IsInputCameraValid(const BIAS::Image<InputStorageType>& img,
                   const BIAS::ProjectionParametersBase* proj)
{
  const ProjectionParametersSpherical* pps = 
    dynamic_cast<const ProjectionParametersSpherical*>(proj);
  
  if(pps == NULL)
    return false;

  return true;
}

template <class InputStorageType, class OutputStorageType>
int BIAS::SphericalRectification<InputStorageType, OutputStorageType>::
DetermineRectificationParameters_()
{

  Pose poseA = SUPERPARENT::ppBA_->GetPose();
  Pose poseB = SUPERPARENT::ppBB_->GetPose();
  Pose rectPoseA, rectPoseB;
  if(SUPERPARENT::CalculateRectifiedBases(poseA, poseB, 
                                          rectPoseA, rectPoseB, false)!=0)
    return -1;

  delete PARENT::rectPPA_;
  delete PARENT::rectPPB_;
  ProjectionParametersGreatCircles* ppgcA;
  ProjectionParametersGreatCircles* ppgcB;
  PARENT::rectPPA_ = ppgcA = new ProjectionParametersGreatCircles();
  PARENT::rectPPB_ = ppgcB = new ProjectionParametersGreatCircles();
  
  PARENT::rectPPA_->SetPose(rectPoseA);
  PARENT::rectPPB_->SetPose(rectPoseB);

  const ProjectionParametersSpherical* ppsA = 
    dynamic_cast<const ProjectionParametersSpherical*>(SUPERPARENT::ppBA_);
  const ProjectionParametersSpherical* ppsB = 
    dynamic_cast<const ProjectionParametersSpherical*>(SUPERPARENT::ppBB_);
  BIASASSERT(ppsA != NULL && ppsB != NULL);

  //  unsigned int size = 0;
  double maxRadius = ppsA->GetRadius();
  if(ppsB->GetRadius()>maxRadius) 
    maxRadius = ppsB->GetRadius();
  //  size =(unsigned int)ceil(maxRadius*2.0*M_PI);
  double numSamplesOnEpiLine = floor(M_PI*maxRadius*2.0);
  double numEpiLines = floor(2.0*M_PI*maxRadius);  

//   PARENT::rectPPA_->SetImageSize(size, size);
//   PARENT::rectPPB_->SetImageSize(size, size);
  PARENT::rectPPA_->SetImageSize((unsigned int)numSamplesOnEpiLine, 
                                 (unsigned int)numEpiLines);
  PARENT::rectPPB_->SetImageSize((unsigned int)numSamplesOnEpiLine, 
                                 (unsigned int)numEpiLines);


  //determine angle offset
  HomgPoint3D z(0,0,1,1);
  HomgPoint3D worldOpticalAxisA = poseA.LocalToGlobal(z);

  //  if(camAngleAlpha>M_PI_2) camAngleAlpha = M_PI_2;
  double minAlphaA, minBetaA, rangeAlphaA, rangeBetaA;
  ppgcA -> GetCenteredAngleInterval(worldOpticalAxisA, 
                                    ppsA->GetMaxCamAngle(), 
                                    minAlphaA, minBetaA, 
                                    rangeAlphaA, rangeBetaA);
  
  HomgPoint3D worldOpticalAxisB = poseB.LocalToGlobal(z);  
  double minAlphaB, minBetaB, rangeAlphaB, rangeBetaB;
  ppgcB -> GetCenteredAngleInterval(worldOpticalAxisB, 
                                    ppsB->GetMaxCamAngle(), 
                                    minAlphaB, minBetaB, 
                                    rangeAlphaB, rangeBetaB);

  double commonAlphaMin, commonBetaMin, commonAlphaRange, commonBetaRange;
  commonAlphaMin = (minAlphaA<minAlphaB) ? minAlphaA : minAlphaB;
  commonBetaMin = (minBetaA<minBetaB) ? minBetaA : minBetaB;

  commonAlphaRange = (rangeAlphaA>rangeAlphaB) ? rangeAlphaA : rangeAlphaB;
  commonBetaRange = (rangeBetaA>rangeBetaB) ? rangeBetaA : rangeBetaB;


  
  ppgcA->SetMinAngles(commonAlphaMin, commonBetaMin);
  ppgcA->SetAngleRanges(commonAlphaRange, commonBetaRange);

  ppgcB->SetMinAngles(commonAlphaMin, commonBetaMin);
  ppgcB->SetAngleRanges(commonAlphaRange, commonBetaRange);

  return 0;
}

// template <class InputStorageType, class OutputStorageType>
// int BIAS::SphericalRectification<InputStorageType, OutputStorageType>::
// DetermineGreatCircleParametersFromSpherical_(
//                                   ProjectionParametersSpherical* ppsIn,
//                                   Pose& rectPose,
//                                   double& minAlpha,
//                                   double& minBeta,
//                                   double& alphaRange,
//                                   double& betaRange,
//                                   double& relativeAlphaRes,
//                                   double& relativeBetaRes) 
// {
  
// }


template class SphericalRectification<unsigned char, unsigned char>;
template class SphericalRectification<float, float>;
template class SphericalRectification<unsigned char, float>;
template class SphericalRectification<float, unsigned char>;

#if defined(BUILD_IMAGE_CHAR)
template class SphericalRectification<unsigned char, char>;
template class SphericalRectification<char, char>;
#endif

#if defined(BUILD_IMAGE_USHORT)
template class SphericalRectification<unsigned short,unsigned short>;
#endif

#if defined(BUILD_IMAGE_SHORT)
template class SphericalRectification<short, short>;
#endif

#if defined(BUILD_IMAGE_SHORT)&&defined(BUILD_IMAGE_USHORT)
template class SphericalRectification<unsigned short, short>;
#endif

#if defined(BUILD_IMAGE_INT)
template class SphericalRectification<int,int>;
#endif

#if defined(BUILD_IMAGE_USHORT)
template class SphericalRectification<unsigned short,float>;
#endif

#if defined(BUILD_IMAGE_USHORT) && defined(BUILD_IMAGE_INT)
template class SphericalRectification<unsigned short,int>;
#endif

#if defined(BUILD_IMAGE_DOUBLE)
template class SphericalRectification<double,double>;
#endif
