/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "RegionMatcher.hh"

#include <float.h>

namespace BIAS
{
using namespace std;

template <class StorageType>
void RegionMatcher::SAD(unsigned int p1[2], unsigned int p2[2], 
                        StorageType **ida1, StorageType **ida2, 
                        unsigned int halfwinsize, double& result)
{
  int nhalfwinsize = - int(halfwinsize);
  result=0.0;
  register StorageType st1, st2;
  for (register int c=nhalfwinsize; c<=(int)halfwinsize; c++){
    for (register int r=nhalfwinsize; r<=(int)halfwinsize; r++){
      st1=ida1[p1[1]+r][p1[0]+c];
      st2=ida2[p2[1]+r][p2[0]+c];
      result+=(double)((st1>=st2) ? st1-st2 : st2-st1);
    }
  }
}

template <> BIASMatcher2D_EXPORT 
void RegionMatcher::SAD<unsigned char>(unsigned int p1[2], unsigned int p2[2], 
				       unsigned char **ida1, 
				       unsigned char **ida2, 
				       unsigned int halfwinsize, 
				       double& result)
{
  int nhalfwinsize = - int(halfwinsize);
  register long unsigned int res=0;
  register unsigned char c1, c2;
  for (register int c=nhalfwinsize; c<=(int)halfwinsize; c++){
    for (register int r=nhalfwinsize; r<=(int)halfwinsize; r++){
      c1=ida1[p1[1]+r][p1[0]+c];
      c2=ida2[p2[1]+r][p2[0]+c];
      res+=(c1>=c2)? c1-c2 : c2-c1;
    }
  }
  result=(double)res;
}

template <class StorageType>
void RegionMatcher::SADN(unsigned int p1[2], unsigned int p2[2], 
                        StorageType **ida1, StorageType **ida2, 
                        unsigned int halfwinsize, double& result)
{
  int nhalfwinsize = - int(halfwinsize);
  result=0.0;
  register StorageType st1, st2;
  for (register int c=nhalfwinsize; c<=(int)halfwinsize; c++){
    for (register int r=nhalfwinsize; r<=(int)halfwinsize; r++){
      st1=ida1[p1[1]+r][p1[0]+c];
      st2=ida2[p2[1]+r][p2[0]+c];
      result+=(st1>=st2) ? st1-st2 : st2-st1;
    }
  }
  // STORAGETYPE_MAX should be the biggest poistive number that can be 
  // expressed with this storage type 
#define STORAGETYPE_MAX 256.0  

  unsigned tmp=(halfwinsize<<1)+1;
  result/=double(tmp*tmp*STORAGETYPE_MAX);
  result=1.0-result;
}

template <> BIASMatcher2D_EXPORT 
void RegionMatcher::SADN<unsigned char>(unsigned int p1[2], unsigned int p2[2], 
				       unsigned char **ida1, 
				       unsigned char **ida2, 
				       unsigned int halfwinsize, 
				       double& result)
{
  int nhalfwinsize = - int(halfwinsize);
  register long unsigned int res=0;
  register unsigned char c1, c2;
  for (register int c=nhalfwinsize; c<=(int)halfwinsize; c++){
    for (register int r=nhalfwinsize; r<=(int)halfwinsize; r++){
      c1=ida1[p1[1]+r][p1[0]+c];
      c2=ida2[p2[1]+r][p2[0]+c];
      res+=(c1>=c2)? c1-c2 : c2-c1;
    }
  }
  // result = 1.0 - res /(UCHAR_MAX*windowarea) = 1 - res / tmp
  // tmp = sqrt(UCHAR_MAX)*(2*halfwinsize+1) ca. 16 * (2 * halfwinsize + 1)
  //     = 32 * halfwinsize + 16
  unsigned tmp=(halfwinsize<<5)+16;
  result=1.0-(double)res/double(tmp*tmp);
}


template <class StorageType>
void RegionMatcher::SSD(unsigned int p1[2], unsigned int p2[2], 
                        StorageType **ida1, StorageType **ida2, 
                        unsigned int halfwinsize, double& result)
{
  int nhalfwinsize = - int(halfwinsize);
  result=0.0;
  for (register int c=nhalfwinsize; c<=(int)halfwinsize; c++){
    for (register int r=nhalfwinsize; r<=(int)halfwinsize; r++){
      result+=pow((double)(ida1[p1[1]+r][p1[0]+c])-
                  (double)(ida2[p2[1]+r][p2[0]+c]), 2.0);
    }
  }
}

template <> BIASMatcher2D_EXPORT 
void RegionMatcher::SSD<unsigned char>(unsigned int p1[2], unsigned int p2[2], 
				       unsigned char **ida1, 
				       unsigned char **ida2, 
				       unsigned int halfwinsize, 
				       double& result)
{
  int nhalfwinsize = - int(halfwinsize);
  register long unsigned int res=0;
  register short diff;
  for (register int c=nhalfwinsize; c<=(int)halfwinsize; c++){
    for (register int r=nhalfwinsize; r<=(int)halfwinsize; r++){
      diff=(short)(ida1[p1[1]+r][p2[0]+c])-(short)(ida2[p2[1]+r][p2[0]+c]);
      res+=(unsigned int)(diff*diff);
    }
  }
  result=(double)res;
}

template <class StorageType>
void RegionMatcher::NCC(unsigned int p1[2], unsigned int p2[2], 
                        StorageType **ida1, StorageType **ida2, 
                        unsigned int halfwinsize, double& result)
{
  int nhalfwinsize = - int(halfwinsize);
  double im1=0, im2=0, S1=0, S2=0;
  double S11=0, S12=0, S22=0, N=0;
  BIASASSERT(ida1!=NULL);
  BIASASSERT(ida2!=NULL);
  for (register int r=nhalfwinsize; r<=(int)halfwinsize; r++)
    for (register int c=nhalfwinsize; c<=(int)halfwinsize; c++){
      im1 = (double)(ida1[p1[1]+r][p1[0]+c]);
      im2 = (double)(ida2[p2[1]+r][p2[0]+c]);
      N ++; S1 += im1; S2 += im2;
      S11 += im1*im1; S12 += im1*im2; S22 += im2*im2;
    }
  result =(N*S12 - S1*S2) / sqrt((N*S11 - S1*S1) * (N*S22 - S2*S2));
  if (result!=result) result=0.0;
}

// faster specialisation because of integer arithmetic
// does work without overflow with halfwindowsizes <= 127
// if sizeof(unsigned char)=1 and sizeof(unsigned long int)=4
template <> BIASMatcher2D_EXPORT void 
RegionMatcher::NCC<unsigned char>(unsigned int p1[2], unsigned int p2[2], 
                                  unsigned char **ida1, unsigned char **ida2, 
                                  unsigned int halfwinsize, double& result)
{
  int nhalfwinsize = - int(halfwinsize);
  register int im1=0, im2=0;
  register unsigned long int S11=0, S12=0, S22=0, N=0, S1=0, S2=0;
  for (register int r=nhalfwinsize; r<=(int)halfwinsize; r++)
    for (register int c=nhalfwinsize; c<=(int)halfwinsize; c++){
      im1 = (int)(ida1[p1[1]+r][p1[0]+c]);
      im2 = (int)(ida2[p2[1]+r][p2[0]+c]);
//        if (im1!=im2) 
//          cerr << "im1 ("<<p1[0]+c<<", "<<p1[1]+r<<") : "<<im1
//               << "\tim2 ("<<p2[0]+c<<", "<<p2[1]+r<<") : "<<im2<<endl;
      N += 1; S1 += im1; S2 += im2;
      S11 += im1*im1; S12 += im1*im2; S22 += im2*im2;
    }
//     cerr << " N " << N << "   S1 " << S1 << "   S11 " << S11 << "   S2 " << S2 
//          << "   S22 " << S22 << "   S12 " << S12 << endl;
  result =((double)N*(double)S12 - (double)S1*(double)S2)/
    sqrt(((double)N*(double)S11 - (double)S1*(double)S1) * 
         ((double)N*(double)S22 - (double)S2*(double)S2));
  if (result!=result) result=0.0;

//   cout<<result<<" at ("<<p1[0]<<", "<<p1[1]<<") & ("
//       <<p2[0]<<", "<<p2[1]<<")"<<endl;
}

template <class StorageType>
void RegionMatcher::NCC(double p1[2], double p2[2], StorageType **ida1, 
                        StorageType **ida2, unsigned hws, double& result)
{
  double im1=0, im2=0, S1=0, S2=0;
  double S11=0, S12=0, S22=0, N=0;
  if (_ncchws!=hws){
    _ncchws=hws;
    _nccwinsize=1+(hws<<1);
    _iim1.newsize(_nccwinsize, _nccwinsize);
    _iim2.newsize(_nccwinsize, _nccwinsize);
  }
  _BilinearRegion(ida1, p1[0], p1[1], _ncchws, _iim1);
  _BilinearRegion(ida2, p2[0], p2[1], _ncchws, _iim2);
  for (register unsigned r=0; r<=hws; r++)
    for (register unsigned c=0; c<=hws; c++){
      im1 = (double)_iim1[r][c];
      im2 = (double)_iim2[r][c];
      N ++; S1 += im1; S2 += im2;
      S11 += im1*im1; S12 += im1*im2; S22 += im2*im2;
    }
  result =(N*S12 - S1*S2) / sqrt((N*S11 - S1*S1) * (N*S22 - S2*S2));
  if (result!=result) result=0.0;
}


template <class StorageType>
void RegionMatcher::NCC(unsigned p1[2], unsigned p2[2], 
                        StorageType **ida1, StorageType **roi1, 
                        StorageType **ida2, StorageType **roi2,
                        unsigned hww, unsigned hwh, double& result)
{
  int nhww = - int(hww);
  int nhwh = - int(hwh);
  double im1=0, im2=0, S1=0, S2=0;
  double S11=0, S12=0, S22=0, N=0;
  for (register int r=nhwh; r<(int)hwh; r++){
    for (register int c=nhww; c<(int)hww; c++){
      if (roi1[p1[1]+r][p1[0]+c]!=0 && roi2[p2[1]+r][p2[0]+c]!=0){
        im1 = (double)(ida1[p1[1]+r][p1[0]+c]);
        im2 = (double)(ida2[p2[1]+r][p2[0]+c]);
        N ++; S1 += im1; S2 += im2;
        S11 += im1*im1; S12 += im1*im2; S22 += im2*im2;
      }
    }
  }
  result =(N*S12 - S1*S2) / sqrt((N*S11 - S1*S1) * (N*S22 - S2*S2));
  if (result!=result) result=0.0;
}


template <class StorageType>
void RegionMatcher::ColorCCV(unsigned int p1[2], unsigned int p2[2], 
                             StorageType **ida1, StorageType **ida2, 
                             unsigned int halfwinsize, double &result)
{
  double Sum1=0, Sh1=0, Ss1=0;
  double Sum2=0, Sh2=0, Ss2=0;  
  double S12=0, N=0;
  StorageType *H1,*S1, *H2,*S2;  
  double I1,I2;  
  double angle,h1,s1, h2,s2;  
  for (register int r=-halfwinsize; r<=(int)halfwinsize; r++)
    for (register int c=-halfwinsize; c<=(int)halfwinsize; c++){
      H1 = (ida1[(p1[1]+r)] + ((p1[0]+c)*3));
      S1 = H1+1; I1 = *(S1+1);
      H2 = (ida2[(p2[1]+r)] + ((p2[0]+c)*3));
      S2 = H2+1; I2 = *(S2+1);
      BIASDOUT(D_REGION_MATCHER_COLORCCV,
               "HSI1:"<<(int)(*H1)<<" "<<(int)(*S1)<<" "<<(int)I1<<
               "  HSI2:"<<(int)(*H2)<<" "<<(int)(*S2)<<" "<<(int)I2);
      // convert H and S to euclidean coords, H=0 is red
      angle = M_PI * (float)(*H1) / 127.5;      
      h1 = cos(angle)* (*S1);      s1 = sin(angle)* (*S1);      
      angle = M_PI * (float)(*H2) / 127.5;      
      h2 = cos(angle)* (*S2);      s2 = sin(angle)* (*S2);      
        BIASDOUT(D_REGION_MATCHER_COLORCCV,
                 "[h1,s1]:["<<(int)h1<<","<<(int)s1<<
                 "] [h2,s2]:["<<(int)h2<<","<<(int)s2);
      angle =  h1*h2 + s1*s2;
      if (angle<0) angle=0;
      S12 += angle + I1 * I2;
      Sh1 += h1; Ss1 += s1; Sum1 += I1;      
      Sh2 += h2; Ss2 += s2; Sum2 += I2;      
      N ++;
    }
  double S1S2= Sh1 * Sh2 + Ss1 * Ss2;
  if (S1S2 < 0 ) S1S2=0;
  S1S2 += Sum1*Sum2;  
  BIASDOUT(D_REGION_MATCHER_COLORCCV,
           "S12:"<<S12<<" Sh1,Ss1:"<<Sh1<<","<<Ss1<<"  Sum1:"<<Sum1<<
           " Sh2,Ss2:"<<Sh2<<","<<Ss2<<"  Sum2:"<<Sum2);
  result=(S12 - S1S2/N);
}



template <class StorageType>
void RegionMatcher::ColorNCC(unsigned int p1[2], unsigned int p2[2], 
			       StorageType **ida1, StorageType **ida2, 
			       unsigned int halfwinsize, double &result)
{
  int nhalfwinsize = - int(halfwinsize);
#define CNCC_EPSILON 1E-08
  double Sum1=0, Sh1=0, Ss1=0;
  double Sum2=0, Sh2=0, Ss2=0;  
  double S11=0, S12=0, S22=0, N=0;
  StorageType *H1,*S1, *H2,*S2;  
  double I1,I2;  
  double angle,h1,s1, h2,s2;  
  for (register int r=nhalfwinsize; r<=(int)halfwinsize; r++)
    for (register int c=nhalfwinsize; c<=(int)halfwinsize; c++){
      H1 = (ida1[(p1[1]+r)] + ((p1[0]+c)*3));
      S1 = H1+1; I1 = *(S1+1);
      H2 = (ida2[(p2[1]+r)] + ((p2[0]+c)*3));
      S2 = H2+1; I2 = *(S2+1);
      BIASDOUT(D_REGION_MATCHER_COLORNCC,
               "HSI1:"<<(int)(*H1)<<" "<<(int)(*S1)<<" "<<(int)I1<<
               "  HSI2:"<<(int)(*H2)<<" "<<(int)(*S2)<<" "<<(int)I2);
      // convert H and S to euclidean coords, H=0 is red
      angle = M_PI * (float)(*H1) / 127.5;      
      h1 = cos(angle)* (float)(*S1);
      s1 = sin(angle)* (float)(*S1);      
      angle = M_PI * (float)(*H2) / 127.5;      
      h2 = cos(angle)* (float)(*S2);
      s2 = sin(angle)* (float)(*S2);      
      BIASDOUT(D_REGION_MATCHER_COLORNCC,
               "[h1,s1]:["<<(int)h1<<","<<(int)s1<<
               "] [h2,s2]:["<<(int)h2<<","<<(int)s2);
      angle =  h1*h2 + s1*s2; // scalar product
   //    angle = fabs(double(*H1)-double(*H2));
//       if (angle>180)
//         angle=255-angle;
//       angle =  cos(angle/255.0) * double(*S1)*double(*S2); // scalar product
      if (angle<0) angle=0;
      S12 += angle + I1 * I2;
      Sh1 += h1; Ss1 += s1; Sum1 += I1;      
      Sh2 += h2; Ss2 += s2; Sum2 += I2;      
      S11 += h1*h1 + s1*s1 + I1 * I1;
      S22 += h2*h2 + s2*s2 + I2 * I2;	
      N ++;
    }


  double S1S2= Sh1 * Sh2 + Ss1 * Ss2;
  if (S1S2 < 0 ) S1S2=0;
  S1S2 = Sum1*Sum2;  
  double S1S1= Sum1*Sum1;
  double S2S2= Sum2*Sum2;
  BIASDOUT(D_REGION_MATCHER_COLORNCC,
           "N*S12:"<<N*S12<<"  S1S2:"<<S1S2<<endl<<
           " Sh1,Ss1:"<<Sh1<<","<<Ss1<<"  Sum1:"<<Sum1<<
           " Sh2,Ss2:"<<Sh2<<","<<Ss2<<"  Sum2:"<<Sum2);
  BIASDOUT(D_REGION_MATCHER_COLORNCC,"-----------------------------");
  BIASDOUT(D_REGION_MATCHER_COLORNCC,
           "N*S11:"<<N*S11<<" S1S1:"<< S1S1<<" N*S22:"<<N*S22<<" S2S2:"<<S2S2);
  double n1= (N*S11 - S1S1);
  double n2= (N*S22 - S2S2);
  BIASDOUT(D_REGION_MATCHER_COLORNCC,"n1:" <<n1<< " n2:"<<n2);
  if ((fabs(n1)<=CNCC_EPSILON) || (fabs(n2)<=CNCC_EPSILON))
    result=0.0;
  else result=(N*S12 - S1S2)/sqrt(n1*n2);
}


template <class StorageType>
void RegionMatcher::CNCC(unsigned int p1[2], unsigned int p2[2], 
                         StorageType **ida1, StorageType **ida2, 
                         unsigned int halfwinsize, double &result)
{
#define CNCC_EPSILON 1E-08
  int nhalfwinsize = - int(halfwinsize);
  register float Sum1=0, Sh1=0, Ss1=0;
  register float Sum2=0, Sh2=0, Ss2=0;  
  register float S11=0, S12=0, S22=0;
  StorageType *H1, *H2;  
  register float h1,h2,s1,s2,L1,L2;
  register unsigned int N=0;
  register float angle;
  for (register int r=nhalfwinsize; r<=(int)halfwinsize; r++)
    for (register int c=nhalfwinsize; c<=(int)halfwinsize; c++){
      H1 = (ida1[(p1[1]+r)] + ((p1[0]+c)*3));
      s1 = 2.0f*((float)*(H1+1) -127.0f); L1 = (float)*(H1+2);
      h1 = 2.0f*((float)*H1 -127.0f);
      H2 = (ida2[(p2[1]+r)] + ((p2[0]+c)*3));
      s2 = 2.0f*((float)*(H2+1) -127.0f); L2 = (float)*(H2+2);
      h2 = 2.0f*((float)*H2 -127.0f);
          BIASDOUT(D_REGION_MATCHER_COLORNCC,
                      "HSL1:"<<(int)(h1)<<" "<<(int)(s1)<<" "<<(int)L1<<
                      "  HSL2:"<<(int)(h2)<<" "<<(int)(s2)<<" "<<(int)L2);
      // convert H and S to euclidean coords, H=0 is red
      //   angle = M_PI * (float)(*H1) / 127.5;      
      //       h1 = cos(angle)* (float)(*S1);
      //       s1 = sin(angle)* (float)(*S1);      
      //       angle = M_PI * (float)(*H2) / 127.5;      
      //       h2 = cos(angle)* (float)(*S2);
      //       s2 = sin(angle)* (float)(*S2);      
          BIASDOUT(D_REGION_MATCHER_COLORNCC,
                      "[h1,s1]:["<<(int)h1<<","<<(int)s1<<
                      "] [h2,s2]:["<<(int)h2<<","<<(int)s2);

      //      angle =  h1*h2 + s1*s2;
      angle = float((int)(h1) * (int)(h2) + (int)(s1) * (int)(s2));
      if (angle<0) angle=0;
      S12 += angle + L1 * L2;
      Sh1 += h1; Ss1 += s1; Sum1 += L1;      
      Sh2 += h2; Ss2 += s2; Sum2 += L2;      
      S11 += h1*h1 + s1*s1 + L1 * L1;
      S22 += h2*h2 + s2*s2 + L2 * L2;	
      N ++;
    }


  float S1S2= Sum1*Sum2;  
  float S1S1= Sum1*Sum1;
  float S2S2= Sum2*Sum2;
  BIASDOUT(D_REGION_MATCHER_COLORNCC,
           "N*S12:"<<N*S12<<"  S1S2:"<<S1S2<<endl<<
           " Sh1,Ss1:"<<Sh1<<","<<Ss1<<"  Sum1:"<<Sum1<<
           " Sh2,Ss2:"<<Sh2<<","<<Ss2<<"  Sum2:"<<Sum2);
  BIASDOUT(D_REGION_MATCHER_COLORNCC,"-----------------------------");
  BIASDOUT(D_REGION_MATCHER_COLORNCC,
           "N*S11:"<<N*S11<<" S1S1:"<< S1S1<<" N*S22:"<<N*S22<<" S2S2:"<<S2S2);
  float n1= (N*S11 - S1S1);
  float n2= (N*S22 - S2S2);
  BIASDOUT(D_REGION_MATCHER_COLORNCC,"n1:" <<n1<< " n2:"<<n2);
  if ((fabs(n1)<=CNCC_EPSILON) || (fabs(n2)<=CNCC_EPSILON))
    result=0.0;
  else result=(N*S12 - S1S2)/sqrt(n1*n2);
}

}
