/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "TrackerBaseAffine.hh"
#include "MathAlgo/SVD.hh"

using namespace BIAS;
using namespace std;

const double DEFAULT_MAX_AFFINECHANGE = 0.001;

template <class StorageType>
TrackerBaseAffine<StorageType>::
TrackerBaseAffine(bool only4params)
{
  _MaxAffineChange = DEFAULT_MAX_AFFINECHANGE;
  _SimilarityTransformOnly           = only4params;
}

template <class StorageType>
int TrackerBaseAffine<StorageType>::
Track_(Vector2<KLT_TYPE>& p1, Vector2<KLT_TYPE>& p2,
       Vector2<KLT_TYPE>& result, KLT_TYPE& error, 
       int &iter,
       const Matrix2x2<KLT_TYPE>& AffinePred,
       Matrix2x2<KLT_TYPE>& AffineResult)
{
  
#ifdef BIAS_DEBUG
  if (AffinePred.NormFrobenius()<1e-10) {
    BIASERR("affine guess seems unitialized, set at least identity");
  }
#endif    
  residuumMAD_ = 0;
  residuumMSD_ = 0;
  
  int res = TrackAffine_(p1.GetData(), p2.GetData(), AffinePred,
                         result.GetData(), AffineResult,
                         error, iter, Cov_, residuumMAD_, residuumMSD_, 
                         _SimilarityTransformOnly, true);  
  return res;
}

// ***************** parts, which are not under LGPL *************************
// The function TrackAffine_ is based on the affine tracker by Thorsten
// Thormaehlen. Some of the worker functions have been rewritten, some
// have been adapted, see below for the conditions on this software.

template <class StorageType>
int TrackerBaseAffine<StorageType>::
TrackAffine_(KLT_TYPE p1[2], KLT_TYPE p2[2], 
             const Matrix2x2<KLT_TYPE>& p2A,
             KLT_TYPE result[2],  Matrix2x2<KLT_TYPE>& resultA,
             KLT_TYPE& error, int &iteration, Matrix<KLT_TYPE>& Cov,
             KLT_TYPE& mad, KLT_TYPE& msd,
             bool SimilarityTransformOnly, bool ComputeCovariance)

{
  //this->AddDebugLevel(D_TRACKERB_AFFINE | D_TRACKERB_KLT);
  BIASDOUT(D_TRACKERB_AFFINE, "Called Affine for "
           <<p1[0]<<";"<<p1[1]<<", guess is "<<p2[0]<<";"<<p2[1]
           <<" affine guess is "<< p2A);

  const int width = _WinSize;
  const int height = _WinSize;
  const int hw = width/2;
  const int hh = height/2;

  const KLT_TYPE fhw = hw;
  const KLT_TYPE fhh = hh;

  if (_minx1 >= p1[0] || _miny1 >= p1[1] || _maxx1 < p1[0] || _maxy1 < p1[1]){
    BIASDOUT(D_TRACKERB_KLT, "original point is too close to border: "
             <<p1[0]<<";"<<p1[1]);
    return  TRACKER_OUTOFIMAGE;
  }

 
  const int max_iterations = _MaxIterations;
  const KLT_TYPE maxerr    = _MaxError;
  const KLT_TYPE th_aff    = _MaxAffineChange;
  
  KLT_TYPE dx=0, dy=0;
  iteration = 0;
  int status = 0;
  
  Matrix<KLT_TYPE> T(SimilarityTransformOnly?4:6,SimilarityTransformOnly?4:6);
  Vector<KLT_TYPE> a(SimilarityTransformOnly?4:6);

  Cov.newsize(SimilarityTransformOnly?4:6,SimilarityTransformOnly?4:6);


  resultA = p2A;
  KLT_TYPE *Axx = &(resultA[0][0]);
  KLT_TYPE *Axy = &(resultA[0][1]); 
  KLT_TYPE *Ayx = &(resultA[1][0]);
  KLT_TYPE *Ayy = &(resultA[1][1]);

  result[0] = p2[0];
  result[1] = p2[1];

  bool convergence = false;

 
  // Allocate memory for windows
  KLT_TYPE *imgdiff = new KLT_TYPE[width * height];
  KLT_TYPE *gradx   = new KLT_TYPE[width * height];
  KLT_TYPE *grady   = new KLT_TYPE[width * height];

  // compute reference window only ONCE
  BilinearRegion1_(p1[0], p1[1], hh);

  do  {
    KLT_TYPE ul_x =  result[0] + *Axy *  fhh  - *Axx * fhw;
    KLT_TYPE ul_y =  result[1] + *Ayy *  fhh  - *Ayx * fhw; 
    KLT_TYPE ll_x =  result[0] - *Axx *  fhw  - *Axy * fhh; 
    KLT_TYPE ll_y =  result[1] - *Ayx *  fhw  - *Ayy * fhh;
    KLT_TYPE ur_x =  result[0] + *Axx *  fhw  + *Axy * fhh; 
    KLT_TYPE ur_y =  result[1] + *Ayx *  fhw  + *Ayy * fhh;
    KLT_TYPE lr_x =  result[0] + *Axx *  fhw  - *Axy * fhh; 
    KLT_TYPE lr_y =  result[1] + *Ayx *  fhw  - *Ayy * fhh;

    if ( ul_x  < _minx2 ||  ul_x  > _maxx2 ||
         ll_x  < _minx2 ||  ll_x  > _maxx2 ||
         ur_x  < _minx2 ||  ur_x  > _maxx2 ||
         lr_x  < _minx2 ||  lr_x  > _maxx2 ||
         ul_y  < _miny2 ||  ul_y  > _maxy2 ||
         ll_y  < _miny2 ||  ll_y  > _maxy2 ||
         ur_y  < _miny2 ||  ur_y  > _maxy2 ||
         lr_y  < _miny2 ||  lr_y  > _maxy2) {
      status = TRACKER_OUTOFIMAGE;
      BIASDOUT( D_TRACKERB_KLT, "out of image");
      break;
    }
    
    // compute deltaI and also brightness normalization factor dev2_
    ComputeIntensityDifferenceAffine_(result[0], result[1],  
                                      *Axx, *Ayx , *Axy, *Ayy,
                                      hw, hh, imgdiff);
    
    // compute gradients and apply dev2_ to gradients
    GetGradientWinAffine_(result[0], result[1], 
                          *Axx, *Ayx , *Axy, *Ayy,
                          hw, hh, gradx, grady);

    if (SimilarityTransformOnly){
    
      _am_compute4by1ErrorVector(imgdiff, gradx, grady, hw, hh, a);
      _am_compute4by4GradientMatrix(gradx, grady, hw, hh, T);
	
      status = _am_gauss_jordan_elimination(T,a);
      	
      *Axx += a[0];
      *Ayx += a[1];
      *Ayy = *Axx;
      *Axy = -(*Ayx);
	
      result[0] += dx = a[2];
      result[1] += dy = a[3];
    } else {
     
      _am_compute6by1ErrorVector(imgdiff, gradx, grady, hw, hh, a);
      _am_compute6by6GradientMatrix(gradx, grady, hw, hh, T);
      
      status = _am_gauss_jordan_elimination(T,a);
	
      *Axx += a[0];
      *Ayx += a[1];
      *Axy += a[2];
      *Ayy += a[3];

      result[0] += dx = a[4];
      result[1] += dy = a[5];          
    }
      
    double posdiag = 
      (*Axx + *Axy) * (*Axx + *Axy) + (*Ayx + *Ayy) * (*Ayx + *Ayy);
    
    double negdiag = 
      (*Axy - *Axx) * (*Axy - *Axx) + (*Ayy - *Ayx) * (*Ayy - *Ayx);

    double ratio = (posdiag>negdiag) ? negdiag / posdiag : posdiag / negdiag;

    // factor 3 is allowed
    if (ratio < 0.09) {
      status = TRACKER_ODDAFFINEWARP;
      BIASDOUT(D_TRACKERB_KLT, "diag ratio is too big: "<< sqrt(ratio));
      break;
    }

    double scalex = (*Axx * *Axx) + (*Ayx * *Ayx);
    double scaley = (*Ayy * *Ayy) + (*Axy * *Axy);
      

    ratio = (scalex>scaley) ? scaley / scalex : scalex / scaley;
    // factor 3 is allowed
    if (ratio < 0.09) {
      status = TRACKER_ODDAFFINEWARP;
      BIASDOUT(D_TRACKERB_KLT, "axes ratio is too big: "<< sqrt(ratio));
      break;
    }

   
    // old upper left corner - new upper left corner 
    ul_x -=  result[0] - *Axx * fhw + *Axy * fhh;  
    ul_y -=  result[1] - *Ayx * fhw + *Ayy * fhh; 
    // old lower left corner - new lower left corner 
    ll_x -=  result[0] - *Axx * fhw - *Axy * fhh;  
    ll_y -=  result[1] - *Ayx * fhw - *Ayy * fhh;
    // old upper right corner - new upper right corner 
    ur_x -=  result[0] + *Axx * fhw + *Axy * fhh;  
    ur_y -=  result[1] + *Ayx * fhw + *Ayy * fhh;
    // old lower right corner - new lower right corner 
    lr_x -=  result[0] + *Axx * fhw - *Axy * fhh;  
    lr_y -=  result[1] + *Ayx * fhw - *Ayy * fhh;
    
    convergence = (fabs(dx) < maxerr && fabs(dy) < maxerr  &&
                   fabs(ul_x) < th_aff && fabs(ul_y) < th_aff &&
                   fabs(ll_x) < th_aff && fabs(ll_y) < th_aff &&
                   fabs(ur_x) < th_aff && fabs(ur_y) < th_aff &&
                   fabs(lr_x) < th_aff && fabs(lr_y) < th_aff);
    
    if (status == TRACKER_NOSTRUCTURE) {
      BIASDOUT( D_TRACKERB_KLT, "no structure for tracking");
      break;
    }
    iteration++;
 
  }  while ( !convergence  && iteration < max_iterations); 
  if (iteration >= max_iterations) {
    status = TRACKER_TOOMANYITER;
    BIASDOUT( D_TRACKERB_KLT, "too many iter: "<<iteration
              <<" maxiter="<<max_iterations);
  }
  mad = 0;
  msd = 0;
  if ((status==TRACKER_TOOMANYITER) || (status==TRACKER_SUCCESS)) {

    /** if we compute the difference here, we must check the range again 
    _am_computeIntensityDifferenceAffine(*_im1, *_im2, p1[0], p1[1], 
                                         result[0], result[1], 
                                         *Axx, *Ayx , *Axy, *Ayy,
                                         hw, hh, imgdiff);
    */


    const unsigned int wh = width*height;
    register KLT_TYPE* fw = imgdiff;
    for (unsigned int h=0; h<wh ; h++) {
      const KLT_TYPE cur = (KLT_TYPE)fabsf((float) (*fw) );
      fw++;
      mad += cur;
      msd += cur*cur;
    }
    
    mad /= double(wh)*dev1_;
    msd /= dev1_*dev1_;
    double sigma = msd / (double)(wh - SimilarityTransformOnly?4:6);
    msd /= double(wh);
 
    if (ComputeCovariance) {
      if (SimilarityTransformOnly)
        _am_compute4by4GradientMatrix(gradx, grady, hw, hh, T);
      else 
        _am_compute6by6GradientMatrix(gradx, grady, hw, hh, T);
      SVD covsvd(T);
      Cov = covsvd.Invert() * sigma;
    } else {
      // pseudo covariance (last step size)
      Cov.SetZero();
      if (SimilarityTransformOnly) {
        Cov[2][2] = dx * dx;
        Cov[3][3] = dy * dy;
      } else {
        Cov[4][4] = dx * dx;
        Cov[5][5] = dy * dy;
      }
    }
  }
  
  error = (fabs(dx)>fabs(dy))?fabs(dx):fabs(dy);

  // Check whether window is out of bounds
  if (result[0]<_minx2 || result[0] > _maxx2  || 
      result[1]<_miny2 || result[1] > _maxy2)
    status = TRACKER_OUTOFIMAGE;


  // Free memory 
  delete imgdiff;
  delete gradx;
  delete grady;

  BIASDOUT( D_TRACKERB_KLT, "status at end is "<<status);
 
  // Return appropriate value 
  return status;
}

template <class StorageType>
void TrackerBaseAffine<StorageType>::
EvaluateResult_(KLT_TYPE& mad, KLT_TYPE& msd, Matrix<KLT_TYPE>& CovMatrix) {
  // these values have already been computed in the track_ function !!!
  if (_SimilarityTransformOnly) {
    CovMatrix.newsize(4,4);
  } else {
    CovMatrix.newsize(6,6);
  }
  CovMatrix = Cov_;

  mad = residuumMAD_;
  msd = residuumMSD_;
}


/********************************************************************** 
* CONSISTENCY CHECK OF FEATURES BY AFFINE MAPPING (BEGIN)
* 
* Created by: Thorsten Thormaehlen (University of Hannover) June 2004    
* thormae@tnt.uni-hannover.de
* 
* Permission is granted to any individual or institution to use, copy, modify,
* and distribute this part of the software, provided that this complete 
* authorship and permission notice is maintained, intact, in all copies. 
*
* This software is provided  "as is" without express or implied warranty.
*
* 
*/

int _am_gauss_jordan_elimination(Matrix<KLT_TYPE>& a,
                                 Vector<KLT_TYPE>& b) {
  const int n = a.GetRows();
  Vector<int> indxc(n),indxr(n),ipiv(n);
  int i,j,k,l,ll;
  KLT_TYPE big,dum,pivinv,tmp;
  int col = 0;
  int row = 0;

  for (j=0;j<n;j++) ipiv[j]=0;
  for (i=0;i<n;i++) {
    big=0.0;
    for (j=0;j<n;j++)
      if (ipiv[j] != 1)
        for (k=0;k<n;k++) {
          if (ipiv[k] == 0) {
            if (fabs(a[j][k]) >= big) {
              big=fabs(a[j][k]);
              row=j;
              col=k;
            }
          } else if (ipiv[k] > 1) return TRACKER_NOSTRUCTURE;
        }
    ++(ipiv[col]);
    if (row != col) {
      for (l=0;l<n;l++) {
        tmp = a[row][l];
        a[row][l] = a[col][l];
        a[col][l] = tmp;
      }
      tmp = b[row];
      b[row] = b[col];
      b[col] = tmp;
    }
    indxr[i]=row;
    indxc[i]=col;
    if (a[col][col] == 0.0) return TRACKER_NOSTRUCTURE;
    pivinv=1.0/a[col][col];
    a[col][col]=1.0;
    for (l=0;l<n;l++) a[col][l] *= pivinv;
    b[col] *= pivinv;
    for (ll=0;ll<n;ll++)
      if (ll != col) {
        dum=a[ll][col];
        a[ll][col]=0.0;
        for (l=0;l<n;l++) a[ll][l] -= a[col][l]*dum;
        b[ll] -= b[col]*dum;
      }
  }
  for (l=n-1;l>=0;l--) {
    if (indxr[l] != indxc[l])
      for (k=0;k<n;k++) {
        tmp = a[k][indxr[l]];
        a[k][indxr[l]] = a[k][indxc[l]];
        a[k][indxc[l]] = tmp;
      }
  }
 
  return TRACKER_SUCCESS;
}

template <class StorageType>
void TrackerBaseAffine<StorageType>::
GetGradientWinAffine_(const KLT_TYPE& x, const KLT_TYPE& y,     
                      const KLT_TYPE& Axx, const KLT_TYPE& Ayx,
                      const KLT_TYPE& Axy, const KLT_TYPE& Ayy,  
                      const int hw, const int hh,   
                      KLT_TYPE* out_gradx,    
                      KLT_TYPE* out_grady) {
  if (_ComputeFilteredImages) {

    double dev2 = dev2_;

    BIASWARNONCE("Inefficient code: affine sparse tracking, needs rewrite."
                 <<" either use precomputed gradients or implment this ");


    // for some reason this approximation does not work (traking interrupts)
    // dont debug this because it is very inefficient either !!!
    BIASABORT; 
    
    


    bool AffineBrightnessInvariance = _AffineBrightnessInvariance;
    _AffineBrightnessInvariance = false;

    register int i, j;
    KLT_TYPE mi, mj;
    Matrix<KLT_TYPE> I(2*hh+1, 2*hw+1, MatrixZero), 
      gx(2*hh+1, 2*hw+1, MatrixZero), gy(2*hh+1, 2*hw+1, MatrixZero);
    for (j = -hh ; j <= hh ; j++) {
      for (i = -hw ; i <= hw ; i++)  {
       


        // FIXME: reeeeally inefficient!



        mi = Axx * i + Axy * j;
        mj = Ayx * i + Ayy * j;

        BilinearRegion2_(x+mi, y+mj, 0);
        gx[j+hh][i+hw] = _gx2[0][0];
        gy[j+hh][i+hw] = _gy2[0][0];
        I[j+hh][i+hw] = _bl2[0][0];
      }
    }
    _AffineBrightnessInvariance = AffineBrightnessInvariance;
    if (_AffineBrightnessInvariance) {
      this->NormalizeRegion_(I, gx, gy);
    }
    dev2_ = dev2;
    KLT_TYPE *pgx = gx.GetData(), *pgy = gy.GetData();
    for (int c = (2*hh+1)*(2*hw+1); c>0; c--) {
      *out_gradx++ = *pgx++;
      *out_grady++ = *pgy++;
    }
  } else {
    register int i, j;
    KLT_TYPE mi, mj;
    if (_AffineBrightnessInvariance) {
      for (j = -hh ; j <= hh ; j++) {
        for (i = -hw ; i <= hw ; i++)  {
          mi = Axx * i + Axy * j;
          mj = Ayx * i + Ayy * j;
          
          // normalize by precomputed dev2_
          *out_gradx++ = _gradim2x->FastBilinearInterpolationGrey(x+mi,
                                                                  y+mj)*dev2_;
          *out_grady++ = _gradim2y->FastBilinearInterpolationGrey(x+mi, 
                                                                  y+mj)*dev2_;
        }
      }
    } else {
      for (j = -hh ; j <= hh ; j++) {
        for (i = -hw ; i <= hw ; i++)  {
          mi = Axx * i + Axy * j;
          mj = Ayx * i + Ayy * j;
          *out_gradx++ = _gradim2x->FastBilinearInterpolationGrey(x+mi, y+mj);
          *out_grady++ = _gradim2y->FastBilinearInterpolationGrey(x+mi, y+mj);
        }
      }
    }
  }
}

template <class StorageType>
void TrackerBaseAffine<StorageType>::
ComputeIntensityDifferenceAffine_(const KLT_TYPE& x2, const KLT_TYPE& y2,   
                                  const KLT_TYPE& Axx, const KLT_TYPE& Ayx, 
                                  const KLT_TYPE& Axy, const KLT_TYPE& Ayy, 
                                  const int hw, const int hh, 
                                  KLT_TYPE* imgdiff) {
  register int i, j;
  BIASASSERT(hh==hw);
  KLT_TYPE* imgdifforig = imgdiff;
  if (_ComputeFilteredImages) {
    for (j = -hh ; j <= hh ; j++) {
      for (i = -hw ; i <= hw ; i++)  {
        // FIXME: reeeeally inefficient to interpolate always with size 1 !!!
        /// write specialized affine function
        BilinearRegion2_(x2 + Axx * i + Axy * j, y2+ Ayx * i + Ayy * j, 0);
        *imgdiff++ = _bl1[j+hh][i+hw] - _bl2[0][0];
      }
    }
  } else { 
    mean1_ = mean2_ = 0;
    BIASASSERT(_bl1.num_rows()==_bl1.num_cols());
    BIASASSERT(_bl1.num_rows()==2*hw+1);
    KLT_TYPE mi,mj;
    KLT_TYPE *pbl1 = _bl1.GetData();
    KLT_TYPE *pbl2 = _bl2.GetData();
    for (j = -hh ; j <= hh ; j++) {
      for (i = -hw ; i <= hw ; i++)  {
        mi = Axx * i + Axy * j;
        mj = Ayx * i + Ayy * j;
        mean2_ += *pbl2 =  _im2->FastBilinearInterpolationGrey(x2+mi, y2+mj);
        *imgdiff++ = *pbl1++ - *pbl2++;
      }
    }
    if (_AffineBrightnessInvariance) {
      pbl2 = _bl2.GetData();
      const KLT_TYPE thearea = (2*hh+1)*(2*hw+1);
      mean2_ /= thearea;
      dev2_ = 0;
      double diff;
      for (j = -hh ; j <= hh ; j++)
        for (i = -hw ; i <= hw ; i++) {
          diff = (*pbl2++ -= mean2_);
          dev2_ += diff * diff;
        }
      if (fabs(dev2_)>1e-10) dev2_ = sqrt(thearea / dev2_);
      else dev2_ = 1.0;
      imgdiff = imgdifforig;
      pbl1 = _bl1.GetData();
      pbl2 = _bl2.GetData();
      for (j = -hh ; j <= hh ; j++) {
        for (i = -hw ; i <= hw ; i++) {
          *imgdiff++ =(*pbl1++) - (*pbl2++ *= dev2_);
        }
      }
    }
  } 
}


void _am_compute6by6GradientMatrix(KLT_TYPE* gradx,
                                   KLT_TYPE* grady,
                                   const int hw, 
                                   const int hh,
                                   Matrix<KLT_TYPE>& T) {
  register int i, j;
  KLT_TYPE gx, gy, gxx, gxy, gyy,  x, y, xx, xy, yy;
 
  
  // init with zero
  for (j = 0 ; j < 6 ; j++)  {
    for (i = j ; i < 6 ; i++)  {
      T[j][i] = 0.0;
    }
  }
  
  for (j = -hh ; j <= hh ; j++) {
    for (i = -hw ; i <= hw ; i++)  {
      gx = *gradx++;
      gy = *grady++;
      gxx = gx * gx;
      gxy = gx * gy;
      gyy = gy * gy;
      x = i; 
      y = j; 
      xx = x * x;
      xy = x * y;
      yy = y * y;
      
      T[0][0] += xx * gxx; 
      T[0][1] += xx * gxy;
      T[0][2] += xy * gxx;
      T[0][3] += xy * gxy;
      T[0][4] += x  * gxx;
      T[0][5] += x  * gxy;
	
      T[1][1] += xx * gyy;
      T[1][2] += xy * gxy;
      T[1][3] += xy * gyy;
      T[1][4] += x  * gxy;
      T[1][5] += x  * gyy;
			 
      T[2][2] += yy * gxx;
      T[2][3] += yy * gxy;
      T[2][4] += y  * gxx;
      T[2][5] += y  * gxy;
	 
      T[3][3] += yy * gyy;
      T[3][4] += y  * gxy;
      T[3][5] += y  * gyy; 

      T[4][4] += gxx; 
      T[4][5] += gxy;
      
      T[5][5] += gyy; 
    }
  }
  
  for (j = 0 ; j < 5 ; j++)  {
    for (i = j+1 ; i < 6 ; i++)  {
      T[i][j] = T[j][i];
    }
  }

}

void _am_compute6by1ErrorVector(KLT_TYPE* imgdiff,
                                KLT_TYPE* gradx,
                                KLT_TYPE* grady,
                                const int hw,
                                const int hh,
                                Vector<KLT_TYPE>& e) {
  register int i, j;
  register KLT_TYPE diff,  diffgradx,  diffgrady;

  /* Set values to zero */  
  for(i = 0; i < 6; i++) e[i] = 0.0; 
  
  /* Compute values */
  for (j = -hh ; j <= hh ; j++) {
    for (i = -hw ; i <= hw ; i++)  {
      diff = *imgdiff++;
      diffgradx = diff * (*gradx++);
      diffgrady = diff * (*grady++);
      e[0] += diffgradx * i;
      e[1] += diffgrady * i;
      e[2] += diffgradx * j; 
      e[3] += diffgrady * j; 
      e[4] += diffgradx;
      e[5] += diffgrady; 
    }
  }
  
  for(i = 0; i < 6; i++) e[i] *= 0.5;
  
}


void _am_compute4by4GradientMatrix(KLT_TYPE* gradx,
                                   KLT_TYPE* grady,
                                   const int hw,
                                   const int hh,
                                   Matrix<KLT_TYPE>& T)
{
  register int i, j;
  KLT_TYPE gx, gy, x, y;
 
  
  // Set values to zero 
  for (j = 0 ; j < 4 ; j++)  {
    for (i = 0 ; i < 4 ; i++)  {
      T[j][i] = 0.0;
    }
  }
  
  for (j = -hh ; j <= hh ; j++) {
    for (i = -hw ; i <= hw ; i++)  {
      gx = *gradx++;
      gy = *grady++;
      x = i; 
      y = j; 
      T[0][0] += (x*gx+y*gy) * (x*gx+y*gy);
      T[0][1] += (x*gx+y*gy)*(x*gy-y*gx);
      T[0][2] += (x*gx+y*gy)*gx;
      T[0][3] += (x*gx+y*gy)*gy;
   
      T[1][1] += (x*gy-y*gx) * (x*gy-y*gx);
      T[1][2] += (x*gy-y*gx)*gx;
      T[1][3] += (x*gy-y*gx)*gy;
     
      T[2][2] += gx*gx;
      T[2][3] += gx*gy;
      
      T[3][3] += gy*gy;
    }
  }
  
  for (j = 0 ; j < 3 ; j++)  {
    for (i = j+1 ; i < 4 ; i++)  {
      T[i][j] = T[j][i];
    }
  }

}

void _am_compute4by1ErrorVector(KLT_TYPE* imgdiff,
                                KLT_TYPE* gradx,
                                KLT_TYPE* grady,
                                const int hw, 
                                const int hh,
                                Vector<KLT_TYPE>& e){
  register int i, j;
  register KLT_TYPE diff,  diffgradx,  diffgrady;

  /* Set values to zero */  
  for(i = 0; i < 4; i++) e[i] = 0.0; 
  
  /* Compute values */
  for (j = -hh ; j <= hh ; j++) {
    for (i = -hw ; i <= hw ; i++)  {
      diff = *imgdiff++;
      diffgradx = diff * (*gradx++);
      diffgrady = diff * (*grady++);
      e[0] += diffgradx * i + diffgrady * j;
      e[1] += diffgrady * i - diffgradx * j;
      e[2] += diffgradx;
      e[3] += diffgrady;
    }
  }
  
  for(i = 0; i < 4; i++) e[i] *= 0.5;
  
}

/**************** end affine part  ********************************/ 


//////////////////////////////////////////////////////////////////////////
//                 instantiation
//////////////////////////////////////////////////////////////////////////

template class BIASMatcher2D_EXPORT TrackerBaseAffine<unsigned char>; 
template class BIASMatcher2D_EXPORT TrackerBaseAffine<float>; 

// fill in instances as required
#ifdef BUILD_IMAGE_INT
template class TrackerBaseAffine<int>; 
#endif
#ifdef BUILD_IMAGE_CHAR
template class TrackerBaseAffine<char>; 
#endif
#ifdef BUILD_IMAGE_SHORT
template class TrackerBaseAffine<short>; 
#endif
#ifdef BUILD_IMAGE_USHORT
template class TrackerBaseAffine<unsigned short>; 
#endif
#ifdef BUILD_IMAGE_UINT
template class TrackerBaseAffine<unsigned int>; 
#endif
#ifdef BUILD_IMAGE_DOUBLE
#endif
