/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <Image/Camera.hh>
#include <Image/CameraParam.hh>
#include <Base/Image/ImageIO.hh>
#include <Base/Image/ImageAttributes.hh>
#include <iostream>
#include <fstream>

#ifdef   WIN32
#include "Base/Common/getopt_W32.h"
#else  //not WIN32
#include <unistd.h>
#endif //WIN32

#include <Utils/Param.hh>

using namespace std;
using namespace BIAS;


void usage()
{
  cout <<"Usage: biasextract [options] Image ..."<<endl
    <<" Options: "<<endl
    <<"-a      : extract ascii string"<<endl
    <<"-t      : extract timestamp as [sec usec]"<<endl
    <<"-T      : extract timestamp as [msec]"<<endl
    <<"-s      : suppress current filename in output"<<endl
    <<"-P      : extract P-Matrix"<<endl
    <<"-p      : Decompose P-Matrix and extract axis-angle orientation"
    <<endl
    <<"-x file : extract pmatrix as xml data"<<endl
    <<"-K      : extract internal parameter matrix"<<endl
    <<"-R      : extract rotation matrix"<<endl
    <<"-r      : extract axis-angle orientation"<<endl
    <<"-C      : extract camera center position"<<endl
    <<"-A      : extract viewing direction"<<endl	  
    <<"-V      : extract 'down' vector"<<endl	  
    <<"-H      : extract 'right' vector"<<endl	  
    <<"-e      : compute K from EXIF data"<<endl
    <<"-c      : compute CameraParam from KMatrix in header"<<endl
    <<"-E      : compute CameraParam from EXIF data"<<endl
    <<"-u      : extract UID"<<endl
    <<"-O      : extract orientation"<<endl       
    <<"-o file : write output to file"<<endl;
}


/** @class biasextract
@ingroup g_tools
@brief extract header data from image files: uid, time stamp orientation,
KMatrix from EXIF data, ..
@author evers */
int main(int argc,char *argv[]) 
{
  Param params(true);
  /**** GENERAL PARAMS *****/
  int generalGroupID = params.GetFreeGroupID();
  bool* help = params.AddParamBool("help", 
                                   "print out help",
                                   false, 'h', 
                                   generalGroupID);
  params.SetGroupName(generalGroupID, "general");
  /**** ADDITIONAL METADATA PARAMS *****/
  int additionalDataGroupID = params.GetFreeGroupID();
  bool* GetASCIIString = params.AddParamBool("GetASCIIString", 
                                             "extract ascii string",
                                             false, 'a', 
                                             additionalDataGroupID);
  bool* GetTimeStamp = params.AddParamBool("GetTimeStamp", 
                                           "extract timestamp as [sec usec]",
                                           false, 't', 
                                           additionalDataGroupID);
  bool* GetTimeMsec = params.AddParamBool("GetTimeMsec", 
                                          "extract timestamp as [msec]",
                                          false, 'T', 
                                          additionalDataGroupID);
  bool* GetUID = params.AddParamBool("GetUID", 
                                     "extract UID",
                                     false, 'u', 
                                     additionalDataGroupID);
  params.SetGroupName(additionalDataGroupID, "extract additional data");

  /**** CAMERA METADATA PARAMS *****/
  int camParamGroupID = params.GetFreeGroupID();

  bool* GetPMatrix = params.AddParamBool("GetPMatrix", 
                                         "Decompose P-Matrix",
                                         false, 'P', 
                                         camParamGroupID);
  bool* GetPandDecomposition = 
    params.AddParamBool("GetPMatrixAndDecompose", 
                        "Decompose P-Matrix and " 
                        "extract axis-angle orientation",
                        false, 'p', 
                        camParamGroupID);
    
  bool* GetR = params.AddParamBool("GetR", 
                                   "extract axis-angle orientation",
                                   false, 'R', 
                                   camParamGroupID);  
  bool* GetAxisAngle = params.AddParamBool("GetAxisAngle", 
                                           "extract axis-angle orientation",
                                           false, 'r', 
                                           camParamGroupID); 

  bool* GetOri = params.AddParamBool("GetOri", 
                                     "extract orientation as yaw, pitch, roll",
                                     false, 'O', 
                                     camParamGroupID);
  

  bool* GetC = params.AddParamBool("GetC", 
                                   "extract camera center position",
                                   false, 'C', 
                                   camParamGroupID);   
  bool* GetK = params.AddParamBool("GetK", 
                                   "extract internal parameter matrix",
                                   false, 'K', 
                                   camParamGroupID);   

  bool* GetKFromEXIF = params.AddParamBool("GetKFromEXIF", 
                                           "compute K from EXIF data",
                                           false, 'e', 
                                           camParamGroupID);
  bool* GetCameraParamFromEXIF = 
    params.AddParamBool("GetCameraParamFromEXIF", 
                        "compute CameraParam from EXIF data",
                        false, 'E', 
                        camParamGroupID);
  bool* GetCameraParamFromK = 
    params.AddParamBool("GetCameraParamFromK", 
                        "compute CameraParam from KMatrix in header",
                        false, 'c', 
                        camParamGroupID);


  bool* GetA = params.AddParamBool("GetA", 
                                   "extract viewing direction",
                                   false, 'A', 
                                   camParamGroupID);
  bool* GetH = params.AddParamBool("GetH", 
                                   "extract 'right' vector",
                                   false, 'H', 
                                   camParamGroupID);
  bool* GetV = params.AddParamBool("GetV", 
                                   "extract 'down' vector",
                                   false, 'V', 
                                   camParamGroupID);
#ifdef BIAS_HAVE_XML2
  string* PXMLName = 
    params.AddParamString("PXMLName", 
                          "write out pmatrix as xml data"
                          " into this file", 
                          "", 'x', 
                          camParamGroupID);


  string* GetProjection = 
    params.AddParamString("GetProjection", 
                        "extract projection and write into this file",
                        "", 0, 
                        camParamGroupID);
#endif

  params.SetGroupName(camParamGroupID, "extract camera parameters");
  
  /**** OUTPUT PARAMS *****/
  int outputGroupID = params.GetFreeGroupID();

  bool* SuppressFilename = 
    params.AddParamBool("SuppressFilename", 
                        "suppress current filename in output",
                        false, 's', 
                        outputGroupID);

  string* Outfile = params.AddParamString("Outfile", 
                                          "write results into this file", 
                                          "", 'o', 
                                          outputGroupID);

  params.SetGroupName(outputGroupID, "output");

  int fup = params.ParseCommandLine(argc, argv);

  if(*help) {
    params.Usage();
    usage();
    return 0;
  }


  if(*GetTimeMsec) *GetTimeStamp = true;
  if(*GetPandDecomposition) {
    *GetPMatrix = true; 
    *GetK = true; 
    *GetR = true; 
    *GetAxisAngle = true;
    *GetC = true;  
  }

  bool GetPXML = false;
#ifdef BIAS_HAVE_XML2
  if(PXMLName->size()>0) GetPXML = true;
#endif
  
  //   //int c;
  //   int c=1;
  //   while(c >0) {
  //     c = getopt(argc,argv,"aceo:ptrsux:ETPKRCAHVO");
  //     if (c>0){
  //       switch(c) {
  //         //       case 'a': GetASCIIString =true;
  //         //         break;
  //         //       case 't': GetTimeStamp = true;
  //         //         break;
  //         //       case 'e': GetKFromEXIF = true;
  //         //         break;
  //         //       case 'E': GetCameraParamFromEXIF = true;
  //         //         break;
  //         //       case 'c': GetCameraParamFromK = true;
  //         //         break;
  //         //       case 'T': 
  //         //         GetTimeStamp = true;
  //         //         GetTimeMsec = true;	
  //         //         break;
  //         //       case 'P': GetPMatrix = true;
  //         //         break;
  //         //       case 'p': 
  //         //         GetPMatrix = true; 
  //         //         GetK = true; 
  //         //         GetR = true; 
  //         //         GetAxisAngle = true;
  //         //         GetC = true;  
  //         //         break;
  //         //       case 'R': GetR = true;  
  //         //         break;
  //         //       case 'r': GetAxisAngle = true;
  //         //         break;
  //         //       case 'C': GetC = true;  
  //         //         break;
  //         //       case 'A': GetA = true;  
  //         //         break;
  //         //       case 'H': GetH = true;  
  //         //         break;
  //         //       case 'V': GetV = true;  
  //         //         break;
  //         //       case 'K': GetK = true;  
  //         //         break;
  //         //      case 'u': GetUID = true;
  //         //         break;
  //         //      case 'O': GetOri = true;
  //         //         break;
  //       case 'o': Outfile = optarg;
  //         break;
  //         //       case 's': SuppressFilename = true;
  //         //         break;
  //         //       case 'x':  GetPXML = true;
  //         //         PXMLName = optarg;
  //         //         break; 
  //       default:
  //         usage();
  //         //exit(1);
  //         return 0; // for Dart testing 
  //       }
  //     }
  //   }


  if (argc-fup < 1) {
    cout <<"==> To few arguments."<<endl;
    params.Usage();
    usage();
    exit(1);
  }
  ofstream ofs;
  if (Outfile->size()>0) {
    ofs.open(Outfile->c_str());
    if (!ofs) {
      perror(Outfile->c_str());
      exit(1);
    }
    cout <<"Writing to file: "<<*Outfile<<endl;
    cout.rdbuf(ofs.rdbuf());
  } else {
    //     ofs = cout;
    if (!(*SuppressFilename))
      cout <<"Writing to stdout "<<endl;
  }

  //  Image<unsigned char> tmp;
  ImageBase tmp;

  for ( int i=fup; i<argc;i++) {
    //     ImageIO::Read(argv[i],pic);
    //    ImageIO::Load(argv[i],pic);
    ImageIO::Load(argv[i],tmp);
    Camera<unsigned char> pic(tmp);    
    pic.ParseMetaData();
    if(*GetASCIIString) {
      cerr<<"ASCII: "<<pic.GetASCIIString()<<endl;
    }

    if (*GetKFromEXIF) {
      ImageAttributes Att;
      Att.InitExifTags();
      if (Att.Read(argv[i])!=BIAS_IMAGEATTRIBUTES_SUCCESS) {
        cerr<<"error reading exif tags ..."<<endl;
      } else {
        KMatrix KExif;
        KExif.SetIdentity();
        bool success = true;
        success = success && (Att.GetFocalLengthXPixel(KExif[0][0])>=0);
        success = success && (Att.GetFocalLengthYPixel(KExif[1][1])>=0);
        success = success && (Att.GetPrincipalPoint(KExif[0][2],
                                                    KExif[1][2] )>=0);
        if (success) {
          cerr<<"K: ";
          cout<<KExif<<endl;
        } else {
          cerr<<"Error, not enough EXIF data to get K:"<<KExif<<endl;
        }
      }
    }
    if (*GetCameraParamFromEXIF) {
#ifdef BIAS_HAVE_XML2
      CameraParam cp;
      if (cp.ReadFromEXIF(argv[i])==0) {
        stringstream cpfilename;
        cpfilename <<argv[i]<<".cameraparam.xml";
        cp.WriteCameraData(cpfilename.str());
      } else {
        cerr<<"Error, not enough EXIF data to construct cameraparam"<<endl;
      }
#else
      cerr<<"Missing XML2 support led to missing class BIAS::CameraParam, "
          <<"cant extract EXIF."<<endl;
#endif
    }
    pic.ParseMetaData();
    if (!(*SuppressFilename)) cout <<argv[i]<<"\t"<<endl;
    if (*GetUID)   cout<<pic.GetUID()<<"  ";  
    if (*GetTimeStamp) {
      long sec,usec;
      pic.GetTime(sec,usec);
      // cout <<"true msec: "<<pic.GetTime()<<endl;
      if (GetTimeMsec) cout<< (sec % 1000) * 1000 + (usec / 1000);
      else cout<<sec<<"  "<<usec;
    }
    if (*GetOri) {
      float y,p,r;
      pic.GetOrientation(y,p,r);
      cout<<y<<"   "<<p<<"   "<<r;
    }
    if (*GetPMatrix || GetPXML) {
      //pic.PrintAppData(std::cout);
      PMatrix P;
      P.SetZero();
      if (pic.IsPValid()) {
        // read composed P directly, ignore possibel decomposition 
        P=pic.GetP();
      };
      // otherwise copose form K,R,C
      // compose, if P is zero, too
      if (!pic.IsPValid() || P.IsZero() ){
        KMatrix K;
        RMatrix R;
        K.SetIdentity();
        R.SetIdentity();
        BIAS::Vector3<double> C(0,0,0);

        // check if valid Pose in meta data
        if (pic.IsRValid()) 
          R=pic.GetR();
        else 
          BIASERR("could not read R, assuming identity");

        if (pic.IsCValid())
          C=pic.GetC();
        else 
          BIASERR("could not read C, assuming zero");

        if (pic.IsKValid())
          K=pic.GetK();
        else 
          BIASERR("could not read K, assuming identity");

        // compose from separation: 
        P.Compose(K, R, C);
      }
      if (*GetPMatrix) 
        cout<< P<<endl;
      if (GetPXML) {
#ifdef BIAS_HAVE_XML2
        P.XMLWrite(*PXMLName);
#else
        cerr<<"No XML support found, so no writing of P as xml !"<<endl;
#endif
      }
    }
    if (*GetK) {
      KMatrix K;
      if (pic.IsKValid()) {
        K = pic.GetK();
      } else {
        pic.GetP().GetK();
      }
      cout<<"K: "<<K<<endl;
    }
    if (*GetR){
      PMatrix P = pic.GetP();
      cout<<"R: "<<P.GetR()<<endl;
    }
    if (*GetAxisAngle){
      PMatrix P = pic.GetP();
      Vector3<double> axis;
      double angle;
      RMatrix R;
      R = P.GetR();
      R.GetRotationAxisAngle(axis, angle);
      cout<<"r[axis, angle]: "<<axis[0]<<" "<<axis[1]
          <<" "<<axis[2]<<" "<<angle<<endl;
      cout<<"|axis|= "<<axis.NormL2()<<endl;
      cout<<"R: "<<R<<endl;
      RMatrix temp;
      temp.Set(axis, angle);
      cout<<"R-axisAngle:\n "<<R-temp<<endl;

    }
    if (*GetC){
      PMatrix P = pic.GetP();
      cout<<"C: "<<P.GetC()<<endl;
    }
    if (*GetA){
      PMatrix P = pic.GetP();
      cout<<"A: "<<P.GetA()<<endl;
    }
    if (*GetH){
      PMatrix P = pic.GetP();
      cout<<"H: "<<P.GetH()<<endl;
    }
    if (*GetV){
      PMatrix P = pic.GetP();
      cout<<"V: "<<P.GetV()<<endl;
    }

    if (*GetCameraParamFromK) {
#ifdef BIAS_HAVE_XML2
      KMatrix K;
      bool HaveK = true;
      if (pic.IsKValid()) {
        K = pic.GetK();
      } else if (pic.IsPValid()) {
        K = pic.GetP().GetK();
      } else {
        HaveK = false;
      }
      if (HaveK) {
        CameraParam cp(K, pic.GetWidth(), pic.GetHeight());
        stringstream cpfilename;
        cpfilename <<argv[i]<<".cameraparam.xml";
        cp.WriteCameraData(cpfilename.str());
      } else {
        cerr<<"Error, no K to construct cameraparam"<<endl;
      }
#else
      cerr<<"Missing XML2 support led to missing class BIAS::CameraParam, "
          <<"cant extract EXIF."<<endl;
#endif
    }
    cout<<endl;
#ifdef BIAS_HAVE_XML2
    if(GetProjection->size() > 0) {
      Projection proj = pic.GetProj();
      if(proj.XMLWrite(*GetProjection)!=0) {
        BIASERR("error writing projection");
      }
    }
#endif

  }//end of for

  if (Outfile->size() > 0)  
    ofs.close();


  return 0;
}
