/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef __BIASPARAMETER__
#define __BIASPARAMETER__
#include <Base/Common/BIASpragmaStart.hh>

#include "bias_config.h" 

#include "ParamTypes.hh"
#include <stdlib.h>
#include <string>
#include <vector>
#include <float.h>
#include <map>

#ifdef WIN32
#  include <Base/Common/getopt_W32.h>
#else
#  include <getopt.h>
#endif

#include <Base/Debug/Error.hh>
#include <Base/Debug/Debug.hh>
#include <Base/Math/Vector.hh>
#ifdef BIAS_HAVE_XML2
#  include <Base/Common/XMLIO.hh>
#endif



namespace BIAS {

#define D_PARAM_ALL 0xffffffff
#define D_PARAM_IO           1<<0
#define D_PARAM_PARSECMDL    1<<1
#define D_PARAM_GET          1<<2
#define D_PARAM_CONF_FILE    1<<3

#define GRP_ALL -2
#define GRP_NOSHOW -1

    /**
    @class ParamCallbackInterface
    @ingroup g_utils
    @brief This class provides an interface to be called if parameter changes
    occured.

    If an application makes use of ParamGui and has to react on parameter 
    changes, it should be derived from this class and could register a 
    callback for each instance to this method.
    */
    class BIASUtils_EXPORT ParamCallbackInterface 
    {
    public: 
        virtual  ~ParamCallbackInterface();
        ParamCallbackInterface();

        virtual  void ParameterChanged(
            const std::string &paramname, 
            const void *data);

    };


    /** @class Param
    @ingroup g_utils
    @brief This class Param provides generic support for parameters.

    This is the documentation for the

    BIAS::Param

    and the

    BIAS::ParamGUI

    classes.



    Philosophy
    ==========

    The idea behind the Param object is a common class to register all
    program parameters in. These parameters can be set from the command
    line, read and written to a file or be edited in a graphical interface
    that is programmed using the wxWidgets framework toolkit.


    Usage
    =====

    There are the following seven parameter types:

    - Integer
    - Boolean
    - String
    - Double
    - Vector of Double
    - Vector of Integer
    - Enumeration

    At first you create your instance of the Param oder ParamGUI object in
    your application. Afterwards you register your parameters by using the
    AddParam... methods corresponding to each of the types above:

    The methods are
    - AddParamInt
    - AddParamBool
    - AddParamString
    - AddParamDouble
    - AddParamVecDbl
    - AddParamVecInt
    - AddParamEnum.

    Each method has the parameters name, a help-string, a default-value,
    a short-command and a group ID as parameters.

    The name must be unique! An error is displayed if you add a new parameter
    with an already existing name.

    The help-string should describe the command in a short way.

    If you use the short-command, the parameter can be set at the command
    line with -<short-command> additionally to --<parameter-name>.

    Using the group ID you can put parameters of similar meaning in single
    groups, which is useful for visualization.

    The first four parameter types come along with standard default values.
    For the vector types you must provide an initial vector when adding a
    new parameter. When setting new vectors later in your program, the
    vector must be of the size given by the initial one.

    Using the method ParseCommandLine you can set the values for the
    registered parameters from the command line. Also you can use the
    default parameters 'readconfig' and 'writeconfig' to read or write a
    configuration from/to a file.

    In your program you can use the GetParam... and SetParam... methods to
    receive a pointer to the data or to set new values for the parameters.


    The GUI
    =======

    The BIAS::ParamGUI class is derived from the BIAS::Param class and
    provides the three methods 

    GetPanel,
    GetNotebook and
    GetFrame.

    The GetPanel method returns a pointer to a wxPanel where the widgets
    for the parameters belonging to the given group are placed.

    The GetNotebook method returns a pointer to a wxNotebook where on each
    notebook page is placed a panel with the parameters of one group.

    These two can be placed in your GUI application as you like to. If you
    want a standalone frame with the notebook inside you use the GetFrame
    method.

    The GetFrame method returns a pointer to a wxFrame in which a wxNotebook
    with all groups of parameters is embedded.


    Why pointers?
    =============

    A big question about this object is why pointers are returned by
    the Add-, Get- and Set-methods. Well, is it wanted, that the data is
    only stored at one location and that is the Param object, so if you
    are interested in a value, you get the pointer to it.


    Internals
    =========

    Classes:

    BIAS::ParamTypeBase
    |
    +-- BIAS::ParamTypeInt
    |
    +-- BIAS::ParamTypeBool
    |
    +-- BIAS::ParamTypeString
    |
    +-- BIAS::ParamTypeDouble
    |
    +-- BIAS::ParamTypeVecDbl
    |
    +-- BIAS::ParamTypeVecInt
    |
    +-- BIAS::ParamTypeEnum


    BIAS::Param
    |
    +-- BIAS::ParamGUI


    STL maps:

    In BIAS::Param the main data struct holding the parameter data is a
    STL map. The key-value-pairs are of the type string,ParamTypeBase*.

    The key-string holds the name of the parameter which is also stored in
    the parameter type class. 

    The ParamTypeEnum class holds another STL map with key of type string and
    value of type int. This is used for easier name->number translation.

    In the GUI class also has a STL map of type
    unsigned int,struct ItemAndWidget to map event IDs to the corresponding
    control widget and parameter data.

    @author Jan-Friso Evers-Senne, Matthias Dunda

    Note about bool: bool options can be used on the cmdl as follow:
    -b                         -> result in "true"
    --bool=true                -> result in "true"
    --bool=<any other string>  -> result in "false"
    --bool <any string>        -> Not allowed.
    */
    class BIASUtils_EXPORT Param : public Debug{

    public:
      /** Extracts lines from passed file.
       * The specified file is expected to contain a list of entries. Those
       * entries are assumed to be seperated by newlines.
       * \param ListFileName Routine tries to open file with this name.
       * \param LinesInFile result of line extraction.
       * \returns -1 if file could not be opened.
       * \returns 0 all good.
       */
      static int ParseListFile(const std::string& ListFileName, 
			       std::vector<std::string>& LinesInFile);
      
        explicit Param(bool DisableDestructorWarning = false);
        virtual ~Param();

        /* All routines for adding, getting and setting parameters return
        a pointer to the data on success or NULL on error! */

        // routines for ADDING parameters
        /** For all adding routines:
        @param name is the unique identifier of the parameter
        @param help is a help or 'tooltip' string for that parameter
        @param deflt is a default value for that parameter. Note that not
        every parameter type has a default at compile time!
        @param cmdshort is a short command option for the command line
        @param Group is a number to group some parameters for visualization,
        e.g. all matchers are in group 1, all trackers are in group 37 etc.
        When printing to screen, you can limit the output to a group of
        interest. The default is 0 - no special group.
        @author mdunda 03 2004 */

        int * AddParamInt( const std::string &name, 
            const std::string &help, int deflt = 0, 
            int min = INT_MIN, int max = INT_MAX,
            char cmdshort = 0, int Group = GRP_NOSHOW);

        bool * AddParamBool( const std::string &name, 
            const std::string &help,
            bool deflt = false, 
            char cmdshort = 0, int Group = GRP_NOSHOW );

        std::string * AddParamString( const std::string &name, 
            const std::string &help,
            std::string deflt = "", 
            char cmdshort = 0, int Group = GRP_NOSHOW );

        double * AddParamDouble( const std::string &name, 
            const std::string &help,
            double deflt = 0.0, 
            double min = -DBL_MAX, double max = DBL_MAX,
            char cmdshort = 0, int Group = GRP_NOSHOW );

        /** Add a parameter that expects a string on command line
        like "<value0>  <value1> <value2> ..." <br>
        A vector of double must be given as default vector! */
        BIAS::Vector<double> * AddParamVecDbl( const std::string &name, 
            const std::string &help,
            const BIAS::Vector<double> &deflt,
            char cmdshort = 0, 
            int Group = GRP_NOSHOW );

        /** The same as above, but default is given by a string,
        syntax is as at command line "v_0 v_1 v_2", don't set empty
        string, cause the length is determined also by the number of values

        @author Daniel Grest, July 2005
        */
        BIAS::Vector<double> * AddParamVecDbl( const std::string &name, 
            const std::string &help,
            const std::string &deflt,
            char cmdshort = 0, 
            int Group = GRP_NOSHOW );


        /**  Add a parameter that expects a string on command line
        like "<value0>  <value1> <value2> ..." <br>
        A vector of integer must be given a default vector! */
        BIAS::Vector<int> * AddParamVecInt( const std::string &name, 
            const std::string &help,
            const BIAS::Vector<int> &deflt,
            char cmdshort=0, int Group=GRP_NOSHOW);

        /** The same as other AddParamVecInt(), but default is given by a string,
        syntax is as at command line "v_0 v_1 v_2". Don't set empty
        string, cause the length is determined also by the number of values

        @author Daniel Grest, July 2005
        */
        BIAS::Vector<int> * AddParamVecInt( const std::string &name, 
            const std::string &help,
            const std::string &deflt,
            char cmdshort=0, int Group=GRP_NOSHOW);



        /** @param enums holds all enumeration names. 
        @param IDs can be a pointer to a vector of integers. It must have
        the same size as 'enums' and gives each name a number, which must
        be unique! If NULL the enums are numbered from 0 to enums.size()-1
        @author mdunda 03 2004 */
        int * AddParamEnum( const std::string &name, 
            const std::string &help,
            const std::vector<std::string> &enums,
            const int deflt = 0,
            const std::vector<int> * IDs = NULL,
            const char cmdshort = 0, 
            const int Group = GRP_NOSHOW );

        /**
        \brief Check if parameter has already been added.
        */
        bool CheckParam(const std::string &name);

        // routines for GETTING parameters
        int * GetParamInt( const std::string &name ) const;
        bool * GetParamBool( const std::string &name )const;
        std::string * GetParamString( const std::string &name )const;
        double * GetParamDouble( const std::string &name )const;
        BIAS::Vector<double> * GetParamVecDbl( const std::string &name )const;
        BIAS::Vector<int> * GetParamVecInt( const std::string &name )const;
        int * GetParamEnum( const std::string &name )const;

        // routines for SETTING parameters
        int * SetParamInt( const std::string &name, const int &value );
        bool * SetParamBool( const std::string &name, const bool &value );
        std::string * SetParamString( const std::string &name,
            const std::string &value );
        double * SetParamDouble( const std::string &name, const double &value );
        BIAS::Vector<double> * SetParamVecDbl( const std::string &name,
            const BIAS::Vector<double> &value );
        BIAS::Vector<int> * SetParamVecInt( const std::string &name,
            const BIAS::Vector<int> &value );
        /// set the enum to the ID 'value'
        int * SetParamEnum( const std::string &name, const int &value );
        /// set the enum to the ID corresponding to the string 'value'
        int * SetParamEnum( const std::string &name, const std::string &value );

        int SetParameterWriteToFile(const std::string &name, bool writeToFile);

        /** @brief if a parametr is marked as enhanced, it is hidden from the
        naive user */
        int SetEnhancedFlag(const std::string &name, bool enhanced);

        /** @brief if a parametr is marked as hidden, it is obsolete and 
        hidden for all users, use this for parameters overwritten by caller 
        class*/
        int SetHiddenFlag(const std::string &name, bool hidden);

        /// sets all entries to their initially (AddParam) given default values
        void SetDefaults();

        /** Set [Min,Max] as new range, especially usefull in derived class
        PramGUI */
        virtual int SetRangeInt(const std::string &name,const int min, 
            const int max);

        /// print all data in group grp including current values to os
        /// if grp = GRP_ALL, print all values
        /// if showenhanced = false, parameters marked as enhanced are not shown
        void ShowData(std::ostream &os = std::cout, int grp=GRP_ALL, 
            bool showenhanced = true);

        /// print Help-Information to stdout
        void Usage(std::ostream &os = std::cout);

        /** @brief update all arguments from command line and parameter file
        1. reads parameter from default file 
        2. if --readconfig is given on commandline, reads parameter 
        from this file
        3. evaluates parameter given on command line

        Command line arguments take precedence over arguments from parameter 
        file. Parameter from parameter file specified with --readconfig=fname
        take precedencor over parameter read from default_filename.
        If no parameterfile is found, the defaults are written to default_name.
        @author woelk 09/2004 */
        int UpdateParameter(int &argc, char *argv[], 
            const std::string &default_filename); 

        /// scan command line arguments for valid parameters
        /** argv is reordered according to getopt() all no recognized arguments 
        are at the end.
        Return value is optind from getopt(), the index of the next not 
        recognized argument. Sometimes, the function however 
        returns negative values. ?
        @todo document concept of return value of ParseCommandLine 
        Reads parameter from file if argument --readconfig is given. 
        @author evers */
        int ParseCommandLine(int &argc, char *argv[]); 

        /// read values for parameters from file
        int ReadParameter(const std::string &filename);

#ifdef BIAS_HAVE_XML2
        /// read values for parameters from xml-file
        int ReadParameterXML(const std::string &filename);
#endif

        /** @brief store complete set of parameter in file
        if write_comments is set, alo the help strings are written to the file
        if show_enhanced is set, also parameters marked as enhanced are written
        @author ?? */
        int WriteParameter(const std::string &filename);

#ifdef BIAS_HAVE_XML2
        int WriteParameterXML(const std::string &filename);
#endif

        /** writes complete command line including program name and 
        unknown command line options,
        no defaults, no params read from file */
        int WriteRunMe(const std::string &filename);

        /** Returns a vector of all group IDs (unique) in ascending order.
        Only parameters are with WriteToFile_ = true are considered.
        @author woelk 03 2004 */
        std::vector<int> UniqueSortedGroups();

        /** @brief returns unused group id 
        @author woelk */
        int GetFreeGroupID();

        /** @brief returns group id of parameter with name 
        @author woelk */
        int GetGroupID(const std::string &name);

        /** @brief 
        @author evers */
        int GetGroupIDForGroupName(const std::string &name);

        /** @brief sets the name for a group
        @author woelk */
        int SetGroupName(const int group_id, const std::string& name);

        /** gets the name of a group
        @author woelk */
        std::string GetGroupName(const int group_id);

        /** @brief returns if the group id is used
        @author woelk */
        bool IsUsedGroupID(const int group_id);

        /**
        @author evers */
        int RegisterCallbackObject(const std::string &name,
            ParamCallbackInterface *obj);

        /** Uses this just before end of your program to avoid error from 
        destructor.
        @author evers */
        void DisableDestructorWarning(){DestructorWarning_ = false;};

        /** Add a short commandline switch to an existing parameter.
        Useful for readconfig/writeconfig.
        @author evers */
        int SetShortCommand(const std::string &name, const char cmd);

        /** @brief returns the size of the longest param name not marked
        as enhanced or hidden
        @author woelk 10/2004 */
        int GetSizeOfLongestParamName();

        /** @brief returns the size of the longest param value string not marked
        as enhanced or hidden
        @author woelk 10/2004 */
        int GetSizeOfLongestParamValueString();

        void SetWriteOptions(bool indent, bool comments, bool enhanced)
        { 
            WriteIndent_ = indent; 
            WriteComments_ = comments; 
            WriteEnhanced_ = enhanced;
        }

  protected:
    std::map<const std::string,BIAS::ParamTypeBase*> & GetParamMap();
    const std::map<const std::string,BIAS::ParamTypeBase*> & GetParamMap() const;

    protected:
        std::vector<std::string> ArgV_;
        std::vector<std::string> GroupNames_;
        struct option *longopts;
        bool DestructorWarning_;
        // WriteOptions
        bool WriteIndent_;
        bool WriteComments_;
        bool WriteEnhanced_;

    private: // avoid DLL export
      /// this STL map holds the name-ParamType-pairs
      std::map<const std::string,BIAS::ParamTypeBase*> ParamMap_;

    };

} //namespace BIAS;

#include <Base/Common/BIASpragmaEnd.hh>


#endif

