/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be usefulen,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <Utils/TriangleMesh.hh>
#include <Base/Math/Vector.hh>
#include <Geometry/RMatrix.hh>
#include <Base/Common/CompareFloatingPoint.hh>
#include <Filter/Gauss.hh>
#include <Base/Common/FileHandling.hh>
#include <Base/Image/ImageIO.hh>
#include <Base/Geometry/HomgPlane3D.hh>
#include <Filter/Rescale.hh>

using namespace BIAS;
using namespace std;


TriangleMesh::TriangleMesh(const double& minCornerAngle,
                           const double& maxAngle,
                           const double depthScale) {
  SetParams (minCornerAngle, maxAngle, depthScale);
}

void TriangleMesh::
SetParams(const double& minCornerAngle,
          const double& maxAngle,
          const double depthScale) {
  minCornerAngle_ = minCornerAngle;
  maxAngle_ = maxAngle; 
  depthScale_ = depthScale;
}

/* @return plane unit 1 normal
   assuming RHS order of p1,p2,p3, e.g.
   p1 p3
   p2
   @author Jan Woetzel 06/2003
*/
inline BIAS::Vector<double> GetPlaneNormalEuclidian( const BIAS::Vector3<double> & p1,
                                              const BIAS::Vector3<double> & p2,
                                              const BIAS::Vector3<double> & p3)
{
  BIAS::Vector3<double> nvec; // normal unit 1 result vector
  nvec = (p2-p1).CrossProduct( p3-p1 );
  if(Equal(nvec.NormL2(), 0.0)) {
    BIASERR("normalvector could not be calculated");
  }
  else
    nvec.Normalize(); // to length 1;
  return nvec;
}


/* @return radian angle between the two direction vectors
   @author JW 06/2003
*/
inline double GetAngle(const BIAS::Vector3<double> & vec1,
                const BIAS::Vector3<double> & vec2)
{
  double cosAlpha = 
    (vec1.GetNormalized()).ScalarProduct( vec2.GetNormalized() );
  // we want an angle between 0 and Pi/2 rad (0..90 deg):
  if (cosAlpha<0) {
    return (M_PI - acos( cosAlpha ));
  } else {
    return  acos( cosAlpha );
  }
 
}


/** @return true if triangle(p1,p2,p3) corner angles are at least minPolyAngle
    @author JW 06/2003 
    assume a RHS of p1,p2,p3, e.g.:
    p1 p3
    p2
    @param minPolyAngle min. angle in rad of each polygion corner.
    sorts out 'stretched' triangles with acute angles (spitze Winkel)
*/
inline bool AcceptPolyCornerAngles( const BIAS::Vector3<double> & p1,
			     const BIAS::Vector3<double> & p2,
			     const BIAS::Vector3<double> & p3,
			     const double minPolyAngle=5.0/180.0*M_PI)
{
  // check min. poly angles of corner p1, p2, p3 in RHS order:
  if (GetAngle(p2-p1 , p3-p1) < minPolyAngle)
    return false; // angle at p1 too small
  else if (GetAngle(p3-p2 , p1-p2) < minPolyAngle)
    return false; // angle at p2 too small
  else if (GetAngle(p1-p3 , p2-p3) < minPolyAngle)
    return false; // angle at p3 too small
  else
    return true;
}


/** @return true if the angel between plane view ray 
    and plane normal is not too big.
    sorts out closing of object discontinuities/occlusions.
    @param p1,p2,p3 triangle corners in RHS order
    @param C virtual view camera center
    @param maxAngle in rad between plane normal and ray of view, e.g. 89(deg).*(M_PI/180.0)
    @author JW 06/2003 */
inline bool AcceptPolyPlanViewAngle(const BIAS::Vector3<double> & p1,
                             const BIAS::Vector3<double> & p2,
                             const BIAS::Vector3<double> & p3,
                             const BIAS::Vector3<double> & C,
                             const double maxAngle=89.0/180.0*M_PI )
{ 
  // check angle between plane normal and ray of view from C to nw:
  // view on p1 (may be any vertex)
  double angle = GetAngle( p1-C, GetPlaneNormalEuclidian(p1, p2, p3) );
  return (angle<=maxAngle);
}


int TriangleMesh::
GenerateDenseMesh(const BIAS::Image<float>& DenseDepthMap,
                  const BIAS::Image<unsigned char>& Texture)
{
  //   if( (DenseDepthMap.GetWidth()!=Texture.GetWidth()) ||
  //       (DenseDepthMap.GetHeight()!=Texture.GetHeight()) ) {
  //     BIASERR("Texturesize != DepthMapsize \n");
  //     return ;
  //   }    
  //init datastructs  
  unsigned int w=DenseDepthMap.GetWidth();
  unsigned int h=DenseDepthMap.GetHeight();
  vector<int> pointIndices;
  pointIndices.resize(w*h);

  meshVertices_.clear();
  vertexNormals_.clear();
  textureCoords_.clear();
  triangleIndices_.clear();

  meshVertices_.reserve(w*h);
  vertexNormals_.reserve(w*h);
  textureCoords_.reserve(w*h);
  triangleIndices_.reserve(w*h*2);

  texture_ = Texture;

  //build list of indices and list of Vertices
  float texCoordOffsetX = 1.0f / float(w-1);
  float texCoordOffsetY = 1.0f / float(h-1);
  Vector2<float> texCoord;
  texCoord.Set(0.0, 1.0);
  
  Vector3<double> vertex;
  HomgPoint3D vertexHom;
    
  float** depthValues = DenseDepthMap.GetImageDataArray();
  unsigned int currentIndex=0;
  unsigned int x,y;
  for( y=0; y<h; y++) {
    for( x=0; x<w; x++) {
      if(depthValues[y][x]<= 0.0) {
        pointIndices[y*w+x] = -1;
      }
      else {
        pointIndices[y*w+x] = currentIndex;
        vertexHom = HomgPoint3D((double)x,(double)y,
                                depthScale_ * (double)depthValues[y][x]);
        vertexHom.GetEuclidian(vertex);
        meshVertices_.push_back(vertex);       
        textureCoords_.push_back(texCoord);
        currentIndex++;
      }
      texCoord[0] += texCoordOffsetX;
    } 
    texCoord[0] = 0.0;
    texCoord[1] -= texCoordOffsetY;
  }
  
  //meshing
  Vector3<int> triangle;
  int index0, index1, index2;
  for( y=0; y<h; y++) {
    for( x=0; x<w-1; x++) {
      /*         v1
       *      i0---i1
       *       |   /
       *       |  / v0
       *       i2
       */
      if(y<h-1) {
        index0 = pointIndices[y*w+x];
        index1 = pointIndices[y*w+x+1];
        index2 = pointIndices[(y+1)*w+x];
        if(index0 > (-1) && index1 > (-1) && index2 > (-1)){
          triangle.Set(index2, index1, index0);
          if (AcceptPolyCornerAngles(meshVertices_[index0],
                                     meshVertices_[index1],
                                     meshVertices_[index2], minCornerAngle_))
            triangleIndices_.push_back(triangle);
        }
      }    
      /*
       *         i0 
       *        / |
       *       /  |
       *      i2--i1
       */
      if(y>0) {
        index2 = pointIndices[y*w+x];
        index1 = pointIndices[y*w+x+1];
        index0 = pointIndices[(y-1)*w+x+1];
        if(index0 > (-1) && index1 > (-1) && index2 > (-1)) {
          triangle.Set(index2, index1, index0);
          if (AcceptPolyCornerAngles(meshVertices_[index0],
                                     meshVertices_[index1],
                                     meshVertices_[index2], minCornerAngle_))
            triangleIndices_.push_back(triangle);
        }
      }
    }

    //   cout<<"vertices "<<meshVertices_.size()<<endl;
    //   cout<<"texCoords "<<textureCoords_.size()<<endl;
    //   cout<<"triangles "<<triangleIndices_.size()<<endl;

  }
  
  if (triangleIndices_.empty()) return -1;
  return 0;

}

int TriangleMesh::
GenerateDenseMesh(const BIAS::Image<float>& DenseDepthMap,
                  BIAS::PMatrix& P,
                  const BIAS::Image<unsigned char>& Texture)
{
  ProjectionParametersPerspective ppp;
  ppp.SetImageSize(Texture.GetWidth(), Texture.GetHeight());
  ppp.SetP(P);
  Projection pP(ppp);
  return GenerateDenseMesh(DenseDepthMap, pP, Texture);  
}

int TriangleMesh::
GenerateDenseMesh(const BIAS::Image<float>& DenseDepthMap,
                  BIAS::Projection& P,
                  const BIAS::Image<unsigned char>& Texture, 
                  const unsigned int distBetweenPoints)
{
  unsigned int w = DenseDepthMap.GetWidth();
  unsigned int h = DenseDepthMap.GetHeight();

  return GenerateDenseMesh(DenseDepthMap,
                           P,
                           Texture,
                           0,0,w,h,
                           distBetweenPoints);
}


Vector2<float> ImageToTexCoordinates(float x0, float y0, unsigned int w,
                                     unsigned int h){
  float texCoordOffsetX = 1.0f / float(w-1);
  float texCoordOffsetY = 1.0f / float(h-1);
  Vector2<float> texCoord;
  texCoord.Set(x0 * texCoordOffsetX, 1.0f - y0 * texCoordOffsetY);
  return texCoord;
}




int TriangleMesh::
GenerateDenseMesh(const BIAS::Image<float>& DenseDepthMap,
                  BIAS::Projection& P,
                  const BIAS::Image<unsigned char>& Texture,
                  const unsigned int x0, const unsigned int y0,
                  const unsigned int width, const unsigned int height, 
                  const unsigned int distBetweenPoints)
{
  Vector3<double> C = P.GetC();
  if( (DenseDepthMap.GetWidth()!=Texture.GetWidth()) ||
    (DenseDepthMap.GetHeight()!=Texture.GetHeight()) ) {
    BIASWARNONCE("Texturesize "<<Texture.GetWidth()<<"x"<<Texture.GetHeight()
                 <<"!= DepthMapsize "<<DenseDepthMap.GetWidth()<<"x"
                 <<DenseDepthMap.GetHeight()<<endl);
  }    
  if(width<=0 || height <=0) {
    BIASERR("tile width is too small!");
    return -1;
  }

    //init datastructs  
    unsigned int w=DenseDepthMap.GetWidth();
    unsigned int h=DenseDepthMap.GetHeight();
    vector<int> pointIndices;
    pointIndices.resize(width*height);

    meshVertices_.clear();
    vertexNormals_.clear();
    textureCoords_.clear();
    triangleIndices_.clear();

    meshVertices_.reserve(width*height);
    vertexNormals_.reserve(width*height);
    textureCoords_.reserve(width*height);
    triangleIndices_.reserve(width*height*2);

    texture_ = Texture;

    //build list of indices and list of Vertices
    float texCoordOffsetX = 1.0f / float(w-1);
    float texCoordOffsetY = 1.0f / float(h-1);
    Vector2<float> texCoord;
    
    texCoord.Set(x0 * texCoordOffsetX, 1.0f - y0 * texCoordOffsetY);

    Vector3<double> vertex;

    float** depthValues = DenseDepthMap.GetImageDataArray();
    unsigned int currentIndex=0;
    unsigned int x,y;
    for( y=0; y<height; y++) {
      for( x=0; x<width; x++) {
        if(depthValues[y0+y][x0+x]< 1e-5) {
          pointIndices[y*width+x] = -1;
        }
        else {
          //          BIASASSERT(depthValues[y][x]>1e-5);
          vertex = P.UnProjectToPoint(HomgPoint2D(x+x0,y+y0), 
                                      depthScale_ * depthValues[y+y0][x0+x]);
          BIASASSERT((vertex-P.GetC()).NormL2()>1e-10)
          if (vertex.NormL2()>1e-10) {
            meshVertices_.push_back(vertex);       
            textureCoords_.push_back(texCoord);
            pointIndices[y*width+x] = currentIndex;
            currentIndex++;
          } else {
            pointIndices[y*width+x] = -1;
            //cout<<"***vertex invalid***"<<currentIndex<<" "<<endl;
          }
        }
        BIASASSERT(meshVertices_.size()==currentIndex);
        texCoord[0] += texCoordOffsetX;
        if (!meshVertices_.empty()) {
          BIASASSERT(meshVertices_.back().NormL2()>1e-10);
        }
      } 
      texCoord[0] = x0 * texCoordOffsetX;
      texCoord[1] -= texCoordOffsetY;
    }
    //cout<<"size of vertex array is "<<meshVertices_.size()<<endl;
    //meshing
    Vector3<int> triangle;
    int index0, index1, index2;
    int rejected = 0;
    for( y=0; y<height; y+=distBetweenPoints) {
      for( x=0; x<width-1; x+=distBetweenPoints) {
        /*         v1
        *      i0---i1
        *       |   /
        *       |  / v0
        *       i2
        */
        if(y<height-1) {
          index0 = pointIndices[y*width+x];
          index1 = pointIndices[y*width+x+1];
          index2 = pointIndices[(y+1)*width+x];
          if(index0 > (-1) && index1 > (-1) && index2 > (-1)){ 
            
            //BIASASSERT(index1<meshVertices_.size());
            //BIASASSERT(index2<meshVertices_.size());
            //BIASASSERT(index0<meshVertices_.size());

            //BIASASSERT((meshVertices_[index0]-meshVertices_[index1]).NormL2()>1e-10);
            //BIASASSERT((meshVertices_[index2]-meshVertices_[index1]).NormL2()>1e-10);
            //BIASASSERT((meshVertices_[index0]-meshVertices_[index2]).NormL2()>1e-10);
            //BIASASSERT(meshVertices_[index0].NormL2()>1e-10);
            //BIASASSERT(meshVertices_[index1].NormL2()>1e-10);
            //BIASASSERT(meshVertices_[index2].NormL2()>1e-10);

            triangle.Set(index2, index1, index0);
            //	  triangleIndices_.push_back(Vector<unsigned int>(triangle));
             if (AcceptPolyCornerAngles(meshVertices_[index0],
                                       meshVertices_[index1],
                                       meshVertices_[index2], minCornerAngle_)
                && AcceptPolyPlanViewAngle(meshVertices_[index0],
                                           meshVertices_[index1],
                                           meshVertices_[index2], C,
                                           maxAngle_)) {
               triangleIndices_.push_back(triangle);
             }
             else { rejected++; }
          }
        }    
        /*
        *         i0 
        *        / |
        *       /  |
        *      i2--i1
        */
        if(y>0) {
          index2 = pointIndices[y*width+x];
          index1 = pointIndices[y*width+x+1];
          index0 = pointIndices[(y-1)*width+x+1];
          if(index0 > (-1) && index1 > (-1) && index2 > (-1)) {
            triangle.Set(index2, index1, index0);
            //	  triangleIndices_.push_back(Vector<unsigned int>(triangle));
             if (AcceptPolyCornerAngles(meshVertices_[index0],
                                       meshVertices_[index1],
                                       meshVertices_[index2], minCornerAngle_)
                && AcceptPolyPlanViewAngle(meshVertices_[index0],
                                           meshVertices_[index1],
                                           meshVertices_[index2], C,
                                           maxAngle_)) {
               triangleIndices_.push_back(triangle);
             }
             else { rejected++; }
          }
        }
      }
    }

//      cout<<"vertices "<<meshVertices_.size()<<endl;
//      cout<<"texCoords "<<textureCoords_.size()<<endl;
//      cout<<"triangles "<<triangleIndices_.size()<<endl;
//      cout<<"rejected "<<rejected<<" triangles due to invalid angles\n";
    if (triangleIndices_.empty()) {
      BIASERR("no triangles generated");
      return -1;
    }
    return 0;
}

int TriangleMesh::GenerateDenseMesh(const Image<float>& DenseDepthMapC,
                                     PMatrix& P,
                                     const Image<unsigned char>& TextureC,
                                     const ROI& CutOut) 
{
  // dirty de-const cast to conveniently access ROI, which is restored later
  Image<float>& DenseDepthMap(const_cast<Image<float>& >(DenseDepthMapC));
  Image<unsigned char>& Texture(const_cast<Image<unsigned char>& >(TextureC));

  int UpperLeftX, UpperLeftY, LowerRightX, LowerRightY;
  CutOut.GetCorners(UpperLeftX, UpperLeftY, LowerRightX, LowerRightY);

  // adapt pmatrix principal point
  PMatrix PCutOut(P);
  if (PCutOut.IsMetric()) {
    //  more exact to use saved K,R,C and recompose than multiplying
    KMatrix K = PCutOut.GetK();
    K[0][2] -= double(UpperLeftX);
    K[1][2] -= double(UpperLeftY);
    RMatrix R =  PCutOut.GetR();
    Vector3<double> C(PCutOut.GetC());
    PCutOut.Compose(K, R, C);
  } else { 
    // non-metrix case, just multiply with correction matrix
    KMatrix K(MatrixIdentity);
    K[0][2] -= double(UpperLeftX);
    K[1][2] -= double(UpperLeftY);
    PCutOut  = K * PCutOut;

    // tell PMatrix that a basic matrix operation has been invoked
    // and chache is invalid
    PCutOut.InvalidateDecomposition();
  }
  // PCutout is now the new projection matrix for the cutout

  // get CutOut from texture
  ROI backupROI = *Texture.GetROI();
  Texture.SetROICorners(UpperLeftX, UpperLeftY, LowerRightX, LowerRightY);
  Image<unsigned char> CutOutTexture;
  Texture.GetCopyOfROI(CutOutTexture);
  *Texture.GetROI() = backupROI;

  // get CutOut from depth map
  backupROI = *DenseDepthMap.GetROI();
  DenseDepthMap.SetROICorners(UpperLeftX, UpperLeftY, 
    LowerRightX, LowerRightY);
  Image<float> CutOutDenseDepthMap;
  DenseDepthMap.GetCopyOfROI(CutOutDenseDepthMap);
  *DenseDepthMap.GetROI() = backupROI;

  // call generic interface
  return GenerateDenseMesh(CutOutDenseDepthMap, PCutOut, CutOutTexture);
}


int TriangleMesh::GenerateImagePlane(BIAS::Projection& P,
                                     const BIAS::Image<unsigned char>& Texture,
                                     const double& w) {
  if (typeid(*P.GetParameters())!=typeid(ProjectionParametersPerspective)) {
    BIASWARN("Interpolation of texture coordinates not yet implemented !");
  }
  HomgPoint2D x(0.0,0.0);
  Vector2<float> texcoord(0.0,1.0);
  for (unsigned int i=0; i<4; i++) {
    switch (i) {
    case 1: {x = HomgPoint2D(Texture.GetWidth()-1, 0); 
    texcoord =  Vector2<float>(1.0, 1.0);break;}
    case 2: {x = HomgPoint2D(0, Texture.GetHeight()-1);
    texcoord =  Vector2<float>(0.0, 0.0); break;}
    case 3: {x = HomgPoint2D(Texture.GetWidth()-1,
                             Texture.GetHeight()-1); 
    texcoord =  Vector2<float>(1.0, 0.0);break;}
    }
    Vector3<double> eucl;
    P.GetParameters()->UnProjectToImagePlane(x).GetEuclidian(eucl);
    meshVertices_.push_back(eucl);   
    textureCoords_.push_back(texcoord);
  }
  Vector3<int> triangle;
  // two triangles
  triangle.Set(0, 1, 2);
  triangleIndices_.push_back(triangle);
  triangle.Set(1, 2, 3);
  triangleIndices_.push_back(triangle);
  texture_ = Texture;
  
   
    return 0;
}


void TriangleMesh::GenerateTexturedCamera(ProjectionParametersBase* rectPPB, 
                                          Image<unsigned char>& rectB,
                                          const double& scale,
                                          const double& opacity,
                                          const double& resolution) {

  ProjectionParametersBase* newparams = rectPPB->Clone();
  if (resolution!=1.0) newparams->Rescale(float(resolution));

  BIASASSERT(opacity<=1.0)
  unsigned int width = 0, height = 0;
  newparams->GetImageSize(width, height);
  Image<float>  depthmap(width, height, 1);
  depthmap.SetZero();
  for (unsigned int y=0; y<height; y++) {
    for (unsigned int x=0; x<width; x++) {
      HomgPoint2D p2(x,y);
      HomgPoint3D p3 = newparams->UnProjectToImagePlane(p2, scale);
      if (p3.NormL2()<0.1) continue;
      depthmap.GetImageDataArray()[y][x]  = (float)
        (p3 - HomgPoint3D(rectPPB->GetC())).NormL2();
      
    }
  }

  BIASASSERT(rectB.GetChannelCount()==3);
  Rescale<unsigned char, unsigned char> R;
  Image<unsigned char> rect(width, height, rectB.GetChannelCount());
  R.SetFactor(resolution);
  R.Filter(rectB, rect);

  
  Projection Proj(*newparams);

  Image<unsigned char> rectAlpha(rect.GetWidth(),
                                  rect.GetHeight(),4, true);

  rectAlpha.SetColorModel(ImageBase::CM_RGBA);
  for (unsigned int y=0; y<rectAlpha.GetHeight(); y++) {
    for (unsigned int x=0; x<rectAlpha.GetWidth(); x++) {
      unsigned char *pD = &rect.GetImageDataArray()[y][3*x];
      unsigned char *pDA = &rectAlpha.GetImageDataArray()[y][4*x];
      *pDA++ = *pD++;
      *pDA++ = *pD++;
      *pDA++ = *pD++;
      *pDA++ = (unsigned char)(255.0*opacity);
    }
  }

  GenerateDenseMesh(depthmap, Proj, rectAlpha);
  delete newparams;
}


/* take 5 points: 4 corners + center, calculate normals for 4 triangles,
   compare normals and if one angle is above threshold, refine recursivly.
If depth values of one point is missing, refine, too
*/
bool TriangleMesh::MakeTriangles_(const Projection& P, 
                                  std::vector<Image<float> > & depthmaps,
                                  float x,float y, float offset, 
                                  float minoffset,
                                  double minCosAngle, double maxDiagRatio,
                                  bool simulateOnly)
{ 

  const int maxlayer = depthmaps.size()-1;
  int step = int(rint(log(offset) / log(2.0))) -2;
  if (step>maxlayer) step = maxlayer;
  if (step<0) step = 0;
  
  Image<float> &depthmap = depthmaps[step];
  HomgPoint2D S[5];
  Vector3<double> qp[5];
  Vector3<double> normal[4];
  bool vetoed = true;
  unsigned int FoundDepthValues = 0;
  unsigned int c;
  float** depthValues = depthmap.GetImageDataArray();
  for (c=0; c<5;c++){
    S[c][2] = 1.0;
    switch (c){
    case 0:
      S[c][0] = x;
      S[c][1] = y;
      break;
    case 1:
      S[c][0] = x-offset;
      S[c][1] = y-offset;
      break;
    case 2:
      S[c][0] = x-offset;
      S[c][1] = y+offset;
      break;
    case 3:
      S[c][0] = x+offset;
      S[c][1] = y+offset;
      break;
    case 4:
      S[c][0] = x+offset;
      S[c][1] = y-offset;
      break;
    } 
    double depth = depthValues[int(S[c][1])][int(S[c][0])];
    if (depth<1e-10) {
      // invalid depth
      break;
    }

    HomgPoint3D backProjected;
    backProjected = 
      P.GetParameters()->UnProjectToPoint(S[c], depthScale_ * 
                                      depthValues[int(S[c][1])][int(S[c][0])]);
    if (backProjected.NormL2()<1e-10) {
      // point is not backprojectable, e.g. out of sphere circle
      break;
      //cout<<"***vertex invalid***"<<currentIndex<<" "<<endl;
    }
    backProjected.GetEuclidian(qp[c]);
    FoundDepthValues++;
    // cout <<qp[c][2]<<"  ";
  }
  // check if loop terminated normaly or break was called
  double cosang[6];
  double diag1 ;
  double diag2, diagratio;

  Vector3<double> meannormal(0,0,0);
  Vector3<double> LoS;
  // add all single-pixel quads
  if (FoundDepthValues==5) {
    vetoed = false;
    // calculate 4 normals
    (qp[1]-qp[0]).CrossProduct((qp[2]-qp[0]),normal[0]);
    (qp[2]-qp[0]).CrossProduct((qp[3]-qp[0]),normal[1]);
    (qp[3]-qp[0]).CrossProduct((qp[4]-qp[0]),normal[2]);
    (qp[4]-qp[0]).CrossProduct((qp[1]-qp[0]),normal[3]);
    for (unsigned int i=0;i<4;i++) {
      normal[i].Normalize();
      meannormal += normal[i];
    }
    meannormal /= 4.0;

    // compare normals

    normal[0].ScalarProduct(normal[1],cosang[0]);
    normal[0].ScalarProduct(normal[2],cosang[1]);
    normal[0].ScalarProduct(normal[3],cosang[2]);
    normal[1].ScalarProduct(normal[2],cosang[3]);
    normal[1].ScalarProduct(normal[3],cosang[4]);
    normal[2].ScalarProduct(normal[3],cosang[5]);
    for (unsigned int c=0; c<6;c++) {
      if (cosang[c] <minCosAngle) vetoed = true; 
    }
    // 25 deg: 0.9, 18 deg: 0.95, 8 deg: 0.96 5 deg: 0.996 
    diag1 = (qp[1]-qp[3]).NormL2();
    diag2 = (qp[2]-qp[4]).NormL2();
    diagratio = diag1>diag2 ? diag1/diag2 : diag2/diag1;
    if (diagratio > maxDiagRatio) {
      vetoed = true;
    }
    LoS = qp[0] - P.GetC();
    LoS.Normalize();
    if (GetAngle(meannormal,LoS)>maxAngle_) {
      // too bad viewing angle
      FoundDepthValues = 0; // reject even if smallest size !
    }
  }
  if ((FoundDepthValues == 5) && ((!vetoed) || offset<=minoffset)){

    
      
    float newx,newy,newoff;
    newoff = offset / 2;
    newx = x-newoff;
    newy = y-newoff;

    if ((newoff<=minoffset) && simulateOnly) return true;
   

    bool result =  true; 

    MakeTriangles_(P,depthmaps,newx,newy,newoff ,minoffset, 
                   minCosAngle, maxDiagRatio, true);
    newy = y+newoff;
    result &= MakeTriangles_(P,depthmaps,newx,newy,newoff ,minoffset, 
                             minCosAngle, maxDiagRatio, true);
    newx = x+newoff;
    result &= MakeTriangles_(P,depthmaps,newx,newy,newoff ,minoffset, 
                             minCosAngle, maxDiagRatio, true);
    newy = y-newoff;
    result &= MakeTriangles_(P,depthmaps,newx,newy,newoff ,minoffset, 
                             minCosAngle, maxDiagRatio, true);
    
    if (simulateOnly) return result;

    if (result) {

      //cout<<"allow at "<<newoff<<endl;
    }


    if (result || (newoff<=minoffset)) {
      // we are allowed to place the quad !
      // add this quad to vertarray
      for (unsigned int c=1; c<5;c++) {
        meshVertices_.push_back(qp[c]);       
        textureCoords_.push_back(ImageToTexCoordinates(float(S[c][0]),
                                                       float(S[c][1]),
                                                       depthmap.GetWidth(),
                                                       depthmap.GetHeight()));
        currentIndex_++;
      }
      Vector3<int> triangle;
      triangle.Set(currentIndex_-4, currentIndex_-3, currentIndex_-2);
      triangleIndices_.push_back(triangle);
      triangle.Set(currentIndex_-2, currentIndex_-1, currentIndex_-4);
      triangleIndices_.push_back(triangle);


      HomgPlane3D thePlaneA;
      thePlaneA.SetFromPoints(HomgPoint3D(qp[1]),
                              HomgPoint3D(qp[2]), HomgPoint3D(qp[3]));
      int x=int(S[0][0]-offset);
      int y = 0;
      for (y=int(S[0][1]-offset); y<=int(S[0][1]+offset); y++) {
        HomgPoint3D Ray(P.UnProjectToRay(HomgPoint2D(x,y)));
        Ray[3] = 0;
        HomgPoint3D Intersection(0,0,0,0);
        thePlaneA.GetLineIntersection(HomgPoint3D(P.GetC()),Ray,Intersection);
        if (depthmaps[0].IsPositionInImage(x,y))
          for (unsigned int i=0; i<depthmaps.size(); i++) {
            depthmaps[i].GetImageDataArray()[y][x] = 
              float((HomgPoint3D(P.GetC())-Intersection).NormL2());
          }
      }
      y--;
      for (x=int(S[0][0]-offset); x<=int(S[0][0]+offset); x++) {
        HomgPoint3D Ray(P.UnProjectToRay(HomgPoint2D(x,y)));
        Ray[3] = 0;
        HomgPoint3D Intersection(0,0,0,0);
        thePlaneA.GetLineIntersection(HomgPoint3D(P.GetC()),Ray,Intersection);
        if (depthmaps[0].IsPositionInImage(x,y))
          for (unsigned int i=0; i<depthmaps.size(); i++) {
            depthmaps[i].GetImageDataArray()[y][x] = 
              float((HomgPoint3D(P.GetC())-Intersection).NormL2());
          }
      }
      x--;
      thePlaneA.SetFromPoints(HomgPoint3D(qp[3]),
                              HomgPoint3D(qp[4]), HomgPoint3D(qp[1]));
      for (; y>=int(S[0][1]-offset); y--) {
        HomgPoint3D Ray(P.UnProjectToRay(HomgPoint2D(x,y)));
        Ray[3] = 0;
        HomgPoint3D Intersection(0,0,0,0);
        thePlaneA.GetLineIntersection(HomgPoint3D(P.GetC()),Ray,Intersection);
        if (depthmaps[0].IsPositionInImage(x,y))
          for (unsigned int i=0; i<depthmaps.size(); i++) {
            depthmaps[i].GetImageDataArray()[y][x] = 
              float((HomgPoint3D(P.GetC())-Intersection).NormL2());
          }
      }
      y++;
      for (; x>=int(S[0][0]-offset); x--) {
        HomgPoint3D Ray(P.UnProjectToRay(HomgPoint2D(x,y)));
        Ray[3] = 0;
        HomgPoint3D Intersection(0,0,0,0);
        thePlaneA.GetLineIntersection(HomgPoint3D(P.GetC()),Ray,Intersection);
        if (depthmaps[0].IsPositionInImage(x,y))
          for (unsigned int i=0; i<depthmaps.size(); i++) {
            depthmaps[i].GetImageDataArray()[y][x] = 
              float((HomgPoint3D(P.GetC())-Intersection).NormL2());
          }
      }
      return true;
    }
  } 
  if (offset>(minoffset)){
    if (simulateOnly) return false;
    float newx,newy,newoff;
    newoff = offset / 2;
    newx = x-newoff;
    newy = y-newoff;

    
    MakeTriangles_(P,depthmaps,newx,newy,newoff ,minoffset, 
                   minCosAngle, maxDiagRatio, false);
    newy = y+newoff;
    MakeTriangles_(P,depthmaps,newx,newy,newoff ,minoffset, 
                   minCosAngle, maxDiagRatio, false);
    newx = x+newoff;
    MakeTriangles_(P,depthmaps,newx,newy,newoff ,minoffset, 
                   minCosAngle, maxDiagRatio, false);
    newy = y-newoff;
    MakeTriangles_(P,depthmaps,newx,newy,newoff ,minoffset, 
                   minCosAngle, maxDiagRatio, false);

    return false;
  }
  return true;
}

int TriangleMesh::
GenerateSimplifiedMesh(const BIAS::Image<float>& DenseDepthMap,
                       const BIAS::Projection& P,
                       const BIAS::Image<unsigned char>& Texture,
                       const unsigned int x0, const unsigned int y0,
                       const unsigned int width, const unsigned int height,
                       double minCosAngle,
                       double maxDiagRatio)
{
  currentIndex_ = 0;
  Vector3<double> C = P.GetC();
  if( (DenseDepthMap.GetWidth()!=Texture.GetWidth()) ||
      (DenseDepthMap.GetHeight()!=Texture.GetHeight()) ) {
    BIASWARN("Texturesize != DepthMapsize \n");
  }    
  if(width<=0 || height <=0) {
    BIASERR("tile width is too small!");
    return -1;
  }



  vector<Image<float> > depthimages;

  int maxlevel = int(log(double(DenseDepthMap.GetWidth()))/log(2.0)-3);
  int spacing = int(pow(2.0, maxlevel));
  cout<<"maximum level is "<<maxlevel<<" spacing is "<< spacing<<endl;
  depthimages.reserve(maxlevel+1);
  depthimages.push_back(DenseDepthMap);


  Gauss<float, float> DepthSmooth; 
  bool smoothOriginal = true;
  if (smoothOriginal) {
    DepthSmooth.SetSigma(1.0); 
    depthimages.back().SetZero();
    DepthSmooth.Filter7x7GreyIgnoreBelowThreshold(DenseDepthMap,
                                                  depthimages.back(), 
                                                  0.0f);
  }


 
  double curSigma = 0.5;
  for (int curLevel = 0; curLevel<maxlevel; curLevel++) {
    double neededSigma = sqrt(pow(2.0, double(2*curLevel))-curSigma*curSigma);
    cout<<"Filtering with sigma "<<neededSigma<<endl;
    curSigma = pow(2.0, double(curLevel));
    DepthSmooth.SetSigma(neededSigma); 
    depthimages.push_back(depthimages.back());
    depthimages.back().SetZero();
    DepthSmooth.Filter7x7GreyIgnoreBelowThreshold(*(depthimages.rbegin()+1),
                                                  depthimages.back(), 
                                                  0.0f);
  }
  for (int curLevel = 0; curLevel<maxlevel; curLevel++) {
    ImageIO::Write("depth"+toString(curLevel),depthimages[curLevel]);
  }

  meshVertices_.clear();
  vertexNormals_.clear();
  textureCoords_.clear();
  triangleIndices_.clear();

  meshVertices_.reserve(width*height);
  vertexNormals_.reserve(width*height);
  textureCoords_.reserve(width*height);
  triangleIndices_.reserve(width*height*2);

  texture_ = Texture;
    
  unsigned int x,y;
  int minoffset = 2;
  for( y=spacing+1; y<height-spacing; y+=spacing) {
    for( x=spacing+1; x<width-spacing; x+=spacing) {
      MakeTriangles_(P, depthimages, float(x), float(y), 
                     float(spacing), float(minoffset), 
                     minCosAngle, maxDiagRatio, false);     
    }
  }
  
  bool multiPass=true;
  if (multiPass) {
    for (int curLevel = 2; curLevel<maxlevel; curLevel++) {
      cout<<"pass "<<curLevel<<" of "<<maxlevel-1<<" ..."<<flush<<endl;
      meshVertices_.clear();
      vertexNormals_.clear();
      textureCoords_.clear();
      triangleIndices_.clear();
      currentIndex_ = 0;
      // second path with consistent depth maps
      for( y=spacing+1; y<height-spacing; y+=spacing) {
        for( x=spacing+1; x<width-spacing; x+=spacing) {
          MakeTriangles_(P, depthimages, float(x), float(y),
                         float(spacing),float(minoffset),
                         minCosAngle, maxDiagRatio, false);     
        }
      }
      
    }
  }

  //      cout<<"vertices "<<meshVertices_.size()<<endl;
  //      cout<<"texCoords "<<textureCoords_.size()<<endl;
  //      cout<<"triangles "<<triangleIndices_.size()<<endl;
  //      cout<<"rejected "<<rejected<<" triangles due to invalid angles\n";
  if (triangleIndices_.empty()) {
    BIASERR("no triangles generated");
    return -1;
  }
  return 0;
}


int TriangleMesh::
MatchWithExistingMesh(TriangleMesh &mesh){

  int ret =0;

 


  return ret;
}
