/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
Multimediale Systeme der Informationsverarbeitung
Institut fuer Informatik
Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#ifndef _TRIANGLEMESH_HH_
#define _TRIANGLEMESH_HH_

#include <Base/Common/W32Compat.hh>
#include <math.h>
#include <vector>
#include <Base/Image/Image.hh>
#include <Geometry/PMatrix.hh>
#include <Geometry/Projection.hh>

namespace BIAS {

  /** \class TriangleMesh 
   * \author bartczak
   * \brief create and represent a 3D triangle mesh
   */
  class BIASUtils_EXPORT TriangleMesh {
  public:
    TriangleMesh(const double& minCornerAngle = M_PI * 5.0 / 180.0, 
                 const double& maxAngle = M_PI * 60.0 / 180.0,
                 const double depthScale = 1.0);

    ~TriangleMesh() {};

    void SetParams(const double& minCornerAngle = M_PI * 5.0 / 180.0, 
                   const double& maxAngle = M_PI * 60.0 / 180.0,
                   const double depthScale = 1.0);

    void SetMesh(BIAS::Image<unsigned char> texture,
                 std::vector<BIAS::Vector3<double> > meshVertices,
                 std::vector<BIAS::Vector3<double> > vertexNormals,
                 std::vector<BIAS::Vector2<float> > textureCoords,
                 std::vector<BIAS::Vector<int> > triangleIndices){
      texture_ = texture;
      meshVertices_ = meshVertices;
      vertexNormals_ = vertexNormals;
      textureCoords_ = textureCoords;
      triangleIndices_ = triangleIndices;
    };  
    
    /** Calculate a triangle mesh from dense depth map without PMatrix.
     * That's a 3D representation of an Image.
     */
    int GenerateDenseMesh(const BIAS::Image<float>& DenseDepthMap,
                          const BIAS::Image<unsigned char>& Texture);
    
    /** Calculate a triangle mesh from dense depth map.
     * No triangles are discarded, hence no smoothing.
     */
    int GenerateDenseMesh(const BIAS::Image<float>& DenseDepthMap,
                          BIAS::PMatrix& P,
                          const BIAS::Image<unsigned char>& Texture);

    /** @brief Calculate a triangle mesh only using the given cutout

    recalculates depthmap, image and p and calls generic interface
    @author koeser  */
    int GenerateDenseMesh(const BIAS::Image<float>& DenseDepthMap,
                          BIAS::PMatrix& P,
                          const BIAS::Image<unsigned char>& Texture,
                          const ROI& cutout);

    /** Calculate a triangle mesh from dense depth map.
     * No triangles are discarded, hence no smoothing.
     */
    int GenerateDenseMesh(const BIAS::Image<float>& DenseDepthMap,
                          BIAS::Projection& P,
                          const BIAS::Image<unsigned char>& Texture,
                           const unsigned int distBetweenPoints=1);

    /** calculates trianglemesh but only for given part of depth map! 
     *  Generated texture coordinates reference are relative to the
     *  full depth map size!
     */
    int GenerateDenseMesh(const BIAS::Image<float>& DenseDepthMap,
                          BIAS::Projection& P,
                          const BIAS::Image<unsigned char>& Texture,
                          const unsigned int x0, const unsigned int y0,
                          const unsigned int width, const unsigned int height,
                          const unsigned int distBetweenPoints=1);

    /** backproject image onto a plane in space
        so far only implemented for perspective cameras */
    int GenerateImagePlane(BIAS::Projection& P,
                           const BIAS::Image<unsigned char>& Texture,
                           const double& w = 1.0);

    /** generates the sensor plane / cylinder / sphere in 3D space
        @author koeser
        @param texture orignal image, must be rgb !!!
    */
    void GenerateTexturedCamera(ProjectionParametersBase* PPB, 
                                Image<unsigned char>& rgbtexture,
                                const double& scale = 1.0,
                                const double& opacity = 1.0,
                                const double& resolution = 1.0);


    void GetTexture(Image<unsigned char>& img) const {
      img = texture_;
    } 

    void GetVertices(std::vector<BIAS::Vector3<double> >& vertices) const {
      vertices = meshVertices_;
    } 

    void GetTexCoords(std::vector<BIAS::Vector2<float> >& texCoo) const {
      texCoo = textureCoords_;
    } 

    void GetTriangleIndices(std::vector<BIAS::Vector<int> >& triangleIndices) const {
      triangleIndices = triangleIndices_;
    } 

    void GetNormals(std::vector<BIAS::Vector3<double> >& normals) const 
    {
      normals = vertexNormals_;
    }

    int GenerateSimplifiedMesh(const Image<float>& DenseDepthMap,
                               const Projection& P,
                               const Image<unsigned char>& Texture,
                               const unsigned int x0, const unsigned int y0,
                               const unsigned int width, 
                               const unsigned int height,  
                               double minCosAngle = 0.8,
                               double maxDiagRatio = 3.0);

    int MatchWithExistingMesh(TriangleMesh &mesh);

  protected:
    bool MakeTriangles_(const Projection& P, 
                        std::vector<Image<float> > & depthmaps,
                        float x,float y, float offset, 
                        float minoffset,
                        double minCosAngle, double maxDiagRatio,
                        bool simulateOnly);

    // exports for dll (cmenk)
    //BIAS_EXPORT_STL_VECTOR( BIASUtils_EXPORT, BIAS::Vector3<double> )
    //BIAS_EXPORT_STL_VECTOR( BIASUtils_EXPORT, BIAS::Vector2<float> )
    //BIAS_EXPORT_STL_VECTOR( BIASUtils_EXPORT, BIAS::Vector<int> )

    BIAS::Image<unsigned char> texture_;
    std::vector<BIAS::Vector3<double> > meshVertices_;
    std::vector<BIAS::Vector3<double> > vertexNormals_;
    std::vector<BIAS::Vector2<float> > textureCoords_;
    std::vector<BIAS::Vector<int> > triangleIndices_;  

    /// minimum angle of any corner to accept a triangle
    double minCornerAngle_; 
    /// maximum angle between triangle normal and viewing ray
    double maxAngle_; 

    //apply this scale to depth values.
    double depthScale_;

    int currentIndex_;

  };  //end class

}//end namespace


#endif
