#ifdef WIN32
# pragma warning (disable: 4251) // STL MAP DLL warning noise (JW)
#endif

#include "VideoServer.hh"

#include <VideoSource/VideoSource_Disk.hh>

#ifdef BIAS_HAVE_DCAM
#  include <VideoSource/VideoSource_DCAM.hh>
#endif // BIAS_HAVE_DCAM

#ifdef BIAS_HAVE_V4L
# include <VideoSource/VideoSource_V4L.hh>
#endif // BIAS_HAVE_V4L

using namespace BIAS;
using namespace std;

VideoServer::VideoServer()
{
  port_ = D_CS_DEFAULT_PORT;
  targetAddress_ = "";
  camType_= 0;
  camParam_ = BIAS::Vector<double>(3);
  camParam_[0] = 15.0;// fps
  camParam_[0] = 0.02; //shutter
  camParam_[0] = 0; //gain

  imgWidth_ = 640;
  imgHeight_ = 480;
  jpeg_ = -1; // no compression
  compressionColor_=true;
  useUDP_ = false;

  resizeFactor_= 1.0;
  UDPServerThrottle_= 1;
  useUDP_ = false;

  msgSize_=0;
  Initialized_ = false;
}

VideoServer::~VideoServer()
{
  camera_->CloseDevice(); 
}


int VideoServer::Init(int port)  {
  int res;
  if (port!=port_)
    port_=port;
  cout <<"Initializing VideoServer on port:"<<port_<<endl;
  switch (camType_) {
  case 0:
    camera_= new VideoSource_Disk;
    break;
#ifdef BIAS_HAVE_DCAM
  case 1:
    camera_= new VideoSource_DCAM;
    break;
#endif // BIAS_HAVE_DCAM
#ifdef BIAS_HAVE_V4L
  case 2:
    camera_= new VideoSource_V4L;
    break;
#endif // BIAS_HAVE_V4L
  default:
    BIASERR("no camera type selected for input");
  }
 
  camera_->SetSize(imgWidth_, imgHeight_);
  //Camera_->SetLeftTop((*ImageParam_)[3], (*ImageParam_)[4]);
  camera_->SetFPS(float(camParam_[0]));
  res = camera_->OpenDevice();
  if (res<0) 
    return res;
  res = camera_->PreGrab();
  if (res<0) 
    return res;

#ifdef BIAS_HAVE_DCAM
  if ( camParam_[1]  >0) {
    camera_->AutoMode(FEATURE_SHUTTER,false);
    camera_->SetShutter(float(camParam_[1]));
    cout <<"Shutter: "<<camera_->GetShutter();
  }
  else 
    camera_->AutoMode(FEATURE_SHUTTER,true);
  if (camParam_[2]>0) {
    camera_->AutoMode(FEATURE_GAIN,false);
    camera_->SetShutter(float(camParam_[2]));
    cout <<"Shutter: "<<camera_->GetGain()<<endl;
  }
  else 
    camera_->AutoMode(FEATURE_GAIN,true);
#endif // BIAS_HAVE_DCAM

  imgWidth_=camera_->GetWidth();
  imgHeight_=camera_->GetHeight();
 
#ifdef BIAS_HAVE_LIBJPEG
  if (jpeg_ > 0) {
    if (Compressor_.Init() < 0) {
      BIASERR("Could not initialize JPEG-compression!!!");
      return -1;
    }
  } 
#else //BIAS_HAVE_LIBJPEG
  if (jpeg_>0) {
    BIASERR("Compression mode not available without LIBJPEG!");
  }
#endif //BIAS_HAVE_LIBJPEG
  if (useUDP_){
    //    Transmitter_.Connect(TargetAddress_->c_str(),*SendPort_);   
  } else {    
    server_.WaitForConnections(port_);
  }
  camera_->InitImage(CamImage_);
  cout<<"Grabbing image size: "<< imgWidth_<<"x"<< imgHeight_<<endl;
  Initialized_ = true;
  return 0;
}


int VideoServer::ProcessOneImage()  {  
  if (! Initialized_) {
    BIASERR("VideoServer not initialized");
    return -1;
  }
  if (server_.GetConnections()<=0)
    return 1;
  camera_->GrabSingle(CamImage_);
  if (resizeFactor_ == 1.0) 
    SendImage_=CamImage_;
  else {
    RescaleFilter_.SetFactor(resizeFactor_);
    RescaleFilter_.Filter(CamImage_, SendImage_);
  }
  SendImage_.SetUID(BIAS::UUID::GenerateUUID());

  //  if (*useUDP_)
    //    Transmitter_.SetThrottle(*UDPServerThrottle_);
  if (jpeg_>0) {
#ifdef BIAS_HAVE_LIBJPEG  
    // compress here --->> add parameter for jpeg-quality ???
    if (compressionColor_) {
      if (SendImage_.GetColorModel() != ImageBase::CM_RGB) {
        Image<unsigned char> temp;
        ImageConvert::ToRGB(SendImage_, temp);
        Compressor_.Compress(temp,jpeg_);
      } else {
        Compressor_.Compress(SendImage_,jpeg_);
      }
    } else {
      if (SendImage_.GetColorModel() != ImageBase::CM_Grey) {
        Image<unsigned char> temp;
        ImageConvert::ToGrey(SendImage_, temp);
        Compressor_.Compress(temp,jpeg_);
      } else {
        Compressor_.Compress(SendImage_,jpeg_);
      }
    }
    void *data = NULL;
    long length;
    Compressor_.GetCompressedData(data,length);
    if (*useUDP_){
      //      Transmitter_.Send(&Header_, (unsigned char*) data);
    } else {
      server_.SendMsg("JPEG_IMAGE", (char*)data, length);      
    }
    msgSize_=length;
#else //BIAS_HAVE_LIBJPEG
    BIASERR("LIBJPEG not configured --> no compression available !!!");
    return -1;
#endif //BIAS_HAVE_LIBJPEG
  } else {   
    if (useUDP_){
      // Transmitter_.Send(&Header_, SendImage_.GetImageData());     
    } else {    
      stringstream dataToSend;
      dataToSend<<SendImage_;   
      msgSize_= dataToSend.str().length();
      server_.SendMsg("BIAS_IMAGE",(char*)&(dataToSend.str()[0]),
                              msgSize_);
    }
  }

  return 0;
}


