/* 
This file is part of the BIAS library (Basic ImageAlgorithmS).

Copyright (C) 2003, 2004    (see file CONTACTS for details)
  Multimediale Systeme der Informationsverarbeitung
  Institut fuer Informatik
  Christian-Albrechts-Universitaet Kiel


BIAS is free software; you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation; either version 2.1 of the License, or
(at your option) any later version.

BIAS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BIAS; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef  WIN32
#  error The PMD-cameras can only be run under windows.
#endif //not WIN32

//THE header
#include "VideoSource_usbPMD.hh"

//std stuff
#include <string>
#include <sstream>
#include <iostream>

//BIAS stuff
#include <Base/Image/ImageIO.hh>
#include <Image/Camera.hh>

//#include <Base/Common/BIASpragma.hh>
//#include <Base/Common/FileHandling.hh>

//PMDtec stuff
#include <PMDDll.h>
#include <defines.hpp>

using namespace BIAS;
using namespace std;


BIAS::VideoSource_usbPMD::VideoSource_usbPMD() 
  : VideoSource()
{
  //for VideoSource-class
  VideoSource::SetSize(-1,-1,1);
  VideoSource::SetColorModel(BIAS::ImageBase::CM_Grey);
  Width_ = -1;
  Height_ = -1;

  //own standards
  type_ = MultiCam_3K;
  use2D_ = 0;
  doCalcs_ = 1;
  phaseAlgorithm_ = 4;
  modulationsFrequenz_ = 15;
  bIsOpen_ = false;
  devicePort_ = "USB";
  bitFile_ = "";
  dist_ = NULL;
  mod_ = NULL;
  grey_ = NULL;
  total_ = NULL;
  grey2D_ = NULL;

  //addept to new vals when 2d-imaging is supported
  Width2D_ = 640;
  Height2D_ = 480;
}

BIAS::VideoSource_usbPMD::~VideoSource_usbPMD()
{
  if (bIsOpen_) CloseDevice();
}

int BIAS::VideoSource_usbPMD::OpenDevice()
{
  if (bIsOpen_) {
    BIASERR("Device has been opened allready!!!");
    return -1;
  }

  if (bitFile_ == "") {
    BIASERR("No bitfile for usbPMD given!!!");
    return -1;
  }

  use2D_ = 0;
  doCalcs_ = 1;
  phaseAlgorithm_ = 4;

  //TODO: activate 2D-output -> set 2.param to 1 (docum. see PMDDll.h)
  if (z2D3D_setChip(type_, use2D_, doCalcs_, phaseAlgorithm_)<0) {
    BIASERR("Connection to PMD-Camera via USB failed!");
    return -1;
  }
  //first param ?will get? used, if more than one cam is attaced via USB (N/A now)!
  //Parameter NULL is passed for standard phase-shifts! (consult docs before using)
  if (z2D3D_init(devicePort_.c_str(), bitFile_.c_str(), type_, use2D_,
                 modulationsFrequenz_, NULL, doCalcs_, phaseAlgorithm_) < 0) {
    BIASERR("Error initializing camera!!!");
    return -1;
  }

  switch(type_) {
  case PMD_1K :
    Width_ = 64;
    Height_ = 16;
    break;
  case PMD_3K :
    Width_ = 66;
    Height_ = 48;
    break;
  case MultiCam_3K :
    Width_ = 64;
    Height_ = 48;
    break;
  case MultiCam_19K :
    Width_ = 160;
    Height_ = 120;
    break;
  default:
    BIASERR("???Unknown PMD-type but init(...) succeeded???");
    return -1;
  }

  //set 3D-exposuretime to 2[ms] (= 2/(999 + 1) * 1000)
  z2D3D_setexpo( 2, 999 );
  //set 2D-exposuretime to  5,1[ms] ( = 5ms * frametime(=1) + 10s * rowtime(=10) )
  z2D3D_SetExposureTimeInMS_2D( 10, 1 );

  //set standard application mode
  if (z2D3D_SetApplicationMode(1) < 0) {
    BIASERR("Error setting application-mode 0!!!");
    //needed if OpenDevice was not completed
    if (z2D3D_deinit() < 0) 
      BIASERR("Disonnection of PMD-Camera via USB failed!");

    return -1;
  }

  //don't mirror/rotate anything
  unsigned long mirrorRotOpts = 0;
  if (z2D3D_MirrorRotOpts(&mirrorRotOpts , 1) < 0) {
    BIASERR("Error setting mirror-modes to none!!!");
    //needed if OpenDevice was not completed
    if (z2D3D_deinit() < 0) 
      BIASERR("Disonnection of PMD-Camera via USB failed!");

    return -1;
  }

  //activate metric depth-image
  if (z2D3D_Scale(1) < 0) {
    BIASERR("Error activating metric depth-output!!!");
    //needed if OpenDevice was not completed
    if (z2D3D_deinit() < 0) 
      BIASERR("Disonnection of PMD-Camera via USB failed!");

    return -1;
  }

  //prepare mem for one image (processing image by image)
  if (z2D3D_Prepare(1) < 0) {
    BIASERR("Error reserving memory for PMD-library!!!");
    //needed if OpenDevice was not completed
    if (z2D3D_deinit() < 0) 
      BIASERR("Disonnection of PMD-Camera via USB failed!");

    return -1;
  }

  dist_ = new double[Width_*Height_*ColorChannels_]; //depth image
  mod_ = new double[Width_*Height_*ColorChannels_]; //Coefficient image
  grey_ = 
    new signed short[Width_*Height_*ColorChannels_]; //PMD-grey image
  total_ = 
    new signed short[Width2D_*Height2D_*ColorChannels_]; //not needed (raw img data)
  grey2D_ = 
    new unsigned char[Width2D_*Height2D_*ColorChannels_]; //not supported right now

  if ((dist_ == NULL)||(mod_ == NULL)||(grey_ == NULL)||
      (total_ == NULL)||(grey2D_ == NULL))
  {
    BIASERR("Error allocating memory for temp-images!!!");
    //needed if OpenDevice was not completed
    if (z2D3D_deinit() < 0) 
      BIASERR("Disonnection of PMD-Camera via USB failed!");

    return -1;
  }

  //MIP 2D-grey image (not filled right now, templated <unsigned char>)
  img2DPMD_.Init(Width_, Height_, ColorChannels_, ImageBase::ST_unsignedchar);
  img2DPMD_.SetColorModel(ImageBase::CM_Grey);
  //MIP depth image (templated <float>)
  DepthImg_.Init(Width_, Height_, ColorChannels_, ImageBase::ST_float);
  DepthImg_.SetColorModel(ImageBase::CM_Grey);
  //MIP image holding Modulation-Coeefincients (templated <float>)
  ModCoeffImg_.Init(Width_, Height_, ColorChannels_, ImageBase::ST_float);
  ModCoeffImg_.SetColorModel(ImageBase::CM_Grey);

  bIsOpen_ = true;

  return 0;
}

int BIAS::VideoSource_usbPMD::OpenDevice(int camNr)
{
  //set string for desired port
  std::stringstream str("USB");
  str << camNr;
  devicePort_ = str.str();
  //call standard open
  return OpenDevice();
}

int BIAS::VideoSource_usbPMD::CloseDevice()
{
  if (!bIsOpen_) {
    BIASERR("Cannot close device, it has not been opened successfully!!!");
    return -1;
  }

  if (z2D3D_deinit() < 0) {
    BIASERR("Disonnection of PMD-Camera via USB failed!");
    return -1;
  }

  delete dist_; //depth image
  delete mod_; //Coefficient image
  delete grey_; //PMD-grey image
  delete total_; //not needed (raw img data)
  delete grey2D_; //not supported right now

  bIsOpen_ = false;

  return 0;
}

int BIAS::VideoSource_usbPMD::PreGrab()
{
  return 0;
}


int BIAS::VideoSource_usbPMD::PostGrab()
{
	return 0;
}

int BIAS::VideoSource_usbPMD::InitImage(ImageBase &img) {
  if (!img.IsEmpty())
    BIASERR("VideoSource::InitImage() should be called with uninitialized image");
  img.Init(Width_, Height_, ColorChannels_, ImageBase::ST_unsignedchar);
  img.SetColorModel(ImageBase::CM_Grey);
  return 0;
}


int BIAS::VideoSource_usbPMD::SetShutter(float exptime){
  //exptime [ms] = value * (prescaler + 1) / 1000 --> some conversion needed
  unsigned long value, prescaler;
  prescaler = 999;
  value = (unsigned long)(1000.0 * double(exptime));
  return z2D3D_setexpo(value, prescaler);
}

int BIAS::VideoSource_usbPMD::SetShutter2D(float exptime){
  //exptime [ms] = 5ms * frametime + 10s * rowtime --> some conversion needed
  unsigned long rowtime, frametime;
  rowtime = (unsigned long)(exptime * 1000.0) / 5;
  frametime = (unsigned long)(exptime * 10000.0) % 50;
  return z2D3D_SetExposureTimeInMS_2D(rowtime, frametime);
}

int BIAS::VideoSource_usbPMD::SetModulationFrequency(int dModulationFrequency){
  unsigned long f = (unsigned long)(dModulationFrequency);
  return z2D3D_ModulationFrequency(&f, 1);
}


int BIAS::VideoSource_usbPMD::GetModulationFrequency(){
  unsigned long f;
  int res = z2D3D_ModulationFrequency(&f, 0);
  //check fo errors
  if (res < 0) return res;
  //everything's fine, return
  return (int)(f);
}

float BIAS::VideoSource_usbPMD::GetShutter(){
  unsigned long *value = new unsigned long;
  unsigned long *prescaler = new unsigned long;
  //get values from lib
  if (z2D3D_getexpo(value, prescaler) < 0)
    BIASERR("Error getting shutter from cam!");

  //convert integer values to exposuretime in sec.
  float res = float(*value) * (float(*prescaler) + 1.0f) / 1000000.0f;

  delete value;
  delete prescaler;

  return res;
}

float BIAS::VideoSource_usbPMD::GetShutter2D(){
  //exptime [ms] = 5ms * frametime + 10s * rowtime --> some conversion needed
  unsigned long *rowtime = new unsigned long;
  unsigned long *frametime = new unsigned long;
  if (z2D3D_GetExposureTimeInMS_2D(rowtime, frametime) < 0)
    BIASERR("Error getting shutter from 2D-cam!");

  //convert integer values to exposuretime in sec.
  float res = (float(*frametime) * 5.0f + 10.0f * float(*rowtime))/1000.0f;

  delete rowtime;
  delete frametime;

  return res;
}

int VideoSource_usbPMD::SetGain(float gain)
{
  BIASWARN("Setting gain not supported for PMD cam");
  return 0;
}

float VideoSource_usbPMD::GetGain(){
  BIASWARN("Getting gain not supported for PMD cam");
  return 1.0;
}

int BIAS::VideoSource_usbPMD::GrabSingle(Camera<unsigned char> &image)
{
  int res=0;

  //clac next image (-1 for all stored imgs)
  if (z2D3D_Do(-1) < 0) {
    BIASERR("Error calculating next image!!!");
    return -1;
  }

  //get next image (index 0 in memory)
  res = z2D3D_Fetch(0, dist_, mod_, grey_, total_, grey2D_);

  if (res < 0) {
    BIASERR("Error getting image data from PMD-lib!!!");
    return res;
  }

  //convert from double/short to float/unsigned char
  ImageFloatFromDouble_(dist_, DepthImg_);
  ImageFloatFromDouble_(mod_, ModCoeffImg_);
  ImageUcharFromShort_(grey_, image);

  return 0;
}

void BIAS::VideoSource_usbPMD::ImageUcharFromDouble_(double* data,
                                                     Image<unsigned char> &dst)
{ 
  if (dst.IsEmpty()) {
    dst.Init(Width_, Height_, ColorChannels_, ImageBase::ST_unsignedchar);
  }
  
  double* pus=data;
  double min = DBL_MAX;
  double max = DBL_MIN;
  for (unsigned int i=0; i<dst.GetWidth()*dst.GetHeight(); i++) {
    if (*pus<min) min = *pus;
    if (*pus>max) max = *pus;
    pus++;
  }
  double divisor = max - min;

  pus=data;
  if (divisor != 0) {
    unsigned char** puc=dst.GetImageDataArray();
    for (unsigned int y=0; y<dst.GetHeight(); y++) {
      for (int x=0; x<int(dst.GetWidth()); x++) {
        puc[y][int(dst.GetWidth())-1-x] =
          (unsigned char)(255.0*((*pus)-min)/divisor);
        pus++;
      }
    }
  } else {
    dst.SetZero();
  }
}

void BIAS::VideoSource_usbPMD::ImageUcharFromShort_(signed short* data,
                                                    Image<unsigned char> &dst)
{ 
  if (dst.IsEmpty()) {
    dst.Init(Width_, Height_, ColorChannels_, ImageBase::ST_unsignedchar);
  }
  
  signed short* pus=data;
  signed short min = *pus;
  signed short max = *pus;
  for (unsigned int i=0; i<dst.GetWidth()*dst.GetHeight(); i++) {
    if (*pus<min) min = *pus;
    if (*pus>max) max = *pus;
    pus++;
  }
  double divisor = double(max - min);

  pus=data;
  if (divisor != 0) {
    unsigned char** puc=dst.GetImageDataArray();
    for (unsigned int y=0; y<dst.GetHeight(); y++) {
      for (int x=0; x<int(dst.GetWidth()); x++) {
        puc[y][int(dst.GetWidth())-1-x] = 
          (unsigned char)(255.0*(double(*pus)-min)/divisor);
        pus++;
      }
    }
  } else {
    dst.SetZero();
  }
}

void BIAS::VideoSource_usbPMD::ImageFloatFromDouble_(double* data,
                                                     Image<float> &dst)
{ 
  if (dst.IsEmpty()) {
    dst.Init(Width_, Height_, ColorChannels_, ImageBase::ST_float);
  }
  
  double* pus=data;
  float** puc=dst.GetImageDataArray();
  for (unsigned int y=0; y<dst.GetHeight(); y++) {
    for (int x=0; x<int(dst.GetWidth()); x++) {
      puc[y][int(dst.GetWidth())-1-x] = (float)(*pus);
      pus++;
    }
  }
}
